/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef CALCULATEPOLAR_SRC_METHODS_TRIMINTERPOLATION_H_
#define CALCULATEPOLAR_SRC_METHODS_TRIMINTERPOLATION_H_

#include <moduleBasics/runtimeIO.h>

#include <memory>
#include <vector>
#include <string>
#include <utility>

class trimInterpolation{
 public:
    struct interpolationPoint {                             /**< struct which holds CL, CM and storage for numeric values which need to be interpolated */
        double CL;                                          /**< CL value of the values in the struct [-] */
        double CM;                                          /**< CM value of the values in the struct [-] */
        std::vector<double> values;                         /**< vector for single numeric values which get interpolated */
        std::vector<std::vector<double>> vectorValues;      /**< vector for vectors of numeric values which get interpolated */
    };
    /** \brief Overall function to call, which handles the trim calculation and returns trimmed values
     *  \param aPolarSet vector of the untrimmed polars stored in the interpolationPoint struct
     *  \param allowExtrapolation switch to allow extrapolation if CM(0) is outside of the polar sets
     */
    std::vector<interpolationPoint> interpolateTrim(const std::vector<std::vector<interpolationPoint>>& aPolarSet, const bool allowExtrapolation);
    /** \brief Function that cuts the polar sets to same size, so that each polar point can be interpolated
     */
    void cutCLRange();
    /** \brief Function that finds between which entrys CM=0 lies and stores the information in the interpolationPartners vector
     */
    void findInterpolationPartners();
    /** \brief Function interpolates all single numeric and vectors of numeric values in the polar sets
     */
    void interpolatePolarSet();

    trimInterpolation();
    virtual ~trimInterpolation() {}

 private:
    std::vector<std::vector<trimInterpolation::interpolationPoint>> thePolarSet;    /**< the polar set containing all untrimmed polars for a sigle flight condition */
    std::vector<trimInterpolation::interpolationPoint> theTrimmedPolar;             /**< the trimmed polar set containing the interpolated polar for a single flight condition */
    std::vector<std::pair<size_t, size_t>> interpolationPartners;             /**< vector containing the left and right index of the interpolation positions in the polar set */
    bool extrapolationAllowed;                                                      /**< switch wheter extropalation is allowed */
};

#endif  // CALCULATEPOLAR_SRC_METHODS_TRIMINTERPOLATION_H_
