/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <cmath>
#include <vector>
#include <numbers>
#include <iostream>
#include <algorithm>
#include "../include/constraint_analysis/ca_functions.h"

namespace Mattingly {

    /*The following method is based on the proposed constraint analysis of Mattingly "Aircraft Engine Design"*/
    auto constraint_analysis::constant_altitude_speed_cruise(const std::vector<double>& W_over_S_data, const std::vector<double>& C_D_data, double weight_fraction, double thrust_lapse, double Mach, double altitude) -> const std::vector<double>{

        std::vector<double> T_over_W_data;

        // Lambda function for calculating T_over_W
        auto calculate_T_over_W = [this](double W_over_S, double C_D, double weight_fraction, double thrust_lapse, double Mach, double altitude) {
            return (weight_fraction / thrust_lapse) * (((0.5 * atm.getDensity(altitude)*pow((Mach * atm.getSpeedOfSound(altitude)),2)) / weight_fraction) * (1.0 / W_over_S)) * C_D;
        };

        // For each W_over_S calculate the T_over_W
        for (size_t i = 0; i < W_over_S_data.size(); ++i) {
            double W_over_S = W_over_S_data[i];

            double C_D = C_D_data[i];

            T_over_W_data.push_back(calculate_T_over_W(W_over_S, C_D, weight_fraction, thrust_lapse, Mach, altitude));
            }
                
        return T_over_W_data;
        }

    auto constraint_analysis::constant_speed_climb(const std::vector<double>& W_over_S_data, const std::vector<double>& C_D_data, double weight_fraction, double thrust_lapse, double Mach, double altitude, double const_speed_climb_rate) -> const std::vector<double>{

        std::vector<double> T_over_W_data;

        // Lambda function for calculating T_over_W
        auto calculate_T_over_W = [this](double W_over_S, double C_D, double weight_fraction, double thrust_lapse, double Mach, double altitude, double const_speed_climb_rate) {
            double u = Mach * atm.getSpeedOfSound(altitude);
            double rho = atm.getDensity(altitude);
            double q = 0.5 * rho * pow(u, 2.0);
            return (weight_fraction / thrust_lapse) * (C_D / (weight_fraction / q * W_over_S) + 1/u * const_speed_climb_rate);
        };

        // For each W_over_S calculate the T_over_W
        for (size_t i = 0; i < W_over_S_data.size(); ++i) {
            double W_over_S = W_over_S_data[i];

            double C_D = C_D_data[i];

            T_over_W_data.push_back(calculate_T_over_W(W_over_S, C_D, weight_fraction, thrust_lapse, Mach, altitude, const_speed_climb_rate));
            }
            
        return T_over_W_data;
        }

    auto constraint_analysis::constant_altitude_speed_turn(const std::vector<double>& W_over_S_data, double K_1, double K_2, double C_D0, double weight_fraction, double thrust_lapse, double Mach, double altitude, double n) -> const std::vector<double>{
        std::vector<double> T_over_W_data;

        // Lambda function for calculating T_over_W
        auto calculate_T_over_W = [this](double W_over_S, double K_1, double K_2, double C_D0, double weight_fraction, double thrust_lapse, double Mach, double altitude, double n) {
            return (weight_fraction / thrust_lapse) * (K_1 * pow(n, 2) * weight_fraction/(0.5 * atm.getDensity(altitude)*pow((Mach * atm.getSpeedOfSound(altitude)),2)) * W_over_S + (C_D0/(weight_fraction/(0.5 * atm.getDensity(altitude)*pow((Mach * atm.getSpeedOfSound(altitude)),2)) * W_over_S)));
          };

        // For each W_over_S calculate the T_over_W
        for (double W_over_S : W_over_S_data) {
            T_over_W_data.push_back(calculate_T_over_W(W_over_S, K_1, K_2, C_D0, weight_fraction, thrust_lapse, Mach, altitude, n)); // Speichere Ergebnis in T_over_W_data
        }
            
        return T_over_W_data;
        }

    auto constraint_analysis::horizontal_acceleration(const std::vector<double>& W_over_S_data, const std::vector<double>& C_D_data, double weight_fraction, double thrust_lapse, double Mach, double altitude, double horizontal_acceleration) -> const std::vector<double>{

        std::vector<double> T_over_W_data;

        // Lambda function for calculating T_over_W
        auto calculate_T_over_W = [this](double W_over_S, double C_D, double weight_fraction, double thrust_lapse, double Mach, double altitude, double horizontal_acceleration) {
            return (weight_fraction / thrust_lapse) * ((0.5 * atm.getDensity(altitude)*pow((Mach * atm.getSpeedOfSound(altitude)),2)) / weight_fraction) * (1.0 / W_over_S) * C_D * 1/g_0 * horizontal_acceleration;
        };

        // For each W_over_S calculate the T_over_W
        for (size_t i = 0; i < W_over_S_data.size(); ++i) {
            double W_over_S = W_over_S_data[i];

            double C_D = C_D_data[i];

            T_over_W_data.push_back(calculate_T_over_W(W_over_S, C_D, weight_fraction, thrust_lapse, Mach, altitude, horizontal_acceleration));
            }

        return T_over_W_data;
        }

    auto constraint_analysis::takeoff_ground_roll(const std::vector<double>& W_over_S_data, double weight_fraction, double thrust_lapse, double Mach, double altitude, double s_G, double CL_max_TO) -> const std::vector<double>{

        std::vector<double> T_over_W_data;

        // Lambda function for calculating T_over_W
        auto calculate_T_over_W = [this](double W_over_S, double weight_fraction, double thrust_lapse, double Mach, double altitude, double s_G, double CL_max_TO) {
            return pow(weight_fraction, 2.0)/thrust_lapse * pow(k_T0, 2.0)/(s_G * rho * g_0 * CL_max_TO) * W_over_S;
        };

        // For each W_over_S calculate the T_over_W
        for (double W_over_S : W_over_S_data) {
            T_over_W_data.push_back(calculate_T_over_W(W_over_S, weight_fraction, thrust_lapse, Mach, altitude, s_G, CL_max_TO)); // Speichere Ergebnis in T_over_W_data
        }
            
        return T_over_W_data;
        }

    auto constraint_analysis::braking_roll(double C_D_L, double C_L_L, double weight_fraction, double thrust_lapse, double Mach, double altitude, double my_B, double s_G) -> const double
        {
        double W_over_S;

        // Lambda function for calculating W_over_S
        auto calculate_W_over_S = [this](double C_D_L, double C_L_L, double weight_fraction, double thrust_lapse, double Mach, double altitude, double my_B, double s_G) {
            return ((s_G * atm.getDensity(altitude)* g_0 * (C_D_L - my_B * C_L_L))/ (weight_fraction * log(1 + ((C_D_L - my_B * C_L_L)/(my_B*C_L_L/pow(k_T0,2))))));
        };

        W_over_S = calculate_W_over_S(C_D_L, C_L_L, weight_fraction, thrust_lapse, Mach, altitude, my_B, s_G);
            
        return W_over_S;
    }

    auto constraint_analysis::service_ceiling(const std::vector<double>& W_over_S_data, double K_1, double K_2, double C_D0, double weight_fraction, double thrust_lapse, double Mach, double altitude, double alt_sc) -> const std::vector<double>{

        std::vector<double> T_over_W_data;

        // Lambda function for calculating T_over_W
        auto calculate_T_over_W = [this](double W_over_S, double K_1, double K_2, double C_D0, double weight_fraction, double thrust_lapse, double Mach, double altitude, double alt_sc) {
            return (weight_fraction/thrust_lapse) * (K_1 * (weight_fraction/(0.5 * atm.getDensity(altitude)*pow((Mach * atm.getSpeedOfSound(altitude)),2))) * W_over_S + K_2 + C_D0/(weight_fraction/(0.5 * atm.getDensity(altitude)*pow((Mach * atm.getSpeedOfSound(altitude)),2))*W_over_S) + 1/(Mach * atm.getSpeedOfSound(alt_sc)) * rate_of_climb_SC);
        };

        // For each W_over_S calculate the T_over_W
        for (double W_over_S : W_over_S_data) {
            T_over_W_data.push_back(calculate_T_over_W(W_over_S, K_1, K_2, C_D0, weight_fraction, thrust_lapse, Mach, altitude, alt_sc)); // Speichere Ergebnis in T_over_W_data
        }
            
        return T_over_W_data;
    }

    auto constraint_analysis::takeoff_climb_angle(const std::vector<double>& W_over_S_data, const std::vector<double>& C_D_data, double weight_fraction, double thrust_lapse, double Mach, double altitude, double CL_max_TO, double takeoff_climb_angle) -> const std::vector<double>{

        std::vector<double> T_over_W_data;

        // Lambda function for calculating T_over_W
        auto calculate_T_over_W = [this](double W_over_S, double C_D, double weight_fraction, double thrust_lapse, double Mach, double altitude, double CL_max_TO, double takeoff_climb_angle) {
            return (weight_fraction / thrust_lapse) * (C_D/(CL_max_TO / pow(k_T0, 2.0)) + std::sin(takeoff_climb_angle * PI / 180.0));
        };

        // For each W_over_S calculate the T_over_W
        for (size_t i = 0; i < W_over_S_data.size(); ++i) {
            double W_over_S = W_over_S_data[i];

            double C_D = C_D_data[i];

            T_over_W_data.push_back(calculate_T_over_W(W_over_S, C_D, weight_fraction, thrust_lapse, Mach, altitude, CL_max_TO, takeoff_climb_angle)); // Speichere Ergebnis in T_over_W_data
        }
            
        return T_over_W_data;
    }

    auto constraint_analysis::gust(double CL_alpha, double altitude, double V_TO_L, double w_g, double dn_G, double weight_fraction) -> const double{

        double W_over_S;

        // Lambda function for calculating W_over_S
        auto calculate_W_over_S = [this](double CL_alpha, double altitude, double V_TO_L, double w_g, double dn_G, double weight_fraction) {
            return ((CL_alpha * atm.getDensity(altitude) * V_TO_L * w_g)/(dn_G * 2 * weight_fraction));
        };

        W_over_S = calculate_W_over_S(CL_alpha, altitude, V_TO_L, w_g, dn_G, weight_fraction); 
            
        return W_over_S;
    }

    auto constraint_analysis::approach_speed(double V_app, double CL_max, double weight_fraction, double s_A, double C_D, double h_obs, double k_TD, double altitude) -> const double 
    {
        double V_stall = 0.0;

        // this is a 4th order polynomial with coefficients (a*x^4 + b*x^3 + c*x^2 + d*x + e = 0)
        double a = -k_TD * CL_max;
        double b = 0.;
        double c = pow(V_app, 2.) * CL_max + 2. * h_obs * CL_max - s_A * k_TD;
        double d = 0.;
        double e = -s_A * V_app * C_D;

        double del = pow(c, 2.) - 4. * a * e;

        V_stall = std::max(std::sqrt((-c + sqrt(del)) / (2. * a)), std::sqrt((-c - sqrt(del)) / (2. * a)));        
        
        return pow(V_stall,2.)/(2.*weight_fraction/(atm.getDensity(altitude)*CL_max));
    }

    auto constraint_analysis::get_V_stall(double V_app, double k_TD, double CL_max, double CD, double h_obs, double s_A) -> const double
    {
        // this is a 4th order polynomial with coefficients (a*x^4 + b*x^3 + c*x^2 + d*x + e = 0)
        double a = -k_TD * CL_max;
        double b = 0.;
        double c = pow(V_app, 2.) * CL_max + 2. * h_obs * CL_max - s_A * k_TD;
        double d = 0.;
        double e = -s_A * V_app * CD;

        double del = pow(c, 2.) - 4. * a * e;

        return std::max(std::sqrt((-c + sqrt(del)) / (2. * a)), std::sqrt((-c - sqrt(del)) / (2. * a)));
    }

}

