/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "../constraint_analysis.h"
#include <moduleBasics/report.h>
#include <moduleBasics/html.h>
#include <fstream>
#include <sstream>
#include <vector>
#include <filesystem>
#include <iomanip>

std::vector<std::vector<std::string>> read_csv(const std::string& file_path) {
    std::ifstream csv_file(file_path);
    if (!csv_file.is_open()) {
        throw std::runtime_error("Could not open CSV file: " + file_path);
    }
    std::vector<std::vector<std::string>> csv_data;
    std::string line;
    while (std::getline(csv_file, line)) {
        std::stringstream line_stream(line);
        std::string cell;
        std::vector<std::string> row;
        while (std::getline(line_stream, cell, ',')) {
            row.push_back(cell);
        }
        csv_data.push_back(row);
    }
    return csv_data;
}

std::string format_number(const std::string& value, int precision) {
try {
    double num = std::stod(value);
    std::ostringstream stream;
    stream << std::fixed << std::setprecision(precision) << num;
    return stream.str();
} catch (...) {
    return value;
}
}

void generate_html_table(std::ostream& html_stream, const std::vector<std::vector<std::string>>& csv_data) {
    if (csv_data.empty()) return;

    // Generate table header dynamically based on the first row
    std::vector<html::Element> header_row;
    for (const auto& cell : csv_data.front()) {
        header_row.push_back(html::Element(cell));
    }
    html_stream << html::table_begin(header_row);

    // Generate table rows dynamically
    for (auto it = std::next(csv_data.begin()); it != csv_data.end(); ++it) {
        const auto& row = *it;

        if (!row.empty()) {
        html_stream << "<tr>\n";
        for (size_t i = 0; i < row.size(); ++i) {
            std::string formatted_value = row[i];
            if (i == 1) {  // 2nd column (index 1) - 4 decimal places
                formatted_value = format_number(row[i], 4);
            } else if (i == 2) {  // 3rd column (index 2) - 1 decimal place
                formatted_value = format_number(row[i], 1);
            }
            html_stream << html::table_data(formatted_value);
        }
        html_stream << "</tr>\n";
    }
    }

    html_stream << html::table_end;
}

void DefaultAnalysis::create_html_report() {
    try {
        Report report(this->get_RuntimeIO());
        std::ostream& html_stream = report.htmlReportStream();
        std::filesystem::path csv_dir = this->get_RuntimeIO()->getCsvFilesDir();

        html_stream << html::div_start("class=\"box data\"")
                    << html::h3("Final performance parameter:")
                    << "<ul>\n";

        std::string csv_file_path = (csv_dir / "design_point.csv").string();
        auto csv_data = read_csv(csv_file_path);
        generate_html_table(html_stream, csv_data);

        html_stream << html::div_end;

        if (this->get_RuntimeIO()->plotOn) {
            std::filesystem::path plot_dir_relative =
                std::filesystem::relative(this->get_RuntimeIO()->getPlotDir(),
                                        this->get_RuntimeIO()->getHtmlDir());
            std::string plot_path = (plot_dir_relative / "constraint_plot.svg").string();

            html_stream << html::div_start("class=\"box plot\"")
                        << html::h2("Design chart")
                        << html::image(plot_path, "width=\"800px\"")
                        << html::div_end;
        }

        report.generateHtmlReport("");
        myRuntimeInfo->out << "HTML report created." << std::endl;

    } catch (const std::exception& e) {
        myRuntimeInfo->out << "Error creating HTML report: " << e.what() << std::endl;
    }
}
