# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing functions for data postprocessing."""
# Import standard modules.

# Import own modules.
from datapostprocessingmodule import method_data_postprocessing
from datapostprocessingmodule import write_key_data_to_aircraft_exchange_file
from datapostprocessingmodule import prepare_element_tree_for_module_key_parameter


def data_postprocessing(paths_and_names, routing_dict, data_dict, runtime_output):
    """Perform data postprocessing and write data to an aircraft exchange file.

    This function is responsible for data postprocessing that involves the following steps:
        (1) Data preparation: The module manager prepares a list containing all paths to the key parameters that must
        be written to the aircraft exchange file.
        (2) Write data to the aircraft exchange file: The results of the module execution are passed on to the function
        responsible for properly writing the data to the aircraft exchange file.
        (3) Method-specific data postprocessing: User-defined method-specific postprocessing is conducted as needed.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict data_dict: Dictionary containing the result of the module execution (direct operating cost estimation)
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return: None
    """

    """Data preparation."""
    # Changes to this list are the sole responsibility of the module manager!
    paths_to_key_parameters_list = [
        './assessment/cost_estimation/operating_cost/direct_operating_cost/flights_per_year_design_mission',
        './assessment/cost_estimation/operating_cost/direct_operating_cost/route_independent_cost_annual',
        './assessment/cost_estimation/operating_cost/direct_operating_cost/route_dependent_cost_annual',
        './assessment/cost_estimation/operating_cost/direct_operating_cost/direct_operating_cost_annual'
    ]

    module_key_parameters_dict = {
        'assessment': {
            'cost_estimation': {
                'attributes': {'description': 'Cost estimation results',
                               'tool_level': '0'
                               },
                'operating_cost': {
                    'attributes': {'description': 'Operating cost estimation results',
                                   'tool_level': '0'
                                   },
                    'direct_operating_cost': {
                        'attributes': {'description': 'Direct operating cost (sum of route independent and route dependent cost)',
                                    'tool_level': '0'
                                    },
                        'flights_per_year_design_mission': {
                            'attributes': {'description': 'Flights per year (design mission)'},
                            'value': '0',
                            'unit': '1',
                            'lower_boundary': '0',
                            'upper_boundary': '100000'
                        },
                        'route_independent_cost_annual': {
                            'attributes': {'description': 'Route independent (fixed) cost per year'},
                            'value': '0',
                            'unit': 'USD/a',
                            'lower_boundary': '0',
                            'upper_boundary': 'inf'
                        },
                        'route_dependent_cost_annual': {
                            'attributes': {'description': 'Route dependent (variable) cost per year'},
                            'value': '0',
                            'unit': 'USD/a',
                            'lower_boundary': '0',
                            'upper_boundary': 'inf'
                        },
                        'direct_operating_cost_annual': {
                            'attributes': {'description': 'Direct operating cost (DOC) per year'},
                            'value': '0',
                            'unit': 'USD/a',
                            'lower_boundary': '0',
                            'upper_boundary': 'inf'
                        }
                    }
                }
            }
        }
    }
    
    # Check, if study mission exists.
    study_exists = ('study' in data_dict.keys())
    if study_exists:
        paths_to_key_parameters_list.append(
            './assessment/cost_estimation/operating_cost/direct_operating_cost/flights_per_year_study_mission')
        module_key_parameters_dict['assessment']['cost_estimation']['operating_cost']['direct_operating_cost'][
            'flights_per_year_study_mission'] = {
                    'attributes': {'description': 'Flights per year (study mission)'},
                    'value': '0',
                    'unit': '1',
                    'lower_boundary': '0',
                    'upper_boundary': '100000'
                }

    paths_and_names = prepare_element_tree_for_module_key_parameter(paths_and_names, module_key_parameters_dict)

    # Run 'user_method_data_output_preparation' from 'usermethoddatapreparation.py'.
    key_output_dict, method_specific_output_dict = routing_dict['func_user_method_data_output_preparation'](data_dict)
    # Extract tool level from routing dictionary.
    tool_level = routing_dict['tool_level']

    """Write data to aircraft exchange file."""
    # Extract root and path to aircraft exchange file.
    root_of_aircraft_exchange_tree = paths_and_names['root_of_aircraft_exchange_tree']
    path_to_aircraft_exchange_file = paths_and_names['path_to_aircraft_exchange_file']
    # Write key data to aircraft exchange file.
    write_key_data_to_aircraft_exchange_file(root_of_aircraft_exchange_tree, path_to_aircraft_exchange_file,
                                             paths_to_key_parameters_list, key_output_dict, tool_level, runtime_output)

    """Method-specific postprocessing."""
    # Run 'method_data_postprocessing' from 'datapostprocessingmodule'.
    method_data_postprocessing(paths_and_names, routing_dict, data_dict,
                               method_specific_output_dict, runtime_output)
