# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing plotting functionalities for current calculation method."""
# Import standard libraries.
import os
from matplotlib import pyplot as plt
import numpy as np


def method_plot(paths_and_names, routing_dict, data_dict, method_specific_output_dict, runtime_output):
    """Plot function.

    This function is responsible for creating plots.
    [Add further information here...]

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict data_dict: Dictionary containing results of module execution
    :param dict method_specific_output_dict: Dictionary containing method specific output data
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return: None
    """
    runtime_output.print('Plots are generated and saved...')

    # Extract necessary values from 'data_dict'.
    prd_range_array = data_dict['design']['payload_range_diagram_data']['ranges_for_prd']/1000
    prd_payload_array = data_dict['design']['payload_range_diagram_data']['payloads_for_prd']
    seat_kilometer_costs_design = data_dict['design']['seat_kilometer_costs']['seat_kilometer_costs']
    ton_kilometers_offered_design = data_dict['design']['ton_kilometer_costs']['ton_kilometer_offered']
    ton_kilometer_costs_design = data_dict['design']['ton_kilometer_costs']['ton_kilometer_costs']

    """ Plot two values in the same diagram. """
    create_two_axes_plot(seat_kilometer_costs_design, 300, "SKC in USD/number of PAX * km * year",
                         ton_kilometers_offered_design, max(ton_kilometers_offered_design) * 1.1,
                         "Ton kilometers offered in payload weight * km",
                         prd_range_array, "SKC and TKO for design mission",
                         "design_mission", paths_and_names)

    create_two_axes_plot(ton_kilometer_costs_design, 3, "DOC per ton kilometers offered in USD/payload weight * km",
                         ton_kilometers_offered_design, max(ton_kilometers_offered_design) * 1.1,
                         "Ton kilometers offered in payload weight * km",
                         prd_range_array, "DOC per ton kilometers offered (TKO) and TKO for design mission",
                         "design_mission", paths_and_names)

    create_two_axes_plot(prd_payload_array, max(prd_payload_array) * 1.1, "Payload in kg",
                         ton_kilometer_costs_design, 3, "DOC per ton kilometers offered in USD/payload weight * km",
                         prd_range_array, "Payload range diagram and SKC for design mission",
                         "design_mission", paths_and_names)


def create_two_axes_plot(y_value_1, y_limit_1, y_label_1, y_value_2, y_limit_2, y_label_2, range_array, plot_title,
                         mission_flag, paths_and_names):
    """Create plot with two y axis.

    This function creates a plot with two different y axis.

    :param y_value_1:
    :param y_limit_1:
    :param y_label_1:
    :param y_value_2:
    :param y_limit_2:
    :param y_label_2:
    :param range_array:
    :param plot_title:
    :param mission_flag:
    :param paths_and_names:
    :return:
    """
    """ Plot seat mile costs and transport kilometers offered in same diagram. """
    fig1, ax1 = plt.subplots()
    color = '#C50F1F'
    ax1.set_xlabel('Range in km' + r'$\rightarrow$')
    ax1.set_ylabel(y_label_1 + r'$\rightarrow$', color=color)
    ax1.set_xlim(0, max(range_array))
    ax1.set_ylim(0, y_limit_1)
    if np.ndim(y_value_1) == 2:
        if mission_flag == "design_mission":
            ax1.plot(range_array, y_value_1[0], color=color)
        else:
            ax1.plot(range_array, y_value_1[1], color=color)
    else:
        ax1.plot(range_array, y_value_1, color=color)
    ax1.tick_params(axis='y', labelcolor=color)
    ax1.grid(color='lightgray', linestyle=':', linewidth=1)
    ax1.ticklabel_format(style='plain', axis='x')

    ax2 = ax1.twinx()  # instantiate a second axes that shares the same x-axis
    color = 'darkgray'
    ax2.set_ylabel(y_label_2 + r'$\rightarrow$', color=color)  # we already handled the x-label with ax1
    ax2.set_ylim(0, y_limit_2)
    if np.ndim(y_value_2) == 2:
        if mission_flag == "design_mission":
            ax2.plot(range_array, y_value_2[0], color=color)
        else:
            ax2.plot(range_array, y_value_2[1], color=color)
    else:
        ax2.plot(range_array, y_value_2, color=color)
    ax2.tick_params(axis='y', labelcolor=color)

    fig1.tight_layout()  # otherwise the right y-label is slightly clipped
    fig1.subplots_adjust(bottom=0.1, top=0.9)
    plt.title(plot_title, pad=15)
    plt.grid(which='major', axis='x')

    """ Save figures to current aircraft project directory. """
    # Check path to directory for figure outputs
    project_directory = paths_and_names["project_directory"]
    plot_directory = project_directory + '/reporting/plots'
    if not os.path.isdir(plot_directory):
        os.mkdir(plot_directory)
    # Save as *.svg
    tool_name = paths_and_names['tool_name']
    name_of_fig = tool_name + '_' + plot_title + '.svg'
    plt.savefig(plot_directory + '/' + name_of_fig, transparent=True)
    # Save as *.pdf
    name_of_fig = tool_name + '_' + plot_title + '.pdf'
    plt.savefig(plot_directory + '/' + name_of_fig, transparent=True)

    return
