/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_MISSION_EMISSIONS_MISSION_H_
#define ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_MISSION_EMISSIONS_MISSION_H_

#include <string>
#include <vector>
#include <memory>
#include "../ecoDatabase.h"
#include "../../emissionsClass.h"
#include "flightPath.h"
#include "missionIOData.h"
#include <atmosphere/atmosphere.h>
#include <moduleBasics/runtimeIO.h>
#include <energyCarriers/energyCarriers.h>

class MissionIOData;

/** \class Mission
 * \brief Class contains the properties of the mission emissions which are then used within the operational phase
 */
class Mission {
 private:
    std::string mission_csv_name;/**< Name of mission csv file */

 public:
    /** \brief Constructor of class mission
     */
    Mission();
    /** \brief Destructor of class mission
     */
    virtual ~Mission();

    std::shared_ptr<MissionIOData> missionIOData; /**< IOData object for mission emission calculation*/; /**< Pointer to ioData */
    std::shared_ptr<RuntimeIO> rtIOPt; /**< Pointer to runtime IO*/
    atmosphere myAtmosphere;/**< Atmosphere object */
    std::string mission_type; /**< study or design  */
    double delta_isa;/**< Temperature delta to ISA-Conditions [K] */
    const ecoDatabase myEcoDatabase;/**Ecodatabase object*/

    /** \class MissionData
     * \brief Class contains all the mission information at one mission step
     */
    class MissionData {
     public:
        double time;/**< Time [s] */
        double range;/**< Range [m] */
        double altitude;/**< Altitude [m] */
        double TAS_ms;/**< True airspeed [m/s] */
        double CAS_ms;/**< Calibrated airspeed [m/s] */
        double mass;/**< Mass [kg] */
        double fuelConsumed;/**< Consumed fuel until this mission point [kg]  */
        double thrust;/**< Thrust [N]  */
        double fuelFlow;/**< Fuel flow [kg/s]  */
        double TAS_kts;/**< True airspeed [kts] */
        double Mach;/**<Mach number [-] */
        double flight_level;/**<Flight level [100ft]  */
        double ROC;/**< Rate of climb [ft/min]  */
        double C_L;/**< Lift coefficient [-]  */
        double SAR;/**< Specific air range [m/kg] */
        std::string aeroConfig;/**< Aerodynamic configuration (Takeoff, Climb, Clean, Approach, ApproachLG) [-] */
        double bleedAirOfftake;/**< Bleed air offtakes [kg/s]*/
        double shaftPowerOfftake;/**< Shaft power offtakes [W]  */
        std::string engineRating;/**< Current rating of the engine (e.g. TO, Cruise, MaxCont) [-] */
        double AOA;/**<Angle of attack [deg]  */
        double spoilerFactor;/**< Spoiler factor [-] */
        std::string modeName;/**< Current flight mode (e.g. TO, Accelerate,Cruise) [-] */

        double Tamb;                /**< Ambient temperature [K]*/
        double Pamb;                /**< Ambient pressure [Pa]*/
        double NOxFlow;             /**< NOx flow [kg/s]*/
        double N1;                  /**< Engine speed compared to nominal speed [-]*/
        double St3MassFlow;         /**< Mass flow at last compressor exit, cold side heat exchanger inlet [kg/s]*/
        double St3TotalTemperature; /**< Total temperature at last compressor exit, cold side heat exchanger inlet [K]*/
        double St3StaticTemperature;/**< Static temperature at last compressor exit, cold side heat exchanger inlet [K]*/
        double St3TotalPressure;    /**< Total pressure at last compressor exit, cold side heat exchanger inlet [kPa]*/
        double St3Mach;             /**< Mach at last compressor exit, cold side heat exchanger inlet [K]*/
        double St4FuelAir;          /**< Fuel/Air at last compressor exit, cold side heat exchanger inlet [-]*/
        double BPR;                 /**< By pass ratio [-] */

        /** \brief Constructor of class MissionData
         */
        MissionData();
        /** \brief Destructor of class MissionData
         */
        virtual ~MissionData() {}
    };
    std::vector<MissionData> missionInfo;/**< Mission information */

    flightPath myFlightPath;/**< Object of class fight path that contains flight path points information */
    double mission_range;/**< Range of the mission [NM]*/
    int cruise_number_of_steps; /**< Cruise Steps [-]**/
    std::vector<double> relative_end_of_step; /**< Step lengths of the Cruise Steps of the mission*/
    std::vector<double> flight_level; /**< Flight level of the Cruise Steps of the Mission*/

    double fuel_mass_taxi;/**< Taxi fuel [kg]*/
    double fuel_mass_takeoff;/**< Fuel for takeoff [kg]*/
    double fuel_mass_climbout;/**< Fuel for climb-out [kg] */
    double fuel_mass_approach;/**< Fuel for approach [kg] */

    double fuel_mass_taxi_ICAO;/**< Consumed fuel for taxiing acc. to ICAO engine data [kg] */
    double fuel_mass_takeoff_ICAO;/**< Consumed fuel for takeoff acc. to ICAO engine data [kg] */
    double fuel_mass_climbout_ICAO;/**< Consumed fuel for climb acc. to ICAO engine data [kg] */
    double fuel_mass_approach_ICAO;/**< Consumed fuel for approach acc. to ICAO engine data [kg] */

    double fuel_mass_LTO;/**< Consumed fuel for LTO cycle [kg]*/
    double fuel_mass_cruise;/**< Consumed fuel during cruise [kg] */

    double taxi_time;/**< Taxi time [s]*/
    double takeoff_time;/**<Takeoff time [s] */
    double climbout_time;/**<Climb-out time [s] */
    double approach_time;/**< Approach time [s]*/
    double LTO_time; /**< Time for taxi, T/O, C/O and App. (LTO) [min]*/
    std::vector<double> EINOxKerosene;/**<Vector containing emission index for kerosene cumbustion in all mission points [kg/kg]*/

    double thrust_taxi;/**< Thrust for taxi [N]*/
    double thrust_takeoff;/**<Thrust for takeoff [N] */
    double thrust_climbout;/**< Thrust for climb-out [N]*/
    double thrust_approach;/**<Thrust for approach [N] */

    double T3_ID; /**<Temperature at engine stage 3 for idle thrust [K] */
    double T3_TO; /**<Temperature at engine stage 3 for take-off thrust [K] */
    double T3_CL; /**<Temperature at engine stage 3 for climb out thrust [K] */
    double T3_AP; /**< Temperature at engine stage 3 for approach thrust [K]*/

    double P3_ID; /**<Static pressure at engine stage 3 for idle thrust[kPa] */
    double P3_TO; /**<Static pressure at engine stage 3 for take-off thrust [kPa] */
    double P3_CL; /**<Static pressure at engine stage 3 for climb out thrust [kPa] */
    double P3_AP; /**<Static pressure at engine stage 3 for approach thrust [kPa] */

    double P3t_ID; /**<Total pressure at engine stage 3 for idle thrust [kPa] */
    double P3t_TO; /**<Total pressure at engine stage 3 for take-off thrust [kPa] */
    double P3t_CL; /**<Total pressure at engine stage 3 for climb out thrust [kPa] */
    double P3t_AP; /**<Total pressure at engine stage 3 for approach thrust [kPa] */

    double St3Mach_ID; /**<Mach number in stage 3 for idle thrust [-]*/
    double St3Mach_TO; /**<Mach number in stage 3 for takeoff thrust [-]*/
    double St3Mach_CL; /**<Mach number in stage 3 for climb out thrust [-]*/
    double St3Mach_AP; /**<Mach number in stage 3 for approach thrust [-}*/

    double St3MassFlow_ID; /**<Air mass flow in stage 3 for idle thrust [kg/s] */
    double St3MassFlow_TO; /**<Air mass flow in stage 3 for takeoff thrust [kg/s] */
    double St3MassFlow_CL; /**<Air mass flow in stage 3 for climb out thrust [kg/s] */
    double St3MassFlow_AP; /**<Air mass flow in stage 3 for approach thrust [kg/s] */

    double fuelFlow_ID; /**< Fuel flow in stage 3 for idle thrust [kg/s] */
    double fuelFlow_TO; /**< Fuel flow in stage 3 for takeoff thrust [kg/s] */
    double fuelFlow_CL; /**< Fuel flow in stage 3 for climb out thrust [kg/s] */
    double fuelFlow_AP; /**< Fuel flow in stage 3 for approach thrust [kg/s] */

    int number_of_steps;/**< Number of path points [-]*/
    double time_step;/**< Time between path points [-] */
    double flights_per_lifecycle;/**< Number of flights per life cycle [-] */

    /* Emissions */
    Emissions LTO;/**<LTO cycle emissions for one mission */
    Emissions LTO_Total;/**<LTO cycle emissions for entire operating phase over entire life cycle */
    Emissions ICAO_LTO;/**LTO cycle emissions for ICAO LTO cycle */
    Emissions cruise;/**<Emissions during cruise */
    Emissions cruiseTotal;/**<Emissions during cruise for entire operating phase over entire life cycle */
    std::vector<Emissions> emissionsFlow;/** Vector contains emission flow in every mission step defined in mission.csv*/
    double c_soot_LTO_max;/**< Maximum soot concentration during Landing and Takeoff [mg/kg]*/

    /** \brief Executes mission calculation
     * \param missionType Typ of mission (design or study)
     * \param rtIO Access to runtime IO
     * \param rtIO Access to ecological_assessment_results.xml
     */
    void run(const std::string& missionType, const std::shared_ptr<RuntimeIO>& rtIO, std::shared_ptr<node> ecoAssXML);

   /** \brief Mission calculaton routine
    * \param missionType Typ of mission (design or study)
    * \param rtIO Access to runtime IO
    */
    void calculate_mission(const std::string& missionType, const std::shared_ptr<RuntimeIO>& rtIO);

   /** \brief Compares calculated LTO values with ICAO data
    */
    void LTO_output_comparison();

    /** \brief Pulls data from mission_file.csv into tmpMissionData
     */
    void get_mission_data();

    /** \brief Calculates from mission_file.csv data the times and fuel consumption of the individual mission segments
     */
    void get_mission_segment_data();

    /** \brief Writes emissions path to csv file
     */
    void write_emissions_path_csv();

    /** \brief Get LTO engine thermodynamics
     */
    void get_engine_thermodynamics_LTO();

    /** \brief Calculates emissions of kerosene combustion
     */
    void calculate_kerosene_emissions();

   /** \brief Uses P3-T3 Method to calculate NOx
     */
    void kerosene_NOx_P3T3();

    /** \brief Calculates the NOx emissions from NOx flow (gas turb data) and time step
     */
    void kerosene_NOx_gasTurb();

    /** \brief Uses Boeing Fuel Flow Method to calculate NOx \n
     * \cite Sch13
     */
    void kerosene_NOx_BFM2();

    /** \brief Uses Boeing Fuel Flow Method to calculate HC \n
     * \cite Sch13
     */
    void kerosene_HC_BFM2();

    /** \brief Uses Boeing Fuel Flow Method to calculate CO \n
     * \cite Sch13
     */
    void kerosene_CO_BFM2();

    /** \brief Uses DLR Omega method to calculate HC \n
     * \cite Eye04, \cite Pet05
     */
    void kerosene_HC_DLR();

   /** \brief Uses DLR Omega method to calculate CO \n
     * \cite Eye04, \cite Pet05
     */
    void kerosene_CO_DLR();

    /** \brief Uses DLR correlation based on Smoke Number to calculate soot \n
     * \cite Kug05
     */
    void kerosene_soot_DLR();

    /** \brief Uses Whyte Correlation method to calculate soot
     */
    void kerosene_soot_whyte();

    /** \brief Calculates emissions of hydrogen combustion
     */
    void calculate_hydrogen_combustion_emissions();


    /** \brief  Function calculats NOx emissions based on constant emission indey per mission segment \n
     * \cite Koss22
     */
    void hydrogen_combustion_NOx_constants();

    /** \brief  Function scales NOx emissions for hydrogen based on kerosene engine */
    void hydrogen_combustion_NOx_P3T3();

    /** \brief Calculates scaling factor for hydrogen_combustion_NOx_P3T3() function
     * \return calculated scaling factor [-]
     * \cite Mar05
     */
    double calculateNOxHydrogenFactor();

    /** \brief Linear interpolation of the ICAO data from NOx
     * \param REI_TO: EI for takeoff [kg/kg]
     * \param REI_CL: EI for climb [kg/kg]
     * \param REI_AP: EI for approach [kg/kg]
     * \param REI_ID: EI for taxi (engine idle) [kg/kg]
     * \param mission_point: mission point [kg/kg]
     * \return REI: Reference emission index [kg/kg]
     */
    double point_to_point_fit(double REI_TO, double REI_CL, double REI_AP, double REI_ID, int mission_point);

    /** \brief Least squares approximation of linear functions to emission index data
     * \param REI_TO: EI for takeoff [kg/kg]
     * \param REI_CL: EI for climb [kg/kg]
     * \param REI_AP: EI for approach [kg/kg]
     * \param REI_ID: EI for taxi (engine idle) [kg/kg]
     * \param mission_point: mission point [kg/kg]
     * \return REI: Reference emission index [kg/kg]
     */
    double linear_least_square_fit(double REI_TO, double REI_CL, double REI_AP, double REI_ID, int mission_point);

    /** \brief Least squares approximation of bi-linear functions to emission index data
     * \param REI_TO: EI for takeoff [kg/kg]
     * \param REI_CL: EI for climb [kg/kg]
     * \param REI_AP: EI for approach [kg/kg]
     * \param REI_ID: EI for taxi (engine idle) [kg/kg]
     * \param mission_point: mission point [kg/kg]
     * \return REI: Reference emission index [kg/kg]
     */
    double bi_linear_least_square_fit(double REI_TO, double REI_CL, double REI_AP, double REI_ID, int mission_point);

    /** \brief potential_fit creates a graphic in the form y=ax^b
     * \param x_mom: x for mission point [K] or [kPa]
     * \param x_ID: x for taxi (engine idle) [K]
     * \param x_TO: x for takeoff [K]
     * \param x_CL: x for climb [K]
     * \param x_AP: x for approach [K]
     * \param y_ID: y for taxi (engine idle) [kPa]
     * \param y_TO: y for takeoff [kPa]
     * \param y_CL: y for climb [kPa]
     * \param y_AP: y for approach [kPa]
     * \param Y: when Y==false "mission point value" should be a value in "Y-axis" and the value returned will be a value from "X-axis" [-]
     * \return x_mom or y_mom: searched value [K] or [kPa]
     */
    double potential_fit(double x_mom, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP, bool Y);

    /** \brief potential_fit_2 creates a graphic in the form y=ax^b
     * \param x_mom: x for mission point [K] or [kPa]
     * \param x_ID: x for taxi (engine idle) [K]
     * \param x_TO: x for takeoff [K]
     * \param x_CL: x for climb [K]
     * \param x_AP: x for approach [K]
     * \param y_ID: y for taxi (engine idle) [kPa]
     * \param y_TO: y for takeoff [kPa]
     * \param y_CL: y for climb [kPa]
     * \param y_AP: y for approach [kPa]
     * \param Y: when Y==false "mission point value" should be a value in "Y-axis" and the value returned will be a value from "X-axis" [-]
     * \return x_mom or y_mom: searched value [K] or [kPa]
     */
    double potential_fit_2(double x_mom, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP, bool Y);

    /** \brief potential_fit_3 creates a graphic in the form y=ax^b
     * \param x_mom: x for mission point [K] or [kPa]
     * \param x_ID: x for taxi (engine idle) [K]
     * \param x_TO: x for takeoff [K]
     * \param x_CL: x for climb [K]
     * \param x_AP: x for approach [K]
     * \param y_ID: y for taxi (engine idle) [kPa]
     * \param y_TO: y for takeoff [kPa]
     * \param y_CL: y for climb [kPa]
     * \param y_AP: y for approach [kPa]
     * \param Y: when Y==false "mission point value" should be a value in "Y-axis" and the value returned will be a value from "X-axis" [-]
     * \return x_mom or y_mom: searched value [K] or [kPa]
     */
    double potential_fit_3(double x_mom, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP, bool Y);

    /** \brief quadratic_fit creates a graphic in the form y=a+bx+cx^2
     * \param x_mom: x for mission point [K] or [kg/kg]
     * \param x_ID: x for taxi (engine idle)[K]
     * \param x_TO: x for takeoff [K]
     * \param x_CL: x for climb [K]
     * \param x_AP: x for approach [K]
     * \param y_ID: y for taxi (engine idle) [kg/kg]
     * \param y_TO: y for takeoff [kg/kg]
     * \param y_CL: y for climb [kg/kg]
     * \param y_AP: y for approach [kg/kg]
     * \param Y: when Y==false "mission point value" should be a value in "Y-axis" and the value returned will be a value from "X-axis"
     * \return y_mom: searched value [K] or [kg/kg]
     */
    double quadratic_fit(double x_mom, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP, bool Y);

    /** \brief calculates values via linear interpolation
     * \param x: x for mission point
     * \param x_ID: x for taxi (engine idle) [kg/s]
     * \param x_TO: x for takeoff [kg/s]
     * \param x_CL: x for climb [kg/s]
     * \param x_AP: x for approach [kg/s]
     * \param y_ID: y for taxi (engine idle) [mg/m^3]
     * \param y_TO: y for takeoff [mg/m^3]
     * \param y_CL: y for climb [mg/m^3]
     * \param y_AP: y for approach [mg/m^3]
     * \return y_mom: searched value [mg/m^3]
     */
    double linear_interpolation(double x, double x_ID, double x_TO, double x_CL, double x_AP, double y_ID, double y_TO, double y_CL, double y_AP);
};

#endif // ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_MISSION_EMISSIONS_MISSION_H_
