/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <iostream>
#include <memory>
#include <algorithm>
#include <string>
#include <unitConversion/unitConversion.h>
#include <moduleBasics/runtimeIO.h>
#include <aixml/node.h>
#include "dallara.h"


void ClimateModelDallara::run(const std::shared_ptr<RuntimeIO> &rtIO, std::shared_ptr<node> ecoAssXML) {
    myRuntimeInfo->out << "  Calculation of climate impact" << std::endl;
    dallaraIOData = std::make_shared<DallaraIOData>(rtIO->acxml, rtIO->moduleConfig, ecoAssXML);
    calculate_climate_impact_dallara(rtIO->aircraft_energy_carrier_type());
}

void ClimateModelDallara::calculate_climate_impact_dallara(const std::string &fuelType) {
    /* ATR calculation s. Gl. (20) Reference Dallara*/
    for (int emissionNo = 0; emissionNo < 8; emissionNo++) {
        emissions_per_year.push_back(dallaraIOData->flights_per_year * dallaraIOData->emissionMass[emissionNo]);
        get_average_forcing_factors(emissionNo);
        calculate_radiative_forcing(emissionNo, fuelType);
        calculate_temperature_change(emissionNo);
        calculate_agtp_and_agwp(emissionNo);
        calculate_average_temperature_response(emissionNo);
    }
    calculate_total_impact();
}

void ClimateModelDallara::get_average_forcing_factors(int emissionNo) {
    double forcing_factor_cruise(0.);//Cruise average forcing factors [-]
    double forcing_factor_climb(0.);//Climb average forcing factors [-]
    double forcing_factor_descent(0.);//Descent average forcing factors [-]
    /* forcing_factor_cruise (summation of weighted cruise steps)*/
    double relative_step_length;
    for (unsigned int j = 0; j < dallaraIOData->flight_level.size(); j++) {
        if (j == 0) {
            relative_step_length = dallaraIOData->relative_end_of_cruise_step.at(j);
        } else {
            relative_step_length = dallaraIOData->relative_end_of_cruise_step.at(j) - dallaraIOData->relative_end_of_cruise_step.at(j-1);
        }
        forcing_factor_cruise += relative_step_length * get_forcing_factor(emissionNo, dallaraIOData->flight_level.at(j) * 100.);
    }
    myRuntimeInfo->info << "forcing_factor_cruise FINAL: " << forcing_factor_cruise << std::endl;
    /* forcing_factor_climb and forcing_factor_descent (integrate)*/
    uint16_t step_size(100);
    double altitude_step_climb = dallaraIOData->flight_level.front() * 100. / step_size; // [ft]
    double altitude_step_descent = dallaraIOData->flight_level.back() * 100. / step_size; // [ft]
    double altitude_climb_tmp = 0.; // [ft]
    double altitude_descent_tmp = dallaraIOData->flight_level.back() * 100.; // [ft]
    for (uint16_t j = 0; j <= step_size; j++) {
        // trapezoidal rule
        if (j == 0 || j == step_size) {
            forcing_factor_climb += 0.5 * get_forcing_factor(emissionNo, altitude_climb_tmp) * altitude_step_climb;
            forcing_factor_descent += 0.5 * get_forcing_factor(emissionNo, altitude_descent_tmp) * altitude_step_descent;
            myRuntimeInfo->info << "get_forcing_factor_height(i,altitude_descent_tmp): " << get_forcing_factor(emissionNo, altitude_descent_tmp) << std::endl;
            myRuntimeInfo->info << "altitude_descent_tmp: " << altitude_descent_tmp << std::endl;
            myRuntimeInfo->info << "forcing_factor_descent: " << forcing_factor_descent << std::endl;
        } else {
            forcing_factor_climb += get_forcing_factor(emissionNo, altitude_climb_tmp) * altitude_step_climb;
            forcing_factor_descent += get_forcing_factor(emissionNo, altitude_descent_tmp) * altitude_step_descent;
            myRuntimeInfo->info << "get_forcing_factor_height(i,altitude_descent_tmp): " << get_forcing_factor(emissionNo, altitude_descent_tmp) << std::endl;
            myRuntimeInfo->info << "altitude_descent_tmp: " << altitude_descent_tmp << std::endl;
            myRuntimeInfo->info << "forcing_factor_descent: " << forcing_factor_descent << std::endl;
        }
        altitude_climb_tmp += altitude_step_climb;
        altitude_descent_tmp -= altitude_step_descent;
    }
    forcing_factor_climb = forcing_factor_climb / ((dallaraIOData->flight_level.front() - 0.) * 100.);
    forcing_factor_descent = forcing_factor_descent / ((dallaraIOData->flight_level.back() - 0.) * 100.);
    myRuntimeInfo->info << "forcing_factor_climb FINAL: " << forcing_factor_climb << std::endl;
    myRuntimeInfo->info << "forcing_factor_descent FINAL: " << forcing_factor_descent << std::endl;
    double climb_ratio = dallaraIOData->top_of_climb_range / dallaraIOData->range_study;
    double cruise_ratio = (dallaraIOData->top_of_descent_range - dallaraIOData->top_of_climb_range) / dallaraIOData->range_study;
    double descent_ratio = (dallaraIOData->range_study - dallaraIOData->top_of_descent_range) / dallaraIOData->range_study;
    forcing_factor_average[emissionNo] = cruise_ratio*forcing_factor_cruise + climb_ratio*forcing_factor_climb + descent_ratio*forcing_factor_descent;
    myRuntimeInfo->info << "forcing_factor_average FINAL: " << forcing_factor_average[emissionNo] << std::endl;
}

void ClimateModelDallara::calculate_radiative_forcing(int emissionNo, const std::string &fuelType) {
    /* Calculate RF und RF_modified (s. boundary conditions eq.(19))*/
    std::vector<double> radiative_forcing_tmp; // Per unit emission for every year
    std::vector<double> normalized_radiative_forcing_tmp; // Per unit emission for every year
    std::vector<double> annual_radiative_forcing_tmp; // Absolute value of emissionNo for every year
    std::vector<double> annual_normalized_radiative_forcing_tmp; // Absolute value of emissionNo for every year
    for (uint16_t t = 0; t <= dallaraIOData->configuration.tmax; t++) {
        radiative_forcing_tmp.push_back(0.);
        normalized_radiative_forcing_tmp.push_back(0.);
        if (emissionNo== 0 || emissionNo== 1 || emissionNo== 2) { //RF to be calculated via time integral at CO2, CH4 and O3L
            for (int tau = 0; tau <= t; tau++) {
                double sigma = (tau < dallaraIOData->duration_operation) ? 1 : 0;
                if (tau == 0 || tau == t) { //step_size is 1 year!
                    radiative_forcing_tmp.back() += 0.5 * get_response_function(emissionNo, (t - tau), fuelType) * sigma;
                    normalized_radiative_forcing_tmp.back() = efficacy[emissionNo] * radiative_forcing_tmp.back() / RF_2CO2;
                } else {
                    radiative_forcing_tmp.back() += get_response_function(emissionNo, (t - tau), fuelType) * sigma;
                    normalized_radiative_forcing_tmp.back() = efficacy[emissionNo] *  radiative_forcing_tmp.back() / RF_2CO2;
                }
            }
        }  else {
            double sigma = (t < dallaraIOData->duration_operation) ? 1 : 0;
            radiative_forcing_tmp.back() += get_response_function(emissionNo, (t), fuelType) * sigma;
            normalized_radiative_forcing_tmp.back() += efficacy[emissionNo] *  radiative_forcing_tmp.back() / RF_2CO2;
        }
        myRuntimeInfo->info << "emissionNo: " << emissionNo << ", RF: " << radiative_forcing_tmp.back() << std::endl;
        /* Calculate annual values (resulting from total annual emission quantity) */
        annual_radiative_forcing_tmp.push_back(radiative_forcing_tmp.back() * get_scale_factor(emissionNo));
        annual_normalized_radiative_forcing_tmp.push_back(normalized_radiative_forcing_tmp.back() * get_scale_factor(emissionNo));
    }
    /* Add results of this emissionNo to vectors */
    radiative_forcing.push_back(radiative_forcing_tmp);
    normalized_radiative_forcing.push_back(normalized_radiative_forcing_tmp);
    annual_radiative_forcing.push_back(annual_radiative_forcing_tmp);
    annual_normalized_radiative_forcing.push_back(annual_normalized_radiative_forcing_tmp);
}

void ClimateModelDallara::calculate_temperature_change(int emissionNo) {
    std::vector<double> temperature_change_tmp; // Temperature change of a unit emission for every year
    std::vector<double> temperature_change_weighted_tmp; // Weighted temperature change of a unit emission for every year
    std::vector<double> annual_temperature_change_tmp; // Absolute temperature change of emissionNo for every year
    std::vector<double> annual_temperature_change_weighted_tmp; // Absolute weighted temperature change of emissionNo for every year
    for (uint16_t t = 0; t <= dallaraIOData->configuration.tmax; t++) {
        temperature_change_tmp.push_back(0.);
        for (int tau = 0; tau <= t; tau++) {
            double Gt = climate_sensitivity_parameter * (alphaT / tauT1 * exp(-(t - tau) / tauT1) + (1 - alphaT) / tauT2
                            * exp(-(t - tau) / tauT2)); //eq. (15) but with boundary conditions s. eq. (19)+(20)
            if (tau == 0 || tau == t) { //step size is 1 year, temperature_change s. eq. (14)
                temperature_change_tmp.back() += 0.5 * Gt * efficacy[emissionNo] * radiative_forcing.at(emissionNo).at(tau) / RF_2CO2;
            } else {
                temperature_change_tmp.back() += Gt * efficacy[emissionNo] * radiative_forcing.at(emissionNo).at(tau) / RF_2CO2;
            }
        }
        myRuntimeInfo->info << "emissionNo: " << emissionNo << ", year: " << t
                                << ", deltaT: " << temperature_change_tmp.back() << std::endl;
        temperature_change_weighted_tmp.push_back(temperature_change_tmp.back() * weighting_function(t));
        /* Calculate annual values (resulting from total annual emission quantity) */
        annual_temperature_change_tmp.push_back(temperature_change_tmp.back() * get_scale_factor(emissionNo));
        annual_temperature_change_weighted_tmp.push_back(temperature_change_weighted_tmp.back() * get_scale_factor(emissionNo));
    }
    /* Add results of this emissionNo to vectors */
    temperature_change.push_back(temperature_change_tmp);
    temperature_change_weighted.push_back(temperature_change_weighted_tmp);
    annual_temperature_change.push_back(annual_temperature_change_tmp);
    annual_temperature_change_weighted.push_back(annual_temperature_change_weighted_tmp);
}

void ClimateModelDallara::calculate_agtp_and_agwp(int emissionNo) {
    std::vector<double> agtp_tmp; // Sum of weighted temperature change of emissionNo per unit emission
    std::vector<double> agwp_tmp; // Sum of radiative forcing of emissionNo per unit emission
    std::vector<double> annual_agtp_tmp; // Annual AGTP of emissionNo
    std::vector<double> annual_agwp_tmp; // Annual AGWP of emissionNo
    for (uint16_t t = 0; t <= dallaraIOData->configuration.tmax; t++) {
        agwp_tmp.push_back(0.);
        if (t == 0) {
            agtp_tmp.push_back(temperature_change_weighted.at(emissionNo).at(t));
        } else {
            agtp_tmp.push_back(agtp_tmp.back() + temperature_change_weighted.at(emissionNo).at(t));
        }
        for (int tau = 0; tau <= t; tau++) {
            if (tau == 0 || tau == t) {
                agwp_tmp.back() += 0.5 * radiative_forcing.at(emissionNo).at(tau);
            } else {
                agwp_tmp.back() += radiative_forcing.at(emissionNo).at(tau);
            }
        }
        myRuntimeInfo->info << "emissionNo: " << emissionNo << ", year: " << t
                                << ", AGTP: " << agtp_tmp.back() << ", AGWP: " << agwp_tmp.back() << std::endl;
        /* Calculate annual values (resulting from total annual emission quantity) */
        annual_agtp_tmp.push_back(agtp_tmp.back() * get_scale_factor(emissionNo));
        annual_agwp_tmp.push_back(agwp_tmp.back() * get_scale_factor(emissionNo));
    }
    /* Add results of this emissionNo to vectors */
    agtp.push_back(agtp_tmp);
    agwp.push_back(agwp_tmp);
    annual_agtp.push_back(annual_agtp_tmp);
    annual_agwp.push_back(annual_agwp_tmp);
}

void ClimateModelDallara::calculate_average_temperature_response(int emissionNo) {
    std::vector<double> unit_atr; // Average temperature response uATR of emissionNo per unit emission [K]
    unit_atr.push_back(0.5 * temperature_change_weighted.at(emissionNo).at(0) / dallaraIOData->duration_operation);
    for (uint16_t t = 1; t <= dallaraIOData->configuration.tmax; t++) {
        if (t == dallaraIOData->configuration.tmax) {
            unit_atr.back() += 0.5 * temperature_change_weighted.at(emissionNo).at(t) / dallaraIOData->duration_operation;
        } else {
            unit_atr.back() += temperature_change_weighted.at(emissionNo).at(t) / dallaraIOData->duration_operation;
        }
    }
    myRuntimeInfo->info << "emissionNo: " << emissionNo << ", unit_atr: " << unit_atr.back() << std::endl;
    myRuntimeInfo->info << "emissionNo: " << emissionNo << ", average forcing factor: " << forcing_factor_average[emissionNo] << std::endl;
    myRuntimeInfo->info << "emissionNo: " << emissionNo << ", dallaraIOData->emissionMass: " << dallaraIOData->emissionMass[emissionNo] << std::endl;
    /* ATR for different components*/
    ATR_per_year.push_back(get_scale_factor(emissionNo) * unit_atr.back());
    myRuntimeInfo->info << "emissionNo: " << emissionNo << "ATR: " << ATR_per_year.back() << std::endl;
}

void ClimateModelDallara::calculate_total_impact() {
    for (uint16_t t = 0; t <= dallaraIOData->configuration.tmax; t++) {
        total_annual_temperature_change.push_back(0.);
        total_annual_agtp.push_back(0.);
        total_annual_temperature_change_weighted.push_back(0.);
        annual_cooling_NOx.push_back(0.);
        annual_warming_H2O_SO4_Soot.push_back(0.);
        total_annual_radiative_forcing.push_back(0.);
        total_annual_normalized_radiative_forcing.push_back(0.);
        total_annual_agwp.push_back(0.);
        for (unsigned int emissionNo = 0; emissionNo < 8; emissionNo++) { //Summing up the component values
            total_annual_radiative_forcing.back() += annual_radiative_forcing.at(emissionNo).at(t);
            total_annual_normalized_radiative_forcing.back() += annual_normalized_radiative_forcing.at(emissionNo).at(t);
            total_annual_temperature_change.back() += annual_temperature_change.at(emissionNo).at(t);
            total_annual_temperature_change_weighted.back() += annual_temperature_change_weighted.at(emissionNo).at(t);
            total_annual_agtp.back() += annual_agtp.at(emissionNo).at(t);
            total_annual_agwp.back() += annual_agwp.at(emissionNo).at(t);
            if (t == 0) {
                dallaraIOData->ATR += ATR_per_year.at(emissionNo);
            }
            if (emissionNo == 1 || emissionNo == 2) { //NOX Cooling (CH4 and O3L)
                annual_cooling_NOx.back() += annual_temperature_change.at(emissionNo).at(t);
            }
            if (emissionNo == 4 || emissionNo == 5 || emissionNo == 6) { // Warming H2O, SO4, soot
                annual_warming_H2O_SO4_Soot.back() += annual_temperature_change.at(emissionNo).at(t);
            }
        }
    }
}

double ClimateModelDallara::get_forcing_factor(int emissionNo, double flightHeight) {
    myRuntimeInfo->info << "emissionNo: " << emissionNo << " flightHeight: " << flightHeight << std::endl;
    double forcing_factor = 0.;
    uint16_t i_max = 0; // index of last row in dataset
    double forcing_factor_height[68][5]; // size of largest dataset
    if (dallaraIOData->configuration.forcing_factor_dataset == "mode_0") {
        std::copy(&forcing_factor_height_Dall[0][0], &forcing_factor_height_Dall[0][0]+13*5, &forcing_factor_height[0][0]); //copy dallara dataset [Dal11] to forcing_factor_height
        i_max = 12;
    } else if (dallaraIOData->configuration.forcing_factor_dataset == "mode_1") {
        std::copy(&forcing_factor_height_Dahl[0][0], &forcing_factor_height_Dahl[0][0]+68*5, &forcing_factor_height[0][0]); //copy dahlmann dataset [Kos22] to forcing_factor_height
        i_max = 67;
    } else {
        myRuntimeInfo->err << "Forcing factor data set doesn't exist. Check forcing_factor_dataset in config." << std::endl;
        exit(1);
    }
    for (uint16_t i = 0; i <= i_max; i++) { // for all altitudes
        if (i < i_max && forcing_factor_height[i][0] < flightHeight) {
        } else {
            if (emissionNo == 1 || emissionNo == 2) { //CH4 and O3L
                if (i == 0)
                    forcing_factor = forcing_factor_height[i][2] * dallaraIOData->configuration.forcing_factor_variation_CH4O3L;
                else if (i == i_max)
                    forcing_factor = forcing_factor_height[i][2] * dallaraIOData->configuration.forcing_factor_variation_CH4O3L;
                else
                    forcing_factor = (forcing_factor_height[i - 1][2] + (forcing_factor_height[i][2] - forcing_factor_height[i - 1][2])
                                    / (forcing_factor_height[i][0] - forcing_factor_height[i - 1][0])
                                    * (flightHeight - forcing_factor_height[i - 1][0])) * dallaraIOData->configuration.forcing_factor_variation_CH4O3L;
                myRuntimeInfo->info << "i: " << i << " forcing_factor: " << forcing_factor << std::endl;
            } else if (emissionNo == 3) { //O3S
                if (i == 0)
                    forcing_factor = forcing_factor_height[i][3] * dallaraIOData->configuration.forcing_factor_variation_O3S;
                else if (i == i_max)
                    forcing_factor = forcing_factor_height[i][3] * dallaraIOData->configuration.forcing_factor_variation_O3S;
                else
                    forcing_factor = (forcing_factor_height[i - 1][3] + (forcing_factor_height[i][3] - forcing_factor_height[i - 1][3])
                                    / (forcing_factor_height[i][0] - forcing_factor_height[i - 1][0])
                                    * (flightHeight - forcing_factor_height[i - 1][0])) *  dallaraIOData->configuration.forcing_factor_variation_O3S;
                myRuntimeInfo->info << "i: " << i << " forcing_factor: " << forcing_factor << std::endl;
            } else if (emissionNo == 4) { //H2O
                if (i == 0)
                    forcing_factor = forcing_factor_height[i][4];
                else if (i == i_max)
                    forcing_factor = forcing_factor_height[i][4];
                else
                    forcing_factor = (forcing_factor_height[i - 1][4] + (forcing_factor_height[i][4] - forcing_factor_height[i - 1][4])
                                    / (forcing_factor_height[i][0] - forcing_factor_height[i - 1][0])
                                    * (flightHeight - forcing_factor_height[i - 1][0]));
                myRuntimeInfo->info << "i: " << i << " forcing_factor: " << forcing_factor << std::endl;
            } else if (emissionNo == 7) { //AIC
                if (i == 0)
                    forcing_factor = forcing_factor_height[i][1] * dallaraIOData->configuration.forcing_factor_variation_AIC;
                else if (i == i_max)
                    forcing_factor = forcing_factor_height[i][1] * dallaraIOData->configuration.forcing_factor_variation_AIC;
                else
                    forcing_factor = (forcing_factor_height[i - 1][1] + (forcing_factor_height[i][1] - forcing_factor_height[i - 1][1])
                                    / (forcing_factor_height[i][0] - forcing_factor_height[i - 1][0])
                                    * (flightHeight - forcing_factor_height[i - 1][0])) * dallaraIOData->configuration.forcing_factor_variation_AIC;
                myRuntimeInfo->info << "i: " << i << " forcing_factor: " << forcing_factor << std::endl;
            } else {
                forcing_factor = 1.;
                myRuntimeInfo->info << "i: " << i << " forcing_factor: " << forcing_factor << std::endl;
            }
            break;
        }
    }
    return forcing_factor;
}

double ClimateModelDallara::get_response_function(int emissionNo, double time, const std::string &fuelType) {
    if (emissionNo == 0) { //CO2, eq. (6)
        return A_CO2 * (1 + (alphaC1 * (expm1(-time / tauC1)) + alphaC2 * (expm1(-time / tauC2)) + alphaC3 * (expm1(-time / tauC3))));
    } else if (emissionNo == 1) { //CH4, eq. (7)
        return A_CH4 * exp(-time / tauN);
    } else if (emissionNo == 2) { //O3L, eq. (7)
        return A_O3L * exp(-time / tauN);
    } else if (emissionNo == 3) { //O3S, eq. (10)
        return RFrefEref_O3S;
    } else if (emissionNo == 4) { //H2O, eq. (10)
        return RFrefEref_H2O;
    } else if (emissionNo == 5) { //soot, eq. (10)
        return RFrefEref_soot;
    } else if (emissionNo == 6) { //SO4, eq. (10)
        return RFrefEref_SO4;
    } else if (emissionNo == 7) { //AIC, eq. (10)
        double AICfactor(1.);
        if (fuelType == "liquid_hydrogen")
            AICfactor = dallaraIOData->configuration.fuelAICfactor;
        return AICfactor * RFrefLref_AIC;
    } else {
        myRuntimeInfo->err << "Response function of emission number does not exist!" << std::endl;
        exit(1);
    }
}

double ClimateModelDallara::weighting_function(double time) {
    double weighting(0.);
    if (time < dallaraIOData->duration_operation) {
            weighting = 1;
    } else {
        weighting = 1.0 / (pow(1 + dallaraIOData->configuration.devRate, (time - dallaraIOData->duration_operation)));
    }
    return weighting;
}

double ClimateModelDallara::get_scale_factor(int emissionNo) {
    return forcing_factor_average[emissionNo] * emissions_per_year[emissionNo];
}
