# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import sys
import xml.etree.ElementTree

from datapostprocessingmodule import convert_element_tree_to_dictionary


def prepare_external_data(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config, runtime_output):
    """  Function prepare external data for project specific calculations.

        :param dict paths_and_names: Dictionary containing system paths and ElementTrees
        :param dict routing_dict: Dictionary containing routing parameter
        :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
        :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
        :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
        :return dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    """
    ''' initialize local parameter '''
    indexes = []
    percentage = []
    seat_details = None
    lavatory_details = {}
    relevant_classes = {}
    relevant_cargo_accommodation = {}
    class_distribution = dict_ac_exchange['class_distribution'].split('/')
    # Todo ersetzen durch constants from lib
    meter_to_nautical_mile = 1852

    ''' Import passenger and cargo accommodation file '''
    # Initialize element tree with content of file and return root element (if given).
    try:
        # Attempt to create an ElementTree and get the root element from the XML file.
        accommodation_tree = \
            xml.etree.ElementTree.ElementTree(file='fuselage_design_lib/accommodation_definitions.xml')
        dict_ac_exchange['dict_accommodation'] = convert_element_tree_to_dictionary(accommodation_tree.getroot())
    # Exception handling for operating system (OS) error.
    except OSError:
        # Handle an error if the XML file cannot be opened. Print an error message and log it to a log file.
        runtime_output.critical('Error: The passenger and cargo accommodation definition xml-file could not be opened '
                                'or found. \n                                     Program aborted!')
        sys.exit('Accommodation definition file could not be opened.')

    ''' Import fuselage design cs requirements file '''
    # Initialize element tree with content of file and return root element (if given).
    try:
        # Attempt to create an ElementTree and get the root element from the XML file.
        cs_requirement_tree = \
            xml.etree.ElementTree.ElementTree(file='fuselage_design_lib/fuselage_design_cs_requirements.xml')
        dict_ac_exchange['dict_cs_requirements'] = convert_element_tree_to_dictionary(cs_requirement_tree.getroot())
    # Exception handling for operating system (OS) error.
    except OSError:
        # Handle an error if the XML file cannot be opened. Print an error message and log it to a log file.
        runtime_output.critical('Error: The fuselage design cs requirements xml-file could not be opened or found. \n'
                                '                                     Program aborted!')
        sys.exit('Fuselage design cs requirements file could not be opened.')

    runtime_output.print('Import of fuselage design cs requirements xml-file and accommodation definition xml-file '
                         'successfully.')

    ''' Extract project specific passenger and cargo accommodation '''
    # Find relevant passenger classes.
    i = 0
    for passenger_class in class_distribution:
        if not passenger_class == '0':
            indexes.append(i)
            percentage.append(round(float(passenger_class), 3))
        i += 1
    relevant_classes['class_indexes'] = indexes
    relevant_classes['class_percentages'] = percentage

    if sum(percentage) > 1 or sum(percentage) <= 0:
        runtime_output.critical('Error: The given class distribution is greater than 1 (100%). '
                                'Please change class distribution in the design specification. \n'
                                '                                     Program aborted!')
        sys.exit('Selected class distribution not possible.')

    # Get relevant class names and associated accommodation items.
    relevant_classes['passenger_classes'] = {}
    relevant_classes['passenger_classes']['class_names'] = []
    class_names = list(dict_mod_config['passenger_class_name'].values())
    for index in indexes:
        # initialize class dictionaries
        relevant_classes['passenger_classes'][class_names[index]] = {}
        relevant_classes['passenger_classes'][class_names[index]]['seat_name'] = {}
        relevant_classes['passenger_classes'][class_names[index]]['passenger_seats'] = {}
        relevant_classes['passenger_classes'][class_names[index]]['seat_pitch'] = {}
        relevant_classes['passenger_classes'][class_names[index]]['class_lavatory'] = {}
        relevant_classes['passenger_classes'][class_names[index]]['class_wardrobe'] = {}
        # initialize class parameter lists
        relevant_classes['passenger_classes'][class_names[index]]['seat_abreast'] = []
        relevant_classes['passenger_classes'][class_names[index]]['seat_name']['seat_name_left_side'] = []
        relevant_classes['passenger_classes'][class_names[index]]['seat_name']['seat_name_right_side'] = []
        relevant_classes['passenger_classes'][class_names[index]]['seat_name']['seat_name_in_the_center'] = []
        relevant_classes['passenger_classes'][class_names[index]]['passenger_seats']['seat_left_side'] = []
        relevant_classes['passenger_classes'][class_names[index]]['passenger_seats']['seat_right_side'] = []
        relevant_classes['passenger_classes'][class_names[index]]['passenger_seats']['seat_in_the_center'] = []
        relevant_classes['passenger_classes'][class_names[index]]['seat_pitch']['seat_pitch_left_side'] = []
        relevant_classes['passenger_classes'][class_names[index]]['seat_pitch']['seat_pitch_right_side'] = []
        relevant_classes['passenger_classes'][class_names[index]]['seat_pitch']['seat_pitch_in_the_center'] = []
        relevant_classes['passenger_classes'][class_names[index]]['seat_space'] = []
        relevant_classes['passenger_classes'][class_names[index]]['class_lavatory']['lavatory_name'] = []
        relevant_classes['passenger_classes'][class_names[index]]['class_lavatory']['pax_per_lavatory'] = []
        relevant_classes['passenger_classes'][class_names[index]]['class_lavatory']['use_specific_lavatory_number'] = []
        relevant_classes['passenger_classes'][class_names[index]]['class_lavatory']['specific_lavatory_number'] = []
        relevant_classes['passenger_classes'][class_names[index]]['class_lavatory']['lavatory'] = []
        relevant_classes['passenger_classes'][class_names[index]]['class_wardrobe']['use_wardrobe'] = []
        relevant_classes['passenger_classes'][class_names[index]]['class_wardrobe']['wardrobe_space'] = []
        relevant_classes['passenger_classes'][class_names[index]]['class_wardrobe']['dimension'] = []

        relevant_classes['passenger_classes']['class_names'].append(class_names[index])
        for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
            # read seat abreast
            relevant_classes['passenger_classes'][class_names[index]]['seat_abreast'].append(
                dict_mod_config['passenger_class_abreast']['passenger_class_abreast_ID' + str(tube) + '_ID'
                                                           + str(index)])
            if relevant_classes['passenger_classes'][class_names[index]]['seat_abreast'] is None:
                runtime_output.critical('Error: At least one selected class has no entered number of seat abreast.\n'
                                        '                                     Program aborted!')
                sys.exit('Missing number of seat abreast for payload tube with ID=' + str(tube) + 'for '
                         + class_names[index] + ' class.')

            # read seat name
            relevant_classes['passenger_classes'][class_names[index]]['seat_name']['seat_name_left_side'].append(
                dict_mod_config['passenger_seat_on_left_side_name']['passenger_seat_on_left_side_name_ID' + str(tube)
                                                                    + '_ID' + str(index)])
            relevant_classes['passenger_classes'][class_names[index]]['seat_name']['seat_name_right_side'].append(
                dict_mod_config['passenger_seat_on_right_side_name']['passenger_seat_on_right_side_name_ID'
                                                                     + str(tube) + '_ID' + str(index)])
            if relevant_classes['passenger_classes'][class_names[index]]['seat_name']['seat_name_left_side'] is None or\
                    relevant_classes['passenger_classes'][class_names[index]]['seat_name']['seat_name_right_side'] is None:  # noPep8 e501
                runtime_output.critical('Error: At least one selected class has no entered seat name.\n'
                                        '                                     Program aborted!')
                sys.exit('Missing name of passenger seat for ' + class_names[index] + ' class.')

            # read seat data from accommodation dictionary
            for seat in dict_ac_exchange['dict_accommodation']['accommodation_definitions']['passenger_seats']['seat']:
                if seat['name']['value'] == \
                        relevant_classes['passenger_classes'][class_names[index]]['seat_name'][
                            'seat_name_left_side'][tube]:
                    seat_details = seat
                    break

            if seat_details is None:
                runtime_output.critical('Error: The selected class seat for the ' + class_names[index]
                                        + ' class does not exist in accommodation definition file.\n'
                                          '                                     Program aborted!')
                sys.exit('Selected seat does not exist in accommodation file.')
            else:
                relevant_classes['passenger_classes'][class_names[index]]['passenger_seats']['seat_left_side'].append(
                    seat_details)

            # seats on the right side of the aircraft
            for seat in dict_ac_exchange['dict_accommodation']['accommodation_definitions']['passenger_seats']['seat']:
                if seat['name']['value'] == \
                        relevant_classes['passenger_classes'][class_names[index]]['seat_name'][
                            'seat_name_right_side'][tube]:  # noPep8 e501
                    seat_details = seat
                    break

            if seat_details is None:
                runtime_output.critical('Error: The selected class seat for the ' + class_names[index]
                                        + ' class does not exist in accommodation definition file.\n'
                                          '                                     Program aborted!')
                sys.exit('Selected seat does not exist in accommodation file.')
            else:
                relevant_classes['passenger_classes'][class_names[index]]['passenger_seats']['seat_right_side'].append(
                    seat_details)

            # read seat pitch
            relevant_classes['passenger_classes'][class_names[index]]['seat_pitch']['seat_pitch_left_side'].append(
                dict_mod_config['passenger_seat_on_left_side_pitch']['passenger_seat_on_left_side_pitch_ID'
                                                                     + str(tube) + '_ID' + str(index)])
            relevant_classes['passenger_classes'][class_names[index]]['seat_pitch']['seat_pitch_right_side'].append(
                dict_mod_config['passenger_seat_on_right_side_pitch']['passenger_seat_on_right_side_pitch_ID'
                                                                      + str(tube) + '_ID' + str(index)])

            # read outer seat space
            relevant_classes['passenger_classes'][class_names[index]]['seat_space'].append(
                dict_mod_config['passenger_seat_space']['passenger_seat_space_ID' + str(tube) + '_ID' + str(index)])

            # read class associated lavatories
            relevant_classes['passenger_classes'][class_names[index]]['class_lavatory']['lavatory_name'].append(
                dict_mod_config['passenger_lavatory_name']['passenger_lavatory_name_ID'
                                                           + str(tube) + '_ID' + str(index)])
            relevant_classes['passenger_classes'][class_names[index]]['class_lavatory']['pax_per_lavatory'].append(
                dict_mod_config['passenger_per_lavatory']['passenger_per_lavatory_ID'
                                                          + str(tube) + '_ID' + str(index)])
            relevant_classes['passenger_classes'][class_names[index]]['class_lavatory'][
                'use_specific_lavatory_number'].append(dict_mod_config['use_specific_number_of_class_lavatories'][
                                   'use_specific_number_of_class_lavatories_ID' + str(tube) + '_ID' + str(index)])
            relevant_classes['passenger_classes'][class_names[index]]['class_lavatory'][
                'specific_lavatory_number'].append(dict_mod_config['specific_number_of_class_lavatories'][
                                                                   'specific_number_of_class_lavatories_ID'
                                                                   + str(tube) + '_ID' + str(index)])

            # read class associated lavatory from accommodation definition file
            for lavatory in dict_ac_exchange['dict_accommodation']['accommodation_definitions']['lavatories']['lavatory']:  # noPep8 e501
                if lavatory['name']['value'] == \
                        relevant_classes['passenger_classes'][class_names[index]]['class_lavatory']['lavatory_name'][tube]:  # noPep8 e501
                    lavatory_details = lavatory
                    break

            if lavatory_details is None:
                runtime_output.critical('Error: The selected class lavatory for the ' + class_names[index]
                                        + ' class does not exist in accommodation definition file.\n'
                                          '                                     Program aborted!')
            else:
                relevant_classes['passenger_classes'][class_names[index]]['class_lavatory']['lavatory'].append(
                    lavatory_details)

            # read class associated wardrobe parameter
            relevant_classes['passenger_classes'][class_names[index]]['class_wardrobe']['use_wardrobe'].append(
                dict_mod_config['use_wardrobe_for_passenger_class']['use_wardrobe_for_passenger_class_ID'
                                                                    + str(tube) + '_ID' + str(index)])
            if relevant_classes['passenger_classes'][class_names[index]]['class_wardrobe']['use_wardrobe'][tube]:
                relevant_classes['passenger_classes'][class_names[index]]['class_wardrobe']['wardrobe_space'].append(
                    dict_mod_config['wardrobe_space_per_passenger']['wardrobe_space_per_passenger_ID'
                                                                    + str(tube) + '_ID' + str(index)])
                relevant_classes['passenger_classes'][class_names[index]]['class_wardrobe']['dimension'].append(
                    dict_ac_exchange['dict_accommodation']['accommodation_definitions']['wardrobe'])
            else:
                relevant_classes['passenger_classes'][class_names[index]]['class_wardrobe']['wardrobe_space'].append(None)  # noPep8 e501
                relevant_classes['passenger_classes'][class_names[index]]['class_wardrobe']['dimension'].append(None)

            # Check if the current project is a wide body aircraft
            #  -> if true: -> set wide body relevant passenger accommodation.
            if dict_ac_exchange['fuselage_type'] == 'wide_body':
                # read seat name
                relevant_classes['passenger_classes'][class_names[index]]['seat_name'][
                    'seat_name_in_the_center'].append(dict_mod_config['passenger_seat_in_the_center_name'][
                        'passenger_seat_in_the_center_name_ID' + str(tube) + '_ID' + str(index)])

                if relevant_classes['passenger_classes'][class_names[index]]['seat_name']['seat_name_in_the_center'][tube] is None:  # noPep8 e501
                    runtime_output.critical('Error: At least one selected class has no entered seat name.\n'
                                            '                                     Program aborted!')
                    sys.exit('Missing name of passenger seat for ' + class_names[index] + ' class.')

                # seats on the left side of the aircraft
                for seat in dict_ac_exchange['dict_accommodation']['accommodation_definitions']['passenger_seats']['seat']:  # noPep8 e501
                    if seat['name']['value'] == \
                            relevant_classes['passenger_classes'][class_names[index]]['seat_name'][
                                'seat_name_in_the_center'][tube]:
                        seat_details = seat
                        break

                if seat_details is None:
                    runtime_output.critical('Error: The selected class seat for the ' + class_names[index]
                                            + ' class does not exist in accommodation definition file.\n'
                                              '                                     Program aborted!')
                    sys.exit('Selected seat does not exist in accommodation file.')
                else:
                    relevant_classes['passenger_classes'][class_names[index]]['passenger_seats'][
                        'seat_in_the_center'].append(seat_details)

                # read seat pitch
                relevant_classes['passenger_classes'][class_names[index]]['seat_pitch'][
                    'seat_pitch_in_the_center'].append(
                    dict_mod_config['passenger_seat_in_the_center_pitch']['passenger_seat_in_the_center_pitch_ID'
                                                                          + str(tube) + '_ID' + str(index)])

    ''' set design range associated scenario '''
    range_in_nautical_mile = dict_ac_exchange['design_mission_range'] / meter_to_nautical_mile
    relevant_classes['range_scenario'] = (
        'regional' if 0 <= range_in_nautical_mile < 500
        else 'short_range' if 500 <= range_in_nautical_mile < 3000
        else 'medium_range' if 3000 <= range_in_nautical_mile < 5500
        else 'long_range')

    ''' set relevant cargo accommodation '''
    for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
        relevant_cargo_accommodation['uld_type_ID' + str(tube)] = None
        if dict_mod_config['uld_type']['uld_type_ID' + str(tube)] == 'container':
            for container in dict_ac_exchange['dict_accommodation']['accommodation_definitions']['unit_load_devices'][
                                              'ul_container']['container']:
                if container['name']['value'] == dict_mod_config['container_name']['container_name_ID' + str(tube)]:
                    relevant_cargo_accommodation['uld_type_ID' + str(tube)] = \
                        {container['name']['value']: {}}
                    relevant_cargo_accommodation['uld_type_ID' + str(tube)][container['name']['value']] = \
                        container

            if relevant_cargo_accommodation.get('uld_type_ID' + str(tube)) is None:
                runtime_output.critical('Error: The selected uld container type: "'
                                        + dict_mod_config['container_name']['container_name_ID' + str(tube)]
                                        + '" for payload tube with "ID=' + str(tube)
                                        + '" is not valid.\n                                     Program aborted!')
                sys.exit('Uld container type not valid.')

        elif dict_mod_config['uld_type']['uld_type_ID' + str(tube)] == 'pallet':
            for pallet in dict_ac_exchange['dict_accommodation']['accommodation_definitions']['unit_load_devices'][
                                              'cargo_pallet']['pallet']:
                if pallet['name']['value'] == dict_mod_config['pallet_name']['pallet_name_ID' + str(tube)]:
                    relevant_cargo_accommodation['uld_type_ID' + str(tube)] = {pallet['name']['value']: {}}
                    relevant_cargo_accommodation['uld_type_ID' + str(tube)][pallet['name']['value']] = \
                        pallet

            if relevant_cargo_accommodation.get('uld_type_ID' + str(tube)) is None:
                runtime_output.critical('Error: The selected cargo pallet type: "'
                                        + dict_mod_config['pallet_name']['pallet_name_ID' + str(tube)]
                                        + '" for payload tube with "ID=' + str(tube)
                                        + '" is not valid.\n                                     Program aborted!')
                sys.exit('Cargo pallet not valid.')

        elif dict_mod_config['uld_type']['uld_type_ID' + str(tube)] == 'mixed':
            for container in dict_ac_exchange['dict_accommodation']['accommodation_definitions']['unit_load_devices'][
                                              'ul_container']['container']:
                if container['name']['value'] == dict_mod_config['container_name']['container_name_ID' + str(tube)]:
                    relevant_cargo_accommodation['uld_type_ID' + str(tube)] = {
                        container['name']['value']: {}}
                    relevant_cargo_accommodation['uld_type_ID' + str(tube)][container['name']['value']] =\
                        container

            if relevant_cargo_accommodation.get('uld_type_ID' + str(tube)) is None:
                runtime_output.critical('Error: The selected uld container type: "'
                                        + dict_mod_config['container_name']['container_name_ID' + str(tube)]
                                        + '" for payload tube with "ID=' + str(tube)
                                        + '" is not valid.\n                                     Program aborted!')
                sys.exit('Uld container type not valid.')

            else:
                for pallet in dict_ac_exchange['dict_accommodation']['accommodation_definitions']['unit_load_devices'][
                                               'cargo_pallet']['pallet']:
                    if pallet['name']['value'] == dict_mod_config['pallet_name']['pallet_name_ID' + str(tube)]:
                        relevant_cargo_accommodation['uld_type_ID' + str(tube)][pallet['name']['value']] = \
                            {}
                        relevant_cargo_accommodation['uld_type_ID' + str(tube)][pallet['name']['value']] = \
                            pallet

                if len(relevant_cargo_accommodation.get('uld_type_ID' + str(tube))) < 2:
                    runtime_output.critical('Error: The selected cargo pallet type: "'
                                            + dict_mod_config['pallet_name']['pallet_name_ID' + str(tube)]
                                            + '" for payload tube with "ID=' + str(tube)
                                            + '" is not valid.\n                                     Program aborted!')
                    sys.exit('Cargo pallet not valid.')

        elif dict_mod_config['uld_type']['uld_type_ID' + str(tube)] is None:
            runtime_output.warning('Attention: For the payload tube with "ID=' + str(tube)
                                   + ' the selected uld_type is None. '
                                     'Therefore, no cargo bay is designed for this payload tube.')
        else:
            runtime_output.critical('Error: The selected uld type: "'
                                    + dict_mod_config['uld_type']['uld_type_ID' + str(tube)]
                                    + '" for payload tube with "ID=' + str(tube)
                                    + '" is not valid.\n                                     Program aborted!')
            sys.exit('Selected uld type not valid.')

    ''' store relevant data to aircraft exchange dictionary '''
    dict_ac_exchange['relevant_classes'] = relevant_classes
    dict_ac_exchange['relevant_cargo_accommodation'] = relevant_cargo_accommodation

    ''' debug print '''
    runtime_output.debug('Debug: The prepare_external_data function was successfully executed!')

    return dict_ac_exchange
