# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def estimate_masses(dict_ac_exchange, dict_mod_config, landing_gear_description, runtime_output):
    """ Hier beschreibung hin: Estimation of undercarriage masses in accordance to Torenbeek.
                [ref. E. Torenbeek; 1982]

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    """
    ''' initialize local parameter '''
    landing_gear_masses = {}
    gravitational_acceleration = float(9.80665)
    fixed_landing_gear_coefficients = {'main': {}, 'nose': {}, 'tail': {}}
    retractable_landing_gear_coefficients = {'main': {}, 'nose': {}, 'tail': {}}
    if dict_ac_exchange['wing_mounting'] == 'low' or dict_ac_exchange['wing_mounting'] == 'mid':
        wing_position_coefficient = 1.0
    else:
        wing_position_coefficient = 1.8
    if dict_ac_exchange['maximum_take_off_mass'] is None:
        maximum_take_off_mass = (landing_gear_description['landing_gear_loads']['maximum_ramp_weight']
                                 / gravitational_acceleration)
    else:
        maximum_take_off_mass = dict_ac_exchange['maximum_take_off_mass']

    ''' coefficients for the landing gear mass in accordance to Torenbeek (table 8-6) '''
    # fixed landing gear
    fixed_landing_gear_coefficients['main']['A'] = 9.1
    fixed_landing_gear_coefficients['main']['B'] = 0.082
    fixed_landing_gear_coefficients['main']['C'] = 0.019
    fixed_landing_gear_coefficients['main']['D'] = 0
    fixed_landing_gear_coefficients['nose']['A'] = 11.3
    fixed_landing_gear_coefficients['nose']['B'] = 0
    fixed_landing_gear_coefficients['nose']['C'] = 0.0024
    fixed_landing_gear_coefficients['nose']['D'] = 0

    # retractable landing gear
    retractable_landing_gear_coefficients['main']['A'] = 18.1
    retractable_landing_gear_coefficients['main']['B'] = 0.131
    retractable_landing_gear_coefficients['main']['C'] = 0.019
    retractable_landing_gear_coefficients['main']['D'] = 0.0000223
    retractable_landing_gear_coefficients['nose']['A'] = 9.1
    retractable_landing_gear_coefficients['nose']['B'] = 0.082
    retractable_landing_gear_coefficients['nose']['C'] = 0
    retractable_landing_gear_coefficients['nose']['D'] = 0.00000297

    ''' estimate nose gear mass '''
    if dict_ac_exchange['undercarriage_retractability'] == 'retractable':
        nose_landing_gear_mass = (wing_position_coefficient * (retractable_landing_gear_coefficients['nose']['A']
                                                               + retractable_landing_gear_coefficients['nose']['B']
                                                               * pow(maximum_take_off_mass, 0.75)
                                                               + retractable_landing_gear_coefficients['nose']['C']
                                                               * maximum_take_off_mass
                                                               + retractable_landing_gear_coefficients['nose']['D']
                                                               * pow(maximum_take_off_mass, 1.5))
                                  * dict_mod_config['landing_gear_mass_correction_factor'])
    else:
        nose_landing_gear_mass = (wing_position_coefficient * (fixed_landing_gear_coefficients['nose']['A']
                                                               + fixed_landing_gear_coefficients['nose']['B']
                                                               * pow(maximum_take_off_mass, 0.75)
                                                               + fixed_landing_gear_coefficients['nose']['C']
                                                               * maximum_take_off_mass
                                                               + fixed_landing_gear_coefficients['nose']['D']
                                                               * pow(maximum_take_off_mass, 1.5))
                                  * dict_mod_config['landing_gear_mass_correction_factor'])

    nose_gear_strut_mass = (nose_landing_gear_mass
                            / landing_gear_description['nose_gear_description']['nose_gear_arrangement']
                            ['number_of_nose_gear_struts'])
    nose_landing_gear_masses = {'nose_landing_gear_mass': nose_landing_gear_mass,
                                'nose_landing_gear_mass_per_strut': nose_gear_strut_mass}

    ''' estimate main gear mass '''
    if dict_ac_exchange['undercarriage_retractability'] == 'retractable':
        main_landing_gear_mass = (wing_position_coefficient * (retractable_landing_gear_coefficients['main']['A']
                                                               + retractable_landing_gear_coefficients['main']['B']
                                                               * pow(maximum_take_off_mass, 0.75)
                                                               + retractable_landing_gear_coefficients['main']['C']
                                                               * maximum_take_off_mass
                                                               + retractable_landing_gear_coefficients['main']['D']
                                                               * pow(maximum_take_off_mass, 1.5))
                                  * dict_mod_config['landing_gear_mass_correction_factor'])
    else:
        main_landing_gear_mass = (wing_position_coefficient * (fixed_landing_gear_coefficients['main']['A']
                                                               + fixed_landing_gear_coefficients['main']['B']
                                                               * pow(maximum_take_off_mass, 0.75)
                                                               + fixed_landing_gear_coefficients['main']['C']
                                                               * maximum_take_off_mass
                                                               + fixed_landing_gear_coefficients['main']['D']
                                                               * pow(maximum_take_off_mass, 1.5))
                                  * dict_mod_config['landing_gear_mass_correction_factor'])

    main_gear_strut_mass = (main_landing_gear_mass
                            / landing_gear_description['main_gear_description']['main_gear_arrangement']
                            ['number_of_main_gear_struts'])
    main_landing_gear_masses = {'main_landing_gear_mass': main_landing_gear_mass,
                                'main_landing_gear_mass_per_strut': main_gear_strut_mass}

    ''' set outputs to landing gear output dictionary '''
    landing_gear_masses['landing_gear_total_mass'] = round(main_landing_gear_mass + nose_landing_gear_mass, 4)
    landing_gear_masses['nose_landing_gear_masses'] = nose_landing_gear_masses
    landing_gear_masses['nose_landing_gear_masses'] = {
        key: round(value, 4) for key, value in landing_gear_masses['nose_landing_gear_masses'].items()}
    landing_gear_masses['main_landing_gear_masses'] = main_landing_gear_masses
    landing_gear_masses['main_landing_gear_masses'] = {
        key: round(value, 4) for key, value in landing_gear_masses['main_landing_gear_masses'].items()}
    landing_gear_description['landing_gear_masses'] = landing_gear_masses

    runtime_output.print('Total mass of landing gear is: '
                         + str(landing_gear_masses['landing_gear_total_mass']) + 'kg!')
    
    runtime_output.debug('estimate_masses successfully executed!')

    return landing_gear_description
