/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef MISSION_ANALYSIS_SRC_LIBS_MULTI_ENGINE_PROPULSION_MULTI_ENGINE_PROPULSION_H_
#define MISSION_ANALYSIS_SRC_LIBS_MULTI_ENGINE_PROPULSION_MULTI_ENGINE_PROPULSION_H_

#include <map>
#include <memory>
#include <set>
#include <string>
#include <tuple>
#include <utility>
#include <vector>

#include <aixml/endnode.h>
#include <energyCarriers/energyCarriers.h>
#include <engine/engine.h>
#include <moduleBasics/module.h>

/** \class MultiEnginePropulsion
 *  \brief Multi engine propulsion handels all engines including energy carriers and tanks
 */
class MultiEnginePropulsion {
 public:
    /* Constructor and destructor */
    /** \brief Constructor for class MultiEnginePropulsion
     */
    explicit MultiEnginePropulsion(const std::shared_ptr<RuntimeIO>& rtIO);
    /** \brief Destructor for class MultiEnginePropulsion
     */
    virtual ~MultiEnginePropulsion();

    /** \brief Function to get the number of unique engines using the specified energy carrier ID
     *  \details Same engines are counted as one. To get ALL engines use 'get_number_of_engines'
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \return number of unique engines
     */
    size_t unique_engines_per_id(const int& energy_carrier_id);
    /** \brief Function to get the number of engines using the specified energy carrier ID
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \return number of all engines
     */
    uint16_t get_number_of_engines(const int& energy_carrier_id);
    /** \brief Function to get the operating points of all engines using the specified energy carrier ID
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \return vector containing operating points (N [-], Mach [-], Altitude [m])
     */
    std::vector<OperatingPoint> get_operating_points(const int& energy_carrier_id);
    /** \brief Function to set the operating points of all engines using the specified energy carrier ID
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \param operating_points: vector containing operating points (N [-], Mach [-], Altitude [m])
     */
    void set_operating_points(const int& energy_carrier_id, std::vector<OperatingPoint> operating_points);
    /** \brief Function to get the total fuel flow of the aircraft using the specified energy carrier ID
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \return double: fuelflow [kg/s]
     */
    double get_aircraft_fuelflow(const int& energy_carrier_id);
    /** \brief Function to get the total energy flow of the aircraft using the specified energy carrier ID
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \return double: energyflow [J/s]
     */
    double get_aircraft_energyflow(const int& energy_carrier_id);
    /** \brief Function to get the total energy flow of the aircraft using the specified energy carrier ID at a given fuelflow
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \param aircraft_fuelflow double&: Pointer to a given aircraft fuelflow [kg/s]
     *  \return double: energyflow [J/s]
     */
    double get_aircraft_energyflow(const int& energy_carrier_id, const double& aircraft_fuelflow);
    /** \brief Function to get the total thrust of the aircraft using the specified energy carrier ID
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \return double: thrust [N]
     */
    double get_thrust_aircraft(const int& energy_carrier_id);
    /** \brief Function to set the engines (using a given energy carrier) to their operating N1
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \param altitude double&: Pointer to a given altitude [m]
     *  \param mach_number double&: Pointer to a given Mach number [m]
     *  \param atm: Pointer to a given atmosphere object
     *  \param derate double&: Pointer to a given derate of the engine
     *  \param thrust_rating string&: Pointer to a given thrust rating ('climb', 'cruise', 'idle', 'maximum_continuous', 'takeoff')
     *  \param bleed_air_offtake double: bleed offtake during the engine operation
     *  \param shaft_power_offtake double: shaft power offtake during the engine operation
     */
    void calculate_N1_with_penalties(const int& energy_carrier_id,
                                     const double& altitude,
                                     const double& mach_number,
                                     const atmosphere& atm,
                                     const double& derate,
                                     const std::string& thrust_rating,
                                     double bleed_air_offtake,
                                     double shaft_power_offtake);
    /** \brief Function to set the engines (using a given energy carrier) to their operating N1 with a overall target thrust
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \param altitude double&: Pointer to a given altitude [m]
     *  \param mach_number double&: Pointer to a given Mach number [m]
     *  \param atm: Pointer to a given atmosphere object
     *  \param derate double&: Pointer to a given derate of the engine
     *  \param thrust_rating string&: Pointer to a given thrust rating ('climb', 'cruise', 'idle', 'maximum_continuous', 'takeoff')
     *  \param bleed_air_offtake double: bleed offtake during the engine operation (all engines together)
     *  \param shaft_power_offtake double: shaft power offtake during the engine operation (all engines together)
     *  \param thrust_limit double: thrust limit that shall be reached by the engines (all engines together)
     */
    void calculate_N1_with_thrustlimit(const int& energy_carrier_id,
                                       const double& altitude,
                                       const double& mach_number,
                                       const atmosphere& atm,
                                       const double& derate,
                                       const std::string& thrust_rating,
                                       double bleed_air_offtake,
                                       double shaft_power_offtake,
                                       double thrust_limit);

    /** \class Propulsor
     *  \details A propulsor contains the engine including fruther information e.g. about the quantity of equal engines
     */
    class Propulsor {
     public:
        std::string name;                   /**< Name within the aircraft xml of the engine */
        Engine engine;                      /**< Engine lib object */
        std::vector<uint16_t> engine_id;    /**< IDs within the aircraft xml of equal engines */
        double thrust_share;                /**< Thrust share [-] */

        /** \brief Constructor of Propulsor.
         *  \param a_name string: Name within the aircraft xml of the engine
         *  \param an_engine Engine: Engine lib object
         *  \param an_id uint16_t&: IDs within the aircraft xml of equal engines
         *  \param a_thrust_share double&: Thrust share [-]
         */
        Propulsor(const std::string& a_name, Engine an_engine, const uint16_t& an_id, const double& a_thrust_share);

        /** \brief Destructor of Propulsor
         */
        virtual ~Propulsor();
    };

    /** \class Tank
     *  \details A Tank contains and limits the energy amount which can be consumed
     */
    class Tank {
     public:
        uint16_t tank_id;           /**< ID within the aircraft xml */
        double capacity;            /**< Energy capacity of the tank [J] */
        double relative_capacity;   /**< Energy capacity of the tank in relation to the total energy capacity [-] */
        double consumed_energy;     /**< Consumed energy of the tank [J] */

        /** \brief Constructor of Tank.
         *  \param a_name uint16_t: ID within the aircraft xml
         *  \param a_capacity double: Energy capacity of the tank [J]
         */
        Tank(uint16_t a_tank_id, double a_capacity);

        /** \brief Destructor of Tank
         */
        virtual ~Tank();
    };

    /** \class EnergyConsumption
     *  \details An energy consumption class contains all tanks needed for the specific energy carrier that shall be consumed
     */
    class EnergyConsumption {
     public:
        const EnergyCarrier& energy_carrier;
        std::vector<Tank> tanks;    /**< Vector of tank objects containing this energy carrier */

        /** \brief Constructor of EnergyConsumption.
         *  \param an_energy_carrier EnergyCarrier: The energy carrier that shall be consumed
         */
        explicit EnergyConsumption(const EnergyCarrier& an_energy_carrier);

        /** \brief Destructor of EnergyCarrier
         */
        virtual ~EnergyConsumption();
    };

    /** \brief Function to get the energy consumption of all energy carriers
     *  \return map of energy amount [J] consumed for each energy carrier ID
     */
    std::map<int, double> get_consumed_energy_map();
    /** \brief Function to get the energy consumption of a specified energy carrier
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \return double value of the total energy amount [J] consumed for a given energy carrier ID
     */
    double get_consumed_energy(const int& energy_carrier_id);
    /** \brief Function to get the energy flow of all energy carriers
     *  \return map of the total energy flow [J/s] for each energy carrier ID
     */
    std::map<int, double> get_energyflow_map();
    /** \brief Function to consume the fuel from the tanks
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \param delta_t double&: Pointer to a given time in which the energy shall be consumed [t]
     *  \param fuelflow double&: Pointer to a given fuelflow if that is already known
     */
    void drain_tanks(const int& energy_carrier_id, const double& delta_t, const double& fuelflow = NAN);
    /** \brief Function to get the total fuel consumption of all energy carriers
     *  \return double value of the total fuel mass amount [kg]
     */
    double get_consumed_fuel_mass();
    /** \brief Function to get the total energy capacity of the given energy carrier ID
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \return double total energy capacity [J]
     */
    double get_energy_carrier_capacity(const int& energy_carrier_id);
    /** \brief Function to get the propulsors using a given energy carrier
     *  \param energy_carrier_id int&: Pointer to a given energy carrier ID
     *  \return vector containing Propulsor objects using the given energy carrier ID
     */
    std::vector<Propulsor> get_propulsors(const int& energy_carrier_id);

 private:
    /* variables */
    std::map<int, std::vector<Propulsor>> propulsors;   /* Map to gather all propulsors used for the mission. Int key = energy carrier ID */
    std::map<int, EnergyConsumption> energy_map;   /* Map to gather all energy carriers which are consumed during the mission. Int key = energy carrier ID */

    /* functions */
    /** \brief Function to check if vectors have matching energy carrier ID entries
     *  \param v1 vector<int>&: Pointer containing energy carrier IDs
     *  \param v2 vector<int>&: vector pointer containing energy carrier IDs
     *  \return boolean check result
     */
    bool vector_check(const std::vector<int>& v1, const std::vector<int>& v2);
};

#endif // MISSION_ANALYSIS_SRC_LIBS_MULTI_ENGINE_PROPULSION_MULTI_ENGINE_PROPULSION_H_
