/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef MISSION_ANALYSIS_SRC_STANDARD_MISSION_OUTPUT_STANDARD_MISSION_OUTPUT_H_
#define MISSION_ANALYSIS_SRC_STANDARD_MISSION_OUTPUT_STANDARD_MISSION_OUTPUT_H_

#include <memory>
#include <string>
#include <vector>

#include <matplot/matplot.h>
#include <moduleBasics/report.h>

#include "../standard_mission/standard_mission.h"

/** \class StandardMissionOutput
 *  \brief Class for creating program outputs and managing output data
 */
class StandardMissionOutput {
 public:
    /* Constructor and destructor */
    /** \brief Constructor for class StandardMissionOutput
     * \param mission_ptr: reference to calculated mission
     */
    explicit StandardMissionOutput(StandardMission* mission_ptr);
    /** \brief Destructor for class StandardMissionOutput
     */
    virtual ~StandardMissionOutput();

    /** \brief Function to initialize all required pointers, objects and member variables to run the tool
     */
    void initialize();

 private:
    StandardMission* my_mission_ptr; /**< Pointer to global mission data */
    const std::shared_ptr<RuntimeIO>& rtIO; /**< Pointer to global mission analysis settings data */
    Report report_; /* Report object */

    /* File names */
    std::stringstream output_name; /**< Name of the output */
    std::string csv_file_name; /**< Name of the csv file */
    std::string report_file; /**< Name of the report file */
    std::string base_name_output_files; /**< Base name of the output files */
    bool requirements_checked; /**< Switch for requirements checked */

    /* SAR plot variables */
    double start_altitude; /**< Altitude at the start for the plot of the Specific Air Range [ft] */
    double end_altitude; /**< Altitude at the end for the plot of the Specific Air Range [ft] */
    bool SAR_plot_on; /**< Switch if SAR shall be plotted */
    std::vector<double> SAR_mach; /**< Vector of the Mach number for the plot of the Specific Air Range */
    std::vector<double> SAR_FL; /**< Vector of the FL for the plot of the Specific Air Range [100 ft] */
    std::vector<double> SAR_total_mass; /**< Vector of the aircraft mass for the plot of the Specific Air Range [t] */
    std::vector<double> specific_air_ranges; /**< Vector of Specific Air Ranges [m/kg]*/

    /* variable camber location */
    size_t first_filled_position; /**< First position in mission_profile holding variable camber data */

    /* Plot data */
    /** \struct MatplotAxisData
     *  \brief Struct to gather plotting data
     */
    struct MatplotAxisData {
        std::vector<double> data; // Vector filled with data points
        std::string name; // Name of the parameter
    };
    MatplotAxisData altitude_profile, bleed_air_profile, CAS_profile, C_L_profile, CoG_pos_profile, fuel_consumption_profile, i_stab_profile,
                        mach_profile, range_profile, shaft_power_profile, thrust_profile, TAS_profile; /**< plot profiles */
    std::vector<MatplotAxisData> flap_settings; // multiple flap settings possible -> vector needed

    /* Functions */
    /** \brief Function for generating the mission data in csv format
     */
    void generate_mission_profile_data();
    /** \brief Function for generating the plot data for the Specific Air Range in csv format
     */
    void generate_SAR_plot_data();
    /** \brief Function for generating the mission data in csv format
     */
    void generate_mission_plots();
    /** \brief Function for generating the plot scripts for the Specific Air Range in csv format
     */
    void generate_SAR_plots();
    /** \brief Function for generating a HTML report
     */
    void generate_html_report();
    /** \brief Function for generating a HTML report for the requirements mission
     */
    void generate_requirements_html_report();
    /** \brief Function for generating a Tex report
     */
    void generate_tex_report_body();
    /** \brief converts decimal hours into the hh:mm format
     *  \param time double: Time that shall be converted [s] 
     *  \return report std::string containing the new format
     */
    std::string timeFormater(double time);
    /** \brief setup for the matplot axes
     *  \param axes std::shared_ptr<matplot::axes_type>: to matplot axis element
     *  \param y_data const std::vector<std::vector<double>>&: Vector containing data vectors for the y axis
     *  \param y_label const std::string&: Label which shall be displayed on the y axis
     *  \param y2_switch bool: Switch to fill the y2 axis instead of y1 (true = y2 axis, false = y1 axis)
     *  \param x_data const std::vector<double>&: x axis data
     *  \param x_label const std::string&: Label which shall be displayed on the x axis
     */
    void set_matplot_axes(std::shared_ptr<matplot::axes_type> axes, const std::vector<std::vector<double>>& y_data, const std::string& y_label,
                                bool y2_switch = false, const std::vector<double>& x_data = {}, const std::string& x_label = std::string());
    /** \brief adding plots into matplot axes
     *  \param axes std::shared_ptr<matplot::axes_type>: to matplot axis element
     *  \param x_data MatplotAxisData: x axis data
     *  \param y1_data std::vector<MatplotAxisData>: Vector containing data vectors for the y1 axis
     *  \param legend_switch bool: Switch to fill a legend with the data entries
     *  \param y2_data std::vector<MatplotAxisData>: Vector containing data vectors for the y2 axis
     */
    void add_matplot_to_axes(std::shared_ptr<matplot::axes_type> axes, MatplotAxisData x_data, std::vector<MatplotAxisData> y1_data,
                                                bool legend_switch, std::vector<MatplotAxisData> y2_data = {});
    /** \brief gets the minimum or maximum value from a given vector
     *  \param data std::vector<double>&: Time that shall be converted
     *  \param min_max_switch bool: true = min, false = max
     *  \return double min/max value
     */
    double get_min_max_value(const std::vector<double>& data, bool min_max_switch);
};

#endif // MISSION_ANALYSIS_SRC_STANDARD_MISSION_OUTPUT_STANDARD_MISSION_OUTPUT_H_
