/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */
#include "aircraft.h"

#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>
#include <unitConversion/unitConversion.h>
#include "performance_assessment_settings.h"

#ifdef _WIN32
    #include <windows.h>
#endif  // _WIN32
#include <iostream>
#include <cmath>
#include <algorithm>
#include <vector>
#include <utility>
#include <string>

using std::endl;

aircraft::aircraft(const std::shared_ptr<RuntimeIO>& rtIO, const std::shared_ptr<performance_assessment_settings> mySettingsPt)
    :
    rtIO(rtIO),
    mySettingsPt(mySettingsPt),  // Pointer to the settings
    delta_isa(EndnodeReadOnly<double>(
        "requirements_and_specifications/requirements/top_level_aircraft_requirements/" + this->getMissionType() + "/delta_ISA").read(rtIO->acxml).value()),
    myMission(rtIO->getMissionDataDir() + "/" + rtIO->acxml.at("requirements_and_specifications/mission_files/" + this->getMissionType() + "_file/value"), delta_isa),
    myEngines(rtIO),
    polarXML(aixml::openDocument(rtIO->getAeroDataDir() + "/" + rtIO->acxml.at("analysis/aerodynamics/polar/polar_file/value"))),
    aero(rtIO->acxml, polarXML) {
    /* Constructor */
    acftName = rtIO->aircraft_type() + rtIO->aircraft_model();
    MTOM = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass").read(rtIO->acxml).value();
    MLM = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_landing_mass/mass_properties/mass").read(rtIO->acxml).value();
    OME = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/operating_mass_empty/mass_properties/mass").read(rtIO->acxml).value();
    MFM = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_fuel_mass/mass_properties/mass").read(rtIO->acxml).value();
    maxPayload = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/masses_cg_inertia/maximum_payload_mass/mass_properties/mass").read(rtIO->acxml).value();
    MMO = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_operating_mach_number")
        .read(rtIO->acxml).value();
    VMO = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_operating_velocity")
        .read(rtIO->acxml).value();
    MDive = EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/speed/dive_mach_number", 0.).read(rtIO->acxml).value();
    VDive = EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/speed/dive_velocity", 0.).read(rtIO->acxml).value();
    MInitialCruise = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/initial_cruise_mach_number")
        .read(rtIO->acxml).value();
    hInitialCruise = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/initial_cruise_altitude")
        .read(rtIO->acxml).value();
    hMaxOperating = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_operating_altitude")
        .read(rtIO->acxml).value();
    timeToClimb = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/design_mission/time_to_climb")
        .read(rtIO->acxml).value();
    designRange = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/mission/design_mission/range").read(rtIO->acxml).value();
    designPayload = EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/maximum_structrual_payload_mass")
        .read(rtIO->acxml).value();
    rangeMaxPayload = EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/range/range_max_payload_at_maximum_takeoff_mass", 0.).read(rtIO->acxml).value();
    rangeMaxFuel = EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/range/range_max_fuel_at_maximum_takeoff_mass", 0.).read(rtIO->acxml).value();
    payloadRangeMaxfuel = EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/range/payload_maximum_fuel_at_maximum_takeoff_mass",
        0.).read(rtIO->acxml).value();
    rangeFerry = EndnodeReadOnly<double>("/aircraft_exchange_file/assessment/performance/range/range_maximum_fuel_empty", 0.).read(rtIO->acxml).value();
    numberEngines = rtIO->acxml.getVector(("/aircraft_exchange_file/component_design/propulsion/specific/propulsion"), 1).size();
    S_Wing = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/aerodynamics/reference_values/S_ref").read(rtIO->acxml).value();
    span = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/aerodynamics/reference_values/b").read(rtIO->acxml).value();
    optCL = EndnodeReadOnly<double>("/aircraft_exchange_file/analysis/aerodynamics/lift_coefficients/C_LoptimumCruise").read(rtIO->acxml).value();
    frictionCoefficient = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/additional_requirements/landing_gear/friction_coefficient").read(rtIO->acxml).value();
    brakingCoefficient = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/additional_requirements/landing_gear/braking_coefficient").read(rtIO->acxml).value();

    factorVLOF = 1.1;
    factorV2 = 1.1;

    /* Requirements */
    myReqs.OEICruiseAltitude = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_altitude_one_engine_inoperative")
        .read(rtIO->acxml).value();
    myReqs.TOFL = EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/takeoff_distance")
        .read(rtIO->acxml).value();
    myReqs.LDN = EndnodeReadOnly<double>("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/landing_field_length")
        .read(rtIO->acxml).value();
    myReqs.approachSpeed = EndnodeReadOnly<double>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_approach_speed")
        .read(rtIO->acxml).value();
    myReqs.icao_aerodrome_reference_code = EndnodeReadOnly<std::string>(
        "/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/icao_aerodrome_reference_code").read(rtIO->acxml).value();
    this->myReqs.setRequiredClimbGradients(this->numberEngines);
    myChecks.TOFL = false;
    myChecks.LDN = false;
    myChecks.vApproach = false;
    myChecks.initCruiseAlt = false;
    myChecks.maxOptAlt = false;
    myChecks.OEICruiseAlt = false;
    myChecks.initCruiseM = false;
    myChecks.climbGradientSecondSegment = false;
    myChecks.climbGradientFinalSegment = false;
    if (OME + MFM > MTOM) {
        MFM = MTOM - OME;
    }
    // Energy provider
    myEnergyProviders.energyCarrierID_ground =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/ground")).value();
    myEnergyProviders.energyCarrierID_taxi =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/taxi")).value();
    myEnergyProviders.energyCarrierID_takeoff =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/takeoff")).value();
    myEnergyProviders.energyCarrierID_climb =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/climb")).value();
    myEnergyProviders.energyCarrierID_cruise =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/cruise")).value();
    myEnergyProviders.energyCarrierID_descent =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/descent")).value();
    myEnergyProviders.energyCarrierID_landing =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/landing")).value();

    // Determination of V2 for minimum climb gradients
    factorV2 = this->getMinimumFactorForV2();
    if (factorV2 <= 1.128) {  // Par CS25.107: v2 !> 1.2 * vS or 1.128 * vSR (vS1g)
        factorVLOF = factorV2 = 1.128;
    } else {
        factorVLOF = 1.128;
    }
    // Span limit
    decipher_icao_ref_code(myReqs.icao_aerodrome_reference_code);  // TODO(hansmann): Move span check somewhere else (e.g into factory)
    if (span < myReqs.span_limit) {
        myChecks.span_limit = true;
    } else {
        myChecks.span_limit = false;
    }

    /* Set atmosphere */
    this->atm.setAtmosphere(0., ISA_TEMPERATURE + myMission.deltaISA, ISA_PRESSURE);
}

aircraft::~aircraft() {
    // dtor
}

string aircraft::getMissionType() {
    string missionType;
    if (!mySettingsPt->designForMTOM && mySettingsPt->useStudyMissionForAnalysis) {
        missionType = "study_mission";
    } else {
        missionType = "design_mission";
    }
    return missionType;
}

double aircraft::getReferenceStallSpeed(const double& aircraftMass, const double& flightAltitude, const double& maximumLiftCoefficient) {
    return sqrt((aircraftMass * G_FORCE) / (0.5 * this->atm.getDensity(flightAltitude) * maximumLiftCoefficient * this->S_Wing));
}

double aircraft::v1gS_Clean(double Mass, double Altitude) {
    return 1.0 * this->getReferenceStallSpeed(Mass, Altitude, this->aero.getCLmax(0., "clean"));
}

double aircraft::v1gS_TO(double Mass, double Altitude) {
    return 1.0 * this->getReferenceStallSpeed(Mass, Altitude, this->aero.CLmaxTakeoff);
}

double aircraft::vLOF(double Mass, double Altitude) {
    return factorVLOF * this->getReferenceStallSpeed(Mass, Altitude, this->aero.CLmaxTakeoff);
}

double aircraft::v2(double Mass, double Altitude) {
    return factorV2 * this->getReferenceStallSpeed(Mass, Altitude, this->aero.CLmaxTakeoff);
}

double aircraft::vFTO(double Mass, double Altitude) {
    return 1.65 * this->getReferenceStallSpeed(Mass, Altitude, this->aero.CLmaxTakeoff);
}

double aircraft::vAPPR(double Mass, double Altitude) {
    return 1.23 * this->getReferenceStallSpeed(Mass, Altitude, this->aero.CLmaxLanding);
}

double aircraft::ROC(double Mass, double Alt, double Mach, double fACC, string Configuration, string Rating, double bleedOff, double shaftOff, bool OEI,
                     const aerodynamics& thisAERO) {
    // s. Scheiderer P.230 ff.
    // Determine engine carrier for climb phase
    int energy_carrier_id = 0;  // TODO(hansmann): Consideration of varying energy carriers and operating strategies (shall be covered by engine.lib or multi_engine_prop.lib)
    myEngines.calculate_N1_with_penalties(energy_carrier_id, Alt, Mach, atm, 1.0, Rating, bleedOff, shaftOff);  // Set operating point
    double ac_thrust = myEngines.get_thrust_aircraft(energy_carrier_id);
    double tempDrag(0.);
    if (OEI == true) {
        ac_thrust = ac_thrust * static_cast<double>(numberEngines - 1) / static_cast<double>(numberEngines);
        tempDrag += calcWindMillingDrag(0, 0, Mach, Alt);
    }
    try {
        tempDrag += thisAERO.getCruiseDrag(Mach, Alt, 0.0, Mass, Configuration, atm);
    } catch (const std::string& e) {
        tempDrag = ac_thrust;  // In case of missing polar data a gamma of 0 is forced
    }
    if (std::isnan(tempDrag)) {
        tempDrag = ac_thrust;  // In case of missing polar data (but NaN-flavoured)
    }
    double tempGamma = asin((ac_thrust - tempDrag) / (G_FORCE * Mass) / (1. + fACC));  // because not with const. TAS but with const. IAS or Ma is risen
                                                                                    //--> Introduction of facc: Commutation factor: facc = V/g * dV/dh; see table 11.1 p.232
    double tempLiftCoefficient = 2.*(G_FORCE * Mass * cos(tempGamma))
                          / (atm.getDensity(Alt) * pow(Mach * atm.getSpeedOfSound(Alt), 2) * S_Wing);
    try {
        tempDrag = thisAERO.getCLDrag(Mach, Alt, tempLiftCoefficient, Configuration, atm);
    } catch (const std::string& e) {
        tempDrag = ac_thrust;  // In case of missing polar data a ROC of 0 is forced
    }
    double tempROC = ((ac_thrust - tempDrag) / (G_FORCE * Mass) * (Mach * atm.getSpeedOfSound(Alt))) / (1. + fACC);
        myRuntimeInfo->debug << "------------------------------------------" << std::endl;
        myRuntimeInfo->debug << "ROC: " << tempROC << " m/s" << std::endl;
        myRuntimeInfo->debug << "Mass: " << Mass << " kg" << std::endl;
        myRuntimeInfo->debug << "Altitude: " << Alt << " m" << std::endl;
        myRuntimeInfo->debug << "Mach: " << Mach << std::endl;
        myRuntimeInfo->debug << "CL: " << thisAERO.getCL(Alt, Mach, 0.0, Mass, atm) << std::endl;
        myRuntimeInfo->debug << "CD: " << tempDrag / (atm.getDensity(Alt)/2 * pow(atm.getSpeedOfSound(Alt) * Mach, 2.) * this->S_Wing) << std::endl;
        myRuntimeInfo->debug << "Thrust: " << ac_thrust << std::endl;
        myRuntimeInfo->debug << "Drag: " << tempDrag << std::endl;
    return tempROC;
}

double aircraft::calcWindMillingDrag(int energy_carrier_id, int engine_id, double Mach, double Altitude) {
    double engine_inlet_diameter = myEngines.get_propulsors(energy_carrier_id).at(engine_id).engine.get_engine_dimensions().diameter;
    double tempCdWMJ(0.0);
    tempCdWMJ = 0.0785 * pow(engine_inlet_diameter, 2.) / S_Wing
                + 2. / (1. + 0.16 * pow(Mach, 2.)) * 0.92 * (1.0 - 0.92) * (PI / 4. * pow(engine_inlet_diameter, 2.)) / S_Wing;
    // tempCdWMJ = 0.75 * tempCdWMJ;
    return 0.5 * atm.getDensity(Altitude) * pow(Mach * atm.getSpeedOfSound(Altitude), 2.) * S_Wing * tempCdWMJ;
}

double aircraft::getMinimumFactorForV2() {
    double altitude(convertUnit(FOOT, METER, 400.));  // End of 2nd climb segment for standard departure according to FAR
    double factor(1.);  // Start value
    double tempClimbGradient(NAN);
    do {
        factor += 0.0025;
        double Mach = factor * this->v1gS_TO(this->MTOM, altitude) / this->atm.getSpeedOfSound(altitude);
        double tempTAS = convertUnit(MACH, TRUEAIRSPEED, altitude, this->atm, Mach);
        double tempROC_OEI = this->ROC(this->MTOM, altitude, Mach, 0.0, "climb", "takeoff", this->myMission.bleedOffTO,  this->myMission.shaftOffTO, true, this->aero);
        tempClimbGradient = tempROC_OEI / tempTAS * 100.;
    } while (tempClimbGradient < myReqs.climbGradientSecondTOSegment);
    return factor;
}

void aircraft::Requirements::setRequiredClimbGradients(const uint16_t& numberOfEngines) {
    if (numberOfEngines == 1) {
        myRuntimeInfo->err << "Climbing requirements are invalid for a single engine aircraft!" << endl;
        exit(1);
    } else if (numberOfEngines == 2) {
        this->climbGradientSecondTOSegment = 2.4;
        this->climbGradientFinalTOSegment = 1.2;
        this->climbGradientApproachOEI = 2.1;
    } else if (numberOfEngines == 3) {
        this->climbGradientSecondTOSegment = 2.7;
        this->climbGradientFinalTOSegment = 1.5;
        this->climbGradientApproachOEI = 2.4;
    } else if (numberOfEngines >= 4) {
        this->climbGradientSecondTOSegment = 3.0;
        this->climbGradientFinalTOSegment = 1.7;
        this->climbGradientApproachOEI = 2.7;
    }
    this->climbGradientApproachAEO = 3.2;
}

void aircraft::decipher_icao_ref_code(const std::string icao_ref_code) {
    if (icao_ref_code.length() != 6) {  // TODO(hansmann): Add checks for other letters
        throw std::invalid_argument("ICAO ref code malformed (length)!");
    }
    // First letter: Field length
    // (...)
    // Second letter: Wing span
    if (icao_ref_code[1] == 'A') {
        myReqs.span_limit = 15.;
    } else if (icao_ref_code[1] == 'B') {
        myReqs.span_limit = 24.;
    } else if (icao_ref_code[1] == 'C') {
        myReqs.span_limit = 36.;
    } else if (icao_ref_code[1] == 'D') {
        myReqs.span_limit = 52.;
    } else if (icao_ref_code[1] == 'E') {
        myReqs.span_limit = 65.;
    } else if (icao_ref_code[1] == 'F') {
        myReqs.span_limit = 80.;
    } else {
        throw std::invalid_argument("ICAO ref code pos 2 malformed!");  // TODO(hansmann): Add case of intentional noncompliance with ICAO ref code
    }
    // Third letter: Tail height
    // (...)
    // Fourth letter: Approach speed
    // (...)
}
