/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */
#ifndef PERFORMANCE_ASSESSMENT_SRC_TAW_DEFAULTMETHODS_LANDINGPERFORMANCE_LOW_FIDELITY_LANDINGPERFORMANCE_H_
#define PERFORMANCE_ASSESSMENT_SRC_TAW_DEFAULTMETHODS_LANDINGPERFORMANCE_LOW_FIDELITY_LANDINGPERFORMANCE_H_

#include <runtimeInfo/runtimeInfo.h>
#include <unitConversion/unitConversion.h>
#include <vector>
#include <memory>

#include "../../aircraft.h"
#include "../abstractLandingPerformance.h"

class aircraft;

using std::vector;

namespace low {

/**\class landingPerformance class contains the landing performance data
*
*/
class landingPerformance :public abstractLandingPerformance {
 public:
    /**\brief Function to calculate the landing distance for different altitudes and aircraft masses
    *\param theRunwaySlope Slope of the runway (rad)
    *\param theGroundFrictionBraking Friction coefficient at full braking (-)
    *\param theHeadWind Head wind speed (m/s)
    */
    void doLandingPerformance(double theRunwaySlope, double theGroundFrictionBraking, double theHeadWind) override;

    /**\brief Constructor for the landingPerformance object
    *\param theAircraft Address to the aircraft object
    */
    explicit landingPerformance(std::shared_ptr<aircraft> theAircraft);

    /**\brief Destructor of the landingPerformance object
    *
    */
    virtual ~landingPerformance();

 private:
    std::shared_ptr<aircraft> theAcftPt; /**< Pointer to theAircraft object */

    double runwaySlopeAngle; /**< Slope of the runway (rad) */
    double brakingFriction; /**< Friction coefficient for complete stop (-) */
    double windSpeed; /**< Wind speed (m/s) */

    /**\brief Function to calculate the available landing distance
    *   \param landingDistanceRequired double*: Pointer to current landing distance for dry runway at ISA+deltaISA [m]
    *   \param acMass Aircraft mass (kg)
    *   \param Altitude Airfield (aircraft) altitude (m)
    */
    void calcALD(double* landingDistanceRequired, const double& acMass, const double& Altitude);

    /**\brief Function to calculate the deceleration on ground
    *\param mass Aircraft mass (kg)
    *\param alt Aircraft altitude (m)
    *\param Velocity Aircraft speed (m/s)
    *\return Deceleration of the aircraft on ground (m/s^2)
    */
    double calcDeAccelerationGroundRoll(double mass, double alt, double Velocity);

    /**\brief Function  verifies if the landing distance can be maintained
    *\return True or false
    */
    bool checkForLDN();

    /**\brief Function verifies if the approach speed can be maintained
    *\return True or false
    */
    bool checkForApproachSpeed();

    /**\brief Function verifies if in case of missed approach the required gradient of climb with one engine inoperative can be maintained
    *\return True or false
    */
    void checkForApproachClimb();
};
}  // namespace low

#endif  // PERFORMANCE_ASSESSMENT_SRC_TAW_DEFAULTMETHODS_LANDINGPERFORMANCE_LOW_FIDELITY_LANDINGPERFORMANCE_H_

