/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */
#ifndef PERFORMANCE_ASSESSMENT_SRC_TAW_DEFAULTMETHODS_MISSIONDEFINITION_H_
#define PERFORMANCE_ASSESSMENT_SRC_TAW_DEFAULTMETHODS_MISSIONDEFINITION_H_

#include <moduleBasics/module.h>
#include <atmosphere/atmosphere.h>
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>
#include <unitConversion/unitConversion.h>

#include <cstdint>
#include <string>
#include <vector>

class node;

using std::vector;

/** \class missionDefinition contains all information about the flight mission
*
*/
class missionDefinition {
 public:
    std::string missionPhase; /**< name of the mission phase */

    double missionPayload; /**< Payload during the mission (kg) */
    double missionPAX; /**< Number of passengers during the mission (-) */
    double missionCargo; /**< Cargo mass during the mission (kg) */
    double missionRange; /**< Range of the mission (m) */
    double alternateRange; /**< Distance to the alternate airport (km) */
    double deltaISA; /**< Temperature delta to the ISA-Conditions (K) */
    double taxiTimeOrigin; /**< Time for taxiing at the origin (s) */
    double taxiTimeDestination; /**< Time for taxiing at the destination (s) */

    double crMach; /**< Desired cruise speed for the calculation of the fuel amount */

    double bleedOffTO; /**< Bleed air off take at take-off (kg/s) */
    double shaftOffTO; /**< Shaft power off take at take-off (W) */

    double bleedOffCr; /**< Bleed air off take at cruise(kg/s) */
    double shaftOffCr; /**< Shaft power off take at cruise (W) */

    uint16_t numberDepartureSteps; /**< Number of steps during depature in the mission.xml  (-) */
    uint16_t numberCruiseSteps; /**< Number of steps during cruise in the mission.xml (-) */
    uint16_t numberApproachSteps; /**< Number of steps during approach in the mission.xml (-) */

    /** \class averageMissionPoint Parameters at a specific mission point
    *
    */
    class averageMissionPoint {
     public:
        std::string simplifiedMODE; /**< Mission phase (TakeOff, Climb, Accelerate, Cruise, Descend, Landing) */

        double Altitude; /**< Aircraft altitude (m) */
        double Temperature; /**< Air temperature (K) */
        double CAS; /**< Calibrated air speed (m/s) */
        double TAS; /**< True air speed (m/s) */
        double Mach; /**< Mach number (-) */
        double density; /**< Air density (kg / pow(m,3)) */
        double dynamicPressure; /**< Dynamic pressure (kg/m*pow(s,2))*/

        double ROC; /**< Rate of climb (m/s)*/

        double shaftOffTakes; /**< Shaft off take (W)*/
        double bleedOffTakes; /**< Bleed air off take (kg/s) */

        averageMissionPoint() {
            Altitude = 0.0;
            Temperature = 0.0;
            CAS = 0.0;
            TAS = 0.0;
            Mach = 0.0;
            density = 0.0;
            dynamicPressure = 0.0;
            ROC = 0.0;
            shaftOffTakes = 0.0;
            bleedOffTakes = 0.0;
        }
    };

    /** \class missionStep Class with the important parameters during the flight mission
    *
    */
    class missionStep {
     public:
        averageMissionPoint Point;

        std::string missionPhase; /**< Phase during the mission (Departure, Cruise, Approach)*/
        std::string modeStep; /**< Mode of the segments in mission.xml */   // (Takeoff, Climb, Accelerate, ChangeSpeed_to_CAS, Climb_to_cruise,
                                                                            // ChangeSpeed_to_M, Cruise, ChangeFLconstROC, Descend_to_approach,
                                                                            // ChangeSpeed, Descend, LevelGSInterception, Landing)

        std::string rating; /**< Engine rating (TO, MaxCont, Climb, Idle, Cruise)*/
        double derating; /**< Engine derate (-)*/
        std::string configuration; /**< Aerodynamic configuration in the mission.xml (TakeOff, Climb, Clean, Approach, ApproachLG, Landing)*/

        double endAltitude; /**< Altitude at the end of the mission segment in the mission.xml (ft) */
        double endFL; /**< Flight level at the end of the mission segment in the mission.xml (-) */
        double endCAS; /**< Calibrated air speed at the end of the mission segment in the mission.xml (kts) */
        double endTAS; /**<  at the end of the mission segment in the mission.xml (kts) */
        double endMach; /**< Mach number at the end of the mission segment in the mission.xml (-)*/

        double endROC; /**<Rate of climb at the end of the mission segment in the mission.xml (ft / min)*/

        double relSegmentLength; /**< Current relative length in reference to the total cruise length (-) */

        double shaftOffTakes; /**< Shaft power off take (kW) */
        double bleedOffTakes; /**< Bleed air off take (kg/s) */

        missionStep()
            :
            missionPhase(""),
            modeStep(""),
            rating(""),
            configuration("") {
                derating = 0.0;
                endAltitude = 0.0;
                endFL = 0.0;
                endCAS = 0.0;
                endTAS = 0.0;
                endMach = 0.0;
                endROC = 0.0;
                relSegmentLength = 0.0;
                shaftOffTakes = 0.0;
                bleedOffTakes = 0.0;
        }
    };
    vector<missionStep> Steps; /**< List of all mission segments */

    vector<double> cruiseAlt; /**< List of the altitudes in cruise (ft) */
    vector<double> cruiseSegLength; /**< List of segment lengths in mission.xml (-) */

    /** \brief Constructor missionDefinition
    *
    */
    explicit missionDefinition(const std::string &fileNameAndDir, double delta_isa);

    /** \brief Destructor missionDefinition
    *
    */
    ~missionDefinition();

 private:
    const node& missionXML; /**< Constant reference to content of mission.xml-File */
    atmosphere myISA; /**< Object for ISA-Conditions */

    /** \brief Function reads the total mission file for several mission steps
    */
    void readMission();

    /** \brief Function reads a definded mission step from mission file and set the members of class missionStep
    *   \param missionPhase const string: Constant reference to a string of a certain mission phase. Allowed are "Departure"/"Cruise"/"Approach"
    *   \param numberTotalSteps const uint16_t: Constant reference to the total number of missionSteps in the mission phase
    */
    void readSteps(const std::string& missionPhase, const uint16_t& numberTotalSteps);

    /** \brief Fill the parameters for each mission step with the data from mission.xml
    * \param XML Subpath in the mission.xml file
    */
    void getMissionSegment(std::string XML);

    /** \brief Get the parameters for each mission step
    */
    void getMissionPoint();

    /** \brief Get the average bleed air and shaft power off take during cruise
    */
    void getTyicalOffTakes();
};

#endif  // PERFORMANCE_ASSESSMENT_SRC_TAW_DEFAULTMETHODS_MISSIONDEFINITION_H_
