/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */
#ifndef PERFORMANCE_ASSESSMENT_SRC_TAW_DEFAULTMETHODS_PERFORMANCE_ASSESSMENT_OUTPUT_H_
#define PERFORMANCE_ASSESSMENT_SRC_TAW_DEFAULTMETHODS_PERFORMANCE_ASSESSMENT_OUTPUT_H_

#include <moduleBasics/report.h>
#include <moduleBasics/module.h>
#include <matplot/matplot.h>
#include <moduleBasics/plot.h>

#include <string>
#include <memory>
#include <vector>
#include <limits>

#include "../../toolinfo.h"

namespace taw {
class performance_assessment_factory;
}

class performance_assessment_output {
 public:
    performance_assessment_output(
        std::shared_ptr<abstractPayloadRange> abstractPayloadRangePt,
        std::shared_ptr<abstractEnginePerformance> abstractEnginePerformancePt,
        std::shared_ptr<abstractCeilingPerformance> abstractCeilingPerformancePt,
        std::shared_ptr<abstractFlightEnvelopes> abstractFlightEnvelopesPt,
        std::shared_ptr<abstractSARPerformance> abstractSARPerformancePt,
        std::shared_ptr<abstractTakeOffPerformance> abstractTakeOffPerformancePt,
        std::shared_ptr<abstractLandingPerformance> abstractLandingPerformancePt,
        std::shared_ptr<abstractVnDiagram> abstractVnDiagramPt,
        std::shared_ptr<aircraft> anAcftPt);
    ~performance_assessment_output();

    const std::shared_ptr<aircraft> myAcftPt;
    const std::shared_ptr<performance_assessment_settings> mySettingsPt;
    const std::shared_ptr<RuntimeIO>& rtIO;

    std::shared_ptr<abstractPayloadRange> abstractPayloadRangePt;
    std::shared_ptr<abstractEnginePerformance> abstractEnginePerformancePt;
    std::shared_ptr<abstractCeilingPerformance> abstractCeilingPerformancePt;
    std::shared_ptr<abstractFlightEnvelopes> abstractFlightEnvelopesPt;
    std::shared_ptr<abstractSARPerformance> abstractSARPerformancePt;
    std::shared_ptr<abstractTakeOffPerformance> abstractTakeOffPerformancePt;
    std::shared_ptr<abstractLandingPerformance> abstractLandingPerformancePt;
    std::shared_ptr<abstractVnDiagram> abstractVnDiagramPt;

    Report theReport;
    std::string theProgramName;

    vector<string> SAR_plot_names;
    vector<string> engine_performance_plot_names;

    /** \brief Function generates the body of HTML report
    *   \details the function is full virtual and is implemented in each program; it defines the content BETWEEN the <body> tag of the HTML-report
    *   \return report a stringstream of the information for the body-tag of HTML-Report
    */
    void generateHtmlBody();

    /** \brief Method writes the results to the AcftExchangeFile XML.
    * \return void
    */
    void writeXmlOutput();

     /** \brief Method for generating plot data in csv format
    * \return void
    */
    void generatePlotData();

    /** \brief Generate all relevant Plots
    * \return void
    */
    void generatePlots();

    /** \brief Generate the payload-range diagram
    */
    void generatePlotDataPayloadRange();

    /** \brief Generate payload-range diagram
    */
    void generate_plot_payload_range();

    /** \brief Generate a csv file with the engine performance results
    */
    void generate_engine_performance_data();

    /** \brief Generate the plot for engine performance
    */
    void generate_plot_engine_performance();

    /** \brief Generate the plot of the aircraft ceiling
    */
    void generateCeilingPlotData();

    /** \brief Generates the ceiling performance plot
    */
    void generate_plot_ceiling_performance();

    /** \brief Generate the plot data of the flight envelope
    */
    void generate_plot_flight_envelope();

    /** \brief Generate the plot data of the flight envelope
    */
    void generateFlightEnvelopeData();

    /** \brief Generate the specific air range plot
    */
    void generateSARPlotData();

    /** \brief Generate the optimum specific air range plot
    */
    void generate_best_SAR_Data();

    /** \brief Generates the specific air range plot
    */
    void generate_plot_SAR_performance();

    // /** \brief Set the settings for the optimum specific air range plot
    // */
    // void generateBestSARPlotScript();

    /** \brief Generate the plot of the take-off field length
    */
    void generateTOFLPlotData();

    // /** \brief Set the settings for the take-off field length plot
    // */
    void generate_plot_takeoff_field_length();

    /** \brief Generate the plot of the balanced field length
    */
    void generateBLFLPlotData();

    /** \brief Generates the balanced field length plot
    */
    void generate_plot_balanced_field_length();

    /** \brief Generates the csv with the landing field length data
    */
    void generateLDNPlotData();

    /** \brief Generates the plot of the landing field length performance assessment results
    */
    void generate_plot_landing_performance();

    /** \brief Generate the V-n diagram
    */
    void generateVnPlotData();

    /** \brief Generates the plot of the V-n-Diagram 
    */
    void generate_plot_v_n_diagram();
};

#endif  // PERFORMANCE_ASSESSMENT_SRC_TAW_DEFAULTMETHODS_PERFORMANCE_ASSESSMENT_OUTPUT_H_
