/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 * 
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */
#include "performance_assessment_settings.h"

using std::endl;

performance_assessment_settings::performance_assessment_settings(const std::shared_ptr<RuntimeIO>& rtIOin)
    :
    rtIO(rtIOin),
    designForMTOM(rtIO->moduleConfig.at("/module_configuration_file/program_settings/modes/mtom_design/value")),
    useStudyMissionForAnalysis(rtIO->moduleConfig.at("/module_configuration_file/program_settings/modes/use_study_mission_for_analysis/value")) {
    /* Constants for start/runway calculations */
    myField.runwaySlope = PI / 180. * EndnodeReadOnly<double>("/module_configuration_file/program_settings/constants_for_performance_checks/field_performance/runway_slope")
        .read(rtIO->moduleConfig).value();
    myField.headWind = EndnodeReadOnly<double>("/module_configuration_file/program_settings/constants_for_performance_checks/field_performance/head_wind")
        .read(rtIO->moduleConfig).value();

    if (false) {  // Vn-Diagram is temporarily disabled
        /* Constants for the Representation of the Vn Diagram */
        myVnDiagramConstants.maxNClean = EndnodeReadOnly<double>("/module_configuration_file/program_settings/constants_for_performance_checks/vn_diagram/max_n_clean")
            .read(rtIO->moduleConfig).value();
        myVnDiagramConstants.minNClean = EndnodeReadOnly<double>("/module_configuration_file/program_settings/constants_for_performance_checks/vn_diagram/min_n_clean")
            .read(rtIO->moduleConfig).value();
        myVnDiagramConstants.maxNFlaps = EndnodeReadOnly<double>("/module_configuration_file/program_settings/constants_for_performance_checks/vn_diagram/max_n_flaps")
            .read(rtIO->moduleConfig).value();
        myVnDiagramConstants.minNFlaps = EndnodeReadOnly<double>("/module_configuration_file/program_settings/constants_for_performance_checks/vn_diagram/min_n_flaps")
            .read(rtIO->moduleConfig).value();
    }

    /* Engine Performance Settings */
    if (false) {  // Engine performance is tremporarily disabled
        int nPerfSettings = rtIO->moduleConfig.getVector("/module_configuration_file/program_settings/constants_for_performance_checks/engine_performance/ratings/rating", 1)
        .size();
        auto aVector = rtIO->moduleConfig.getVector("/module_configuration_file/program_settings/constants_for_performance_checks/engine_performance/ratings", 1);
        for (int i = 0; i <= nPerfSettings - 1; i++) {
            string tempRating = rtIO->moduleConfig.at("/module_configuration_file/program_settings/constants_for_performance_checks/engine_performance/ratings/rating@"
                + num2Str(i+1)+"/value");
            enginePerformance.Rating.push_back(tempRating);
        }
    }
    overwrite_initial_values = EndnodeReadOnly<bool>("/module_configuration_file/program_settings/performance_checks/flight_envelope_performance/overwrite_initial_values", true)
        .read(rtIO->moduleConfig).value();
}

void performance_assessment_settings::checkForValidity(double Value, double lowerBound, double upperBound, const string &Text) {
    if (std::isnan(Value)) {  // Test for NaN
        myRuntimeInfo->err << "Value is NaN!" << endl;
        myRuntimeInfo->err << "Parameter off:" << endl;
        myRuntimeInfo->err << Text << endl;
        myRuntimeInfo->err << "Abort!" << endl;
        exit(1);
    }
    if (std::isinf(Value)) {  // Test for real number
        myRuntimeInfo->err << "Value is Infinite!" << endl;
        myRuntimeInfo->err << "Parameter aus:" << endl;
        myRuntimeInfo->err << Text << endl;
        myRuntimeInfo->err << "Abort!" << endl;
        exit(1);
    }
    if (Value > upperBound || Value < lowerBound) {
        myRuntimeInfo->err << "Invalid value range!" << endl;
        myRuntimeInfo->err << "Parameter off:" << endl;
        myRuntimeInfo->err << Text << endl;
        myRuntimeInfo->err << "With the value:" << endl;
        myRuntimeInfo->err << Value << " (permitted is " << lowerBound << " <= X <= " << upperBound << ")" << endl;
        myRuntimeInfo->err << "Abort!" << endl;
        exit(1);
    }
}

performance_assessment_settings::~performance_assessment_settings() {
}
