/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

/* The following statements import necessary header file */
#include "conventionalATA30.h"                                                   // Import of header file from src/aircraftSystems folder for conventionalATA30

/* The following statements import necessary libraries */
#include <algorithm>                                                                            // Import c++ standard library for using algorithm
#include <unordered_map>                                                                        // Import c++ standard library for using unordered map

/* The following statements import necessary header files */
#include <runtimeInfo/runtimeInfo.h>                                             // Import of header file from external runtime library for using runtimeInfo
#include <atmosphere/atmosphere.h>                                               // Import of header file from external atmosphere library for using atmosphere
#include <unitConversion/constants.h>                                            // Import of header file from external unitConversion library for constants
#include <unitConversion/unitConversion.h>                                       // Import of header file from external unitConversion library for unitConverison
#include <standardFiles/functions.h>                                             // Import of header file from src folder standardFiles library for functions
#include "wingAntiIcing/liquidWaterContent.h"                                    // Import of header file from src/aircraftSystems/wingAntiIcing folder for liquidWaterContent
#include "wingAntiIcing/wingAntiIcing.h"                                         // Import of header file from src/aircraftSystems/wingAntiIcing folder for wingAntiIcing
#include "../geometricalFunctions.h"                                             // Import of header file from src folder for using geometricalFunctions

/* Constructor for conventialATA30 */
conventionalATA30::conventionalATA30(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO) {
    ATA = "ATA-30";
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

/* Destructor for conventialATA30 */
conventionalATA30::~conventionalATA30() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA30::readParameters(const node& config) {
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/ice_rain_protection_system").read(config).value();
    std::string unit;
    std::string ATA30config = "/module_configuration_file/program_settings/systems_constants/ice_rain_protection";
    unit = std::string(config.at(ATA30config + "/top_operating_altitude/unit"));
    if (unit == "ft") {
        operatingAltitude   = convertUnit(FOOT, METER,
                                                static_cast<double>(config.at(ATA30config + "/top_operating_altitude/value")));
    } else if (unit == "m") {
        operatingAltitude = config.at(ATA30config + "/top_operating_altitude/value");
    } else if (unit == "km") {
        operatingAltitude   = convertUnit(KILO, METER, METER,
                                                  static_cast<double>(config.at(ATA30config + "/top_operating_altitude/value")));
    } else {
        myRuntimeInfo->err << "The unit of the IceRain.operatingAltitude is unknown: " << unit << std::endl;
        exit(1);
    }
    data_->configuration.checkForValidity(operatingAltitude, 0.0, 45000., "IceRain.operatingAltitude");
    engineAntiIce   = config.at(ATA30config + "/engine_anti_ice/value");
    data_->configuration.checkForValidity(engineAntiIce, 0.0, 5.0, "IceRain.engineAntiIce");
    elecPowerDeparture  = config.at(ATA30config + "/elec_power_departure/value");
    data_->configuration.checkForValidity(elecPowerDeparture, 0.0, 300000., "IceRain.elecPowerDeparture");
    elecPowerCruise  = config.at(ATA30config + "/elec_power_cruise/value");
    data_->configuration.checkForValidity(elecPowerCruise, 0.0, 300000., "elecPowerCruise");
    elecPowerApproach  = config.at(ATA30config + "/elec_power_approach/value");
    data_->configuration.checkForValidity(elecPowerApproach, 0.0, 300000., "elecPowerApproach");
    elecPowerLand  = config.at(ATA30config + "/elec_power_land/value");
    data_->configuration.checkForValidity(elecPowerLand, 0.0, 300000., "elecPowerLand");
    wingAntiIcingSwitch = config.at(ATA30config + "/switch_off_wing_anti_icing/value");
    wingSkinThickness = config.at(ATA30config + "/skin_thickness/value");
    data_->configuration.checkForValidity(wingSkinThickness, 0.000001, 1., "skinThickness");
    wingRelSpanAntiIcingStart_str = std::string(config.at(ATA30config + "/rel_span_anti_icing_start/value"));
    wingSkinHeatConductivity = config.at(ATA30config + "/heat_conductivity_wing/value");
    data_->configuration.checkForValidity(wingSkinHeatConductivity, 50, 500., "heatConductivity");
    efficiencyOfElectroThermalSystem = config.at(ATA30config + "/efficiency_factor_electro_thermic_anti_icing/value");
    data_->configuration.checkForValidity(efficiencyOfElectroThermalSystem, 0.1, 1.0, "efficiencyOfElectroThermalSystem");
    designDropDiameter = config.at(ATA30config + "/drop_diameter/value");
    data_->configuration.checkForValidity(designDropDiameter, 15., 40., "dropDiameter");
    numberOfPowerSources = config.at(ATA30config + "/shaft_power_sources/number_of_power_sources/value");
    double operationFactor_gesamt(0.);
    for (int i(0); i < numberOfPowerSources; i++) {
        powerSources.push_back(PowerSource());
        powerSources.back().type =
            std::string(config.at(ATA30config + "/shaft_power_sources/power_source@" + num2Str(i) + "/type/value"));
        powerSources.back().sourceID =
            config.at(ATA30config + "/shaft_power_sources/power_source@" + num2Str(i) + "/source_ID/value");
        powerSources.back().operationFactor =
            config.at(ATA30config + "/shaft_power_sources/power_source@" + num2Str(i) + "/operation_factor/value");
        operationFactor_gesamt += Rounding(powerSources.back().operationFactor, 10);
    }
    if (operationFactor_gesamt != 1.) {
        myRuntimeInfo->err << "The work distribution of the \"PowerSources\" for Ice and Rain Protection does not add up to 100% ("
                           << operationFactor_gesamt * 100 << "%). adjust in config. Abort!"
                           << std::endl;
        exit(1);
    }
    percentageOfOME = checkBoundaries(config.at(ATA30config + "/percentage_of_OME/value"), 0, true, 1, true);
    efficiencyFactorBleedAir    = config.at("/module_configuration_file/program_settings/systems_constants/bleed_air/efficiency_factor_bleed_air_system/value");
    data_->configuration.checkForValidity(efficiencyFactorBleedAir, 0.0, 1.0, "BleedAir.efficiencyFactorBleedAir");
    temperatureBleedAir         = config.at("/module_configuration_file/program_settings/systems_constants/bleed_air/temperature_bleed_air/value");
    unit = std::string(config.at("/module_configuration_file/program_settings/systems_constants/bleed_air/temperature_bleed_air/unit"));
    if (unit == "K") {
        // No conversion required
    } else if (unit == "C") {
        // conversion from Celsius to Kelvin
        temperatureBleedAir += 273.15;
    } else {
        myRuntimeInfo->err << "The unit of bleed-air temperature is not known.!" << std::endl;
        myRuntimeInfo->err << "Bleed-Air Temperature: " << temperatureBleedAir << " " << unit << std::endl;
        exit(1);
    }
    data_->configuration.checkForValidity(temperatureBleedAir, 0.0, 1000.0, "BleedAir.temperatureBleedAir");
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA30::updateXML() {
    updateMass("de_icing", Mass, ATA);
}

/* Calculation of the system power */
void conventionalATA30::calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
        const std::vector<powerProfile>& allPreviousSystemsHydraulic,
        const std::vector<powerProfile>& allPreviousSystemsElectric,
        const powerProfile& allPreviousSystemsHeat) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(allPreviousSystemsBleed);
    UNUSED(allPreviousSystemsHydraulic);
    UNUSED(allPreviousSystemsElectric);
    UNUSED(allPreviousSystemsHeat);
    /* Calculate Systems Power */
    initializePowerProfile();
    if (Operating == true) {
        getPower();
    }
    calculateMaxMissionPower();
}

/* Calculation of the skin mass */
void conventionalATA30::calculateSinkMass(const powerProfile& sourceBleed,
        const std::vector<powerProfile>& sourceElectric,
        const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceElectric);
    UNUSED(sourceHydraulic);
    /* Calculate Systems Mass */
    Mass = scaling_factor * getMass();
    refPoint = getRefPoint();
}

/* Determination of the power */
void conventionalATA30::getPower() {
    int nrEngines(data_->data.Propulsion.numberEngines);
    // NOTE (aigner#1#): NEW: designPower is calculated from the maximum required power
    std::vector<double> electricalPower;
    electricalPower.push_back(elecPowerDeparture);//Departure
    electricalPower.push_back(elecPowerCruise);//Cruise
    electricalPower.push_back(elecPowerApproach);//Approach
    electricalPower.push_back(elecPowerLand);//Land
    /* Wing Anti-Icing */
    std::unordered_map<std::string, double> theRelSpanPositionsAntiIcing;
    // Get relative start and end positions for ice protection system
    geometricalFunctions myGeoFunctions;
    theRelSpanPositionsAntiIcing = myGeoFunctions.getRelSpanAntiIcing(data_->theWing.at(0), wingRelSpanAntiIcingStart_str);
    wingAntiIcing theWingAntiIcing(data_->theWing.at(0).aerodynamicSurface, wingSkinThickness, theRelSpanPositionsAntiIcing["start"],
                                                 theRelSpanPositionsAntiIcing["end"], wingSkinHeatConductivity, data_->theWing.at(0).frontSparPos);
    liquidWaterContent theLiquidWaterContent;
    double designBleedAir(0.), designAltitude(TRANSITION_ALTITUDE);
    double theLWC = theLiquidWaterContent.getLiquidWaterContentForMaximumIcing(Rounding(convertTemperature(KELVIN, CELCIUS,
                                                                                                           data_->mission.myISA.getTemperature(designAltitude)), 2),
                                                                               designDropDiameter, 5.5);
    // calibratedAirspeedUpToTransAltitude is used instead of CAS over transition altitude,
    // because TAS approaching velocities as high as VMO would not be used as design point for IPS system (not designed for worst conditions for each parameter)
    double trueAirspeed_SI(convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, designAltitude, data_->mission.myISA, data_->calibratedAirspeedUpToTransAltitude));
    double designMach = convertUnit(TRUEAIRSPEED, MACH, designAltitude, data_->mission.myISA, trueAirspeed_SI);
    theWingAntiIcing.setPowerConsumption(designMach, designAltitude, &data_->mission.myISA, theLWC);
    designBleedAir = theWingAntiIcing.getBleedAirConsumption(temperatureBleedAir, efficiencyFactorBleedAir);
    bleedAir.designPower = designBleedAir + static_cast<double>(nrEngines) * engineAntiIce;
    myRuntimeInfo->debug << "Bleed air design power of ATA 30: " << bleedAir.designPower << std::endl;
    for (size_t n(0); n < Electric.size(); n++) {
        for (size_t j(0); j < powerSources.size(); j++) {
            double operationFactor_tmp(powerSources.at(j).operationFactor);
            if (powerSources.at(j).type == "Electric" && powerSources.at(j).sourceID == static_cast<int>(n + 1)) {
                Electric.at(n).designPower = *max_element(electricalPower.begin(), electricalPower.end()) * operationFactor_tmp;
            }
        }
    }
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        bleedAir.missionPower.at(i).baseLoad = 0.0;
        heatLoad.missionPower.at(i).baseLoad = 0.0;
        /*Loop over hydraulic circuits*/
        for (size_t n(0); n < Hydraulic.size(); n++) {
            Hydraulic.at(n).missionPower.at(i).baseLoad = 0.0;
        }
        for (size_t n(0); n < Electric.size(); n++) {
            Electric.at(n).missionPower.at(i).baseLoad = 0.0;
        }
        /* Required bleed air of wing anti icing */
        if (data_->mission.Steps.at(i).Point.Altitude < operatingAltitude && wingAntiIcingSwitch == false) {
            /* Pneumatic demand */
            bleedAir.missionPower.at(i).baseLoad += designBleedAir;
        } else {
            bleedAir.missionPower.at(i).baseLoad = 0.0;
        }
        /* Demand of the engines */
        bleedAir.missionPower.at(i).baseLoad += static_cast<double>(nrEngines) * engineAntiIce;
        /* Required electric power */
        for (size_t n(0); n < Electric.size(); n++) {
            double baseLoad_tmp(0.);
            if (data_->mission.Steps.at(i).missionPhase == "Departure") {
                baseLoad_tmp = elecPowerDeparture;
            } else if (data_->mission.Steps.at(i).missionPhase == "Cruise") {
                baseLoad_tmp = elecPowerCruise;
            } else if (data_->mission.Steps.at(i).missionPhase == "Approach") {
                baseLoad_tmp = elecPowerApproach;
            } else if (data_->mission.Steps.at(i).missionPhase == "Landing") {
                baseLoad_tmp = elecPowerLand;
            } else {
                myRuntimeInfo->err << "Unknown mission phase in electrical energy calculation for ice and rain protection. Abort!" << std::endl;
                myRuntimeInfo->err << "mission phase: " << data_->mission.Steps.at(i).missionPhase << std::endl;
                exit(1);
            }
            for (size_t j(0); j < powerSources.size(); j++) {
                double operationFactor_tmp(powerSources.at(j).operationFactor);
                if (powerSources.at(j).type == "Electric" && powerSources.at(j).sourceID == static_cast<int>(n + 1)) {
                    Electric.at(n).missionPower.at(i).baseLoad += baseLoad_tmp * operationFactor_tmp;
                }
            }
        }
    }
}

/* Determination of the mass */
double conventionalATA30::getMass() {
    /* Source: [Sch04]. Mass of ATA30 is about 2% of the total system weight. **/
    /* Source: [Sch04]. System weight depending on MTOM **/
    // TODO(risse#2#): Mantis ID 0000211: comment kri:
    // [Ste10] Eq. (3.11) verweist hierzu auf [Ray18] -> dort 2% von MTOM, s. Eq. (15.43)
    // der prozentsatz wurde im Rahmen von CeRAS aber an A320-Wert aus internem Dokument angepasst
    // -> 32 kg / ~42 to -> also eher 1/1000 (0.2 Prozent MTOM waere ca. 150 kg)
    // am besten nochmal ueberpruefen und ggf. bessere Formel implementieren
    return percentageOfOME * data_->data.Mass.OME;
}

/* Determination of the reference point */
Vec3 conventionalATA30::getRefPoint() {
    /* Assumption: center of gravity at 45% fuselage length **/
    double fuselage_length = geom2::measure::length(data_->theFuselage.back());
    return Vec3(data_->theFuselage.back().origin[0] + 0.45 * fuselage_length, data_->theFuselage.back().origin[1], data_->theFuselage.back().origin[2]);
}
