/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_COMMON_AIRCRAFTSYSTEMS_HIGHLIFTSYSTEM_ACTUATIONANDSUPPORT_H_
#define SYSTEMS_DESIGN_SRC_COMMON_AIRCRAFTSYSTEMS_HIGHLIFTSYSTEM_ACTUATIONANDSUPPORT_H_

#include <string>
#include <vector>

/** \class ActuationAndSupport
 * \brief contains all functions and object for actuation and support of the high lift system
 */
class ActuationAndSupport {
 private:
    double l_m_Panel;/**< [m] length of the panel  */
    double S_ref;/**< [m^2] reference area of the device */
    int numberActuators;/**< number of actuators for the device  */
    double Density;/**< [kg/m^3] air density  */
    double Speed;/**< [m/s] air speed */

    /** \brief generalized linear function
     */
    class linearFunction {
     public:
        double m;
        double c;

        linearFunction() {
            m = 0.0;
            c = 0.0;
        }
    };

    linearFunction liftCoefficient;/**< linear function for the lift coefficient */
    linearFunction momentCoefficient;/**< linear function for the moment coefficient */
    linearFunction dragCoefficient;/**< linear function for the drag coefficient */

    double deviceAoA;/**<[rad] angle of attack of the device with respect to wing chord*/

    double deploymentXFlap;/**< [m] deployment of the flap in x-direction */
    double deploymentZFlap;/**< [m] deployment of the flap in z-direction */
    double deploymentTrack;/**< [m] lenght of the flap track */
    double b1;             /**< [m] dimension of push rod in idealized flap kinematics */
    double a;              /**< [m] dimension of push rod in idealized flap kinematics */

    double deltaXActuator; /**< [m] difference between x-coordinate of point A C of idealized flap kinematics */
    double deltaZActuator; /**< [m] difference between x-coordinate of point A C of idealized flap kinematics */

    double k1;             /**< [-] deltaXActuator divided by deltaXFlap */
    double k2;             /**< [-] deltaZActuator divided by deltaZFlap */

    double xA;             /**< [m] x-coordinate Point A in idealized flap kinematics */
    double yA;             /**< [m] y-coordinate Point A in idealized flap kinematics */
    double xB;             /**< [m] x-coordinate Point B in idealized flap kinematics */
    double yB;             /**< [m] y-coordinate Point B in idealized flap kinematics */
    double xC;             /**< [m] x-coordinate Point C in idealized flap kinematics */
    double yC;             /**< [m] y-coordinate Point C in idealized flap kinematics */
    double xD;             /**< [m] x-coordinate Point D in idealized flap kinematics */
    double yD;             /**< [m] y-coordinate Point D in idealized flap kinematics */

    double alphaA;         /**< [rad] angle between wing chordline and x-axis at the start of the deflection */
    double alphaE;         /**< [rad] angle between wing chordline and x-axis at the end of the deflection */
    double betaA;          /**< [rad] angle of flap kinematic at the start of the deflection */
    double betaE;          /**< [rad] angle of flap kinematic at the end of the deflection */

    double gamma;          /**< [rad] angle of flap kinematic and chordline */

    double delta_rad;      /** ??? */

    double i_Pinion;       /**< [-] gear ratio */
    double r_Pinion;       /**< [m] radius pinion */
    double r_zs;           /**< [m] radius rak slat aktuator */

    double schranke;       /** ??? */

    std::string rotationDirection;/**< Direction of rotation of the actuators */

    double actuatorRotationAngle_Begin;/**<[rad] start rotation angle of the actuator  */
    double actuatorRotationAngle_End;/**<[rad] end rotation angle of the actuator */

    double timeForDeployment;/**< [s] time for actuator deployment  */


    /** \brief function tests the angle phiA
     *
     * \param _phiA_rad double: [rad] angle phiA
     * \return bool
     *
     */
    bool angle_test_phiA(double _phiA_rad) const;

    /** \brief function tests the angle phiE
     *
     * \param _phiA_rad double: [rad] angle phiE
     * \return bool
     *
     */
    bool angle_test_phiE(double _phiE) const;

    /**
     * \brief calculates the kinematics of the slat
    */
    void calculateKinematic_Slat();
    /**
     * \brief calculates the kinematics of the flap
    */
    void calculateKinematic_Flap();

    /** \brief calculates the flap actuator torque
     *
     * \param Angle double: [rad]
     * \return double m_akt: [Nm] flap actuator torque
     *
     */
    double calcMomentFlap(double Angle);

    /** \brief calculates the slat actuator moment
     *
     * \param Angle double: [rad]
     * \return double m_akt: [Nm] slat actuator torque
     *
     */
    double calcMomentSlat(double Angle);


    /** \brief calculates the maximum torqe
     *
     * \return double maxMoment: [Nm] maximum actuation torque
     *
     */
    double getMaxMoment();

    /**
     * \brief calculates beta in dependance of phi for the flap
     * 
     * \param Phi double: phi angle of the flap
     * 
     * \return double: beta angle of the flap
    */
    double beta_von_phiFlap(double Phi);

    /**
     * \brief test of flap kinematics
     * 
     * \param _m double: gradient
     * \param _k double
     * 
     * \return bool
    */
    bool test_mFlap(double _m, double _k) const;

 public:
    double Mass;/**< [kg] mass of the actuation and support system  */

    /** \brief contains member objects wit respect to torque moment
     */
    class moment {
     public:
        double actuationAngle;/**<[rad] actuation angle  */
        double actuationMoment;/**<[Nm] actuation torque moment  */

        moment() {
            actuationAngle = 0.0;
            actuationMoment = 0.0;
        }
    };
    std::vector<moment> actuationMoments;/**< vector containing the actuation moments */

    double maxActuationMoment;/**<[Nm] max. actuation torque moment  */
    double actuationWork;/**<[J] actuation work*/
    double actuationPower;/**<[W] actuation power*/

    double efficiencyFactor;/**< [-] efficiency factor. Set for specific slat/flap types */
    double rotationSpeedActuator;/**< rotational speed of the actuator [s] */

    /** \brief calculates the mass of the actuation and support system
     *
     * \param deviceType string: type of the device
     * \param Sref double: [m^2] reference area of the device
     * \param Stations int: number of stations
     * \return void
     *
     */
    void getMass(const std::string& deviceType, double Sref, int Stations);

    /**
     * \brief calculates actuation moments
     * 
     * \param deviceType string
    */
    void calcActuationMoments(const std::string& deviceType);
    /**
     * \brief calculates work load depending on actuation moments and angles
    */
    void calcWorkLoad();
    /**
     * \brief calculates power from actuation work and time for deployment
    */
    void calcPower();

    /** \brief constructor
     *
     * \param Type string
     * \param _l_m_Panel double
     * \param _S_ref double
     * \param _numberActuators int
     * \param _Density double
     * \param _Speed double
     *
     */
    ActuationAndSupport(const std::string& Type, double _l_m_Panel, double _S_ref, int _numberActuators, double _Density, double _Speed);
    /** \brief destructor
     *
     * \return virtual
     *
     */
    virtual ~ActuationAndSupport();
};

#endif // SYSTEMS_DESIGN_SRC_COMMON_AIRCRAFTSYSTEMS_HIGHLIFTSYSTEM_ACTUATIONANDSUPPORT_H_
