/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "powerProfile.h"

#include <cmath>

#include <standardFiles/functions.h>
#include <unitConversion/constants.h>

missionLoads::missionLoads()
    :
    baseLoad(0.0) {
}

missionLoads::~missionLoads() {
}


powerProfile::powerProfile()
    :
    designPower(0.0),
    maxMissionPower(0.0) {
    //ctor
}

powerProfile::~powerProfile() {
    //dtor
}

double powerProfile::getMaxPower() const {
    double maxPower{0.};
    for (size_t i = 0; i < this->missionPower.size(); i++) {
        double stepPower = fabs(this->missionPower.at(i).baseLoad);
        for (size_t j = 0; j < this->missionPower.at(i).peakLoad.size(); j++) {
            stepPower += fabs(this->missionPower.at(i).peakLoad.at(j));
        }
        if (stepPower >= maxPower) {
            maxPower = stepPower;
        }
    }
    return maxPower;
}


powerProfile getPowerProfileSum(std::vector<powerProfile> profileVector) {
    powerProfile profileSum;
    //Power of the accumulated powerProfile by addition of the individual power from the vector
    for (size_t j(0); j < profileVector.size(); j++) {
        profileSum.designPower += profileVector.at(j).designPower;
        profileSum.maxMissionPower += profileVector.at(j).maxMissionPower;
    }
    //Mission Loads are also added via the vector
    for (size_t i(0); i < profileVector.at(0).missionPower.size(); i++) {
        //Creating the Mission Loads
        profileSum.missionPower.push_back(missionLoads());
    }
    for (size_t j(0); j < profileVector.size(); j++) {
        for (size_t i(0); i < profileVector.at(j).missionPower.size(); i++) {
            profileSum.missionPower.at(i).baseLoad += profileVector.at(j).missionPower.at(i).baseLoad;
            for (size_t n(0); n < profileVector.at(j).missionPower.at(i).peakLoad.size(); n++) {
                //Create a new peak, only if not already present
                if (profileSum.missionPower.at(i).peakLoad.size() < j + 1) {
                    profileSum.missionPower.at(i).peakLoad.push_back(profileVector.at(j).missionPower.at(i).peakLoad.at(n));
                    profileSum.missionPower.at(i).timePeak.push_back(profileVector.at(j).missionPower.at(i).timePeak.at(n));
                } else {
                    if (!accuracyCheck(profileSum.missionPower.at(i).timePeak.at(j), profileVector.at(j).missionPower.at(i).timePeak.at(n), ACCURACY_HIGH)) {
                        profileSum.missionPower.at(i).peakLoad.push_back(profileVector.at(j).missionPower.at(i).peakLoad.at(n));
                        profileSum.missionPower.at(i).timePeak.push_back(profileVector.at(j).missionPower.at(i).timePeak.at(n));
                    } else {
                        profileSum.missionPower.at(i).peakLoad.back() += profileVector.at(j).missionPower.at(i).peakLoad.at(n);
                    }
                }
            }
        }
    }
    return profileSum;
}

void cleanPowerPeaks(powerProfile *uncleanedPowerProfile) {
    for (size_t i(0); i < uncleanedPowerProfile->missionPower.size(); i++) {
        // NOTE (aigner#1#): ATTENTION: Summarizing the same peaks (assumption: peaks with the same duration overlap --> are added)
        for (size_t j(0); j < uncleanedPowerProfile->missionPower.at(i).peakLoad.size(); j++) {
            for (size_t k(j + 1); k < uncleanedPowerProfile->missionPower.at(i).peakLoad.size(); k++) {
                // Check whether two peaks have the same duration and therefore belong together.
                if (accuracyCheck(uncleanedPowerProfile->missionPower.at(i).timePeak.at(j), uncleanedPowerProfile->missionPower.at(i).timePeak.at(k), ACCURACY_HIGH)) {
                    //Adding the two peaks to a single case
                    uncleanedPowerProfile->missionPower.at(i).peakLoad.at(j) += uncleanedPowerProfile->missionPower.at(i).peakLoad.at(k);
                    uncleanedPowerProfile->missionPower.at(i).peakLoad.at(k) -= uncleanedPowerProfile->missionPower.at(i).peakLoad.at(k);
                }
            }
        }
        //Deletion of superfluous zero peaks
        for (size_t j(0); j < uncleanedPowerProfile->missionPower.at(i).peakLoad.size(); j++) {
            if (accuracyCheck(uncleanedPowerProfile->missionPower.at(i).peakLoad.at(j), 0., ACCURACY_HIGH)) {
                uncleanedPowerProfile->missionPower.at(i).peakLoad.erase(uncleanedPowerProfile->missionPower.at(i).peakLoad.begin() + j);
                uncleanedPowerProfile->missionPower.at(i).timePeak.erase(uncleanedPowerProfile->missionPower.at(i).timePeak.begin() + j);
                j--;
            }
        }
    }
}
