/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <string>
#include "virtualSystemsDesign.h"
#include "../toolinfo.h"
#include <moduleBasics/plot.h>
#include <matplot/matplot.h>
#include <cstdlib>

void virtual_systems_design::generate_plot_data() {
    //generate mission plot
    generate_mission_plot();
    //generate accumulated power plot
    accumulated_power_plot();
}

void virtual_systems_design::generate_mission_plot() {
    // Prepare file paths
    std::string plotDataFile = rtIO_->getCsvFilesDir() + "/" + data_->configuration.theProgramName + "_mission_steps_plot.csv";
    std::ofstream plot(plotDataFile);
    if (!plot) {
        myRuntimeInfo->err << plotDataFile + " could not be opened!" << std::endl;
        exit(1);
    }

    // Write headers
    plot << "# " + win2lin(rtIO_->getCsvFilesDir()) + "/" + data_->configuration.theProgramName + "_mission_steps_plot.csv" << std::endl;
    plot << "# Mission profile from the mission XML file" << std::endl;
    plot << "# (1)Step; (2)Missionsphase; (3)Missionsmode; (4)endAltitude; (5)endFL; (6)endCAS; "
            "(7)endTAS; (8)endMach; (9)pointAltitude; (10)pointTemperature; (11)pointCAS; "
            "(12)pointTAS; (13)pointMach; (14)endTemperature" << std::endl;

    // Prepare data vectors
    std::vector<double> steps, endFL, endTemperature, endMach, endCAS, endTAS;

    for (size_t i = 0; i < data_->mission.Steps.size(); ++i) {
        const auto& step = data_->mission.Steps[i];
        double flightLevel = convertUnit(METER, FOOT, step.endAltitude) / 100.0;

        // CSV output
        plot << i + 1 << "; " << step.missionPhase << "; " << step.modeStep << "; "
             << step.endAltitude << "; " << flightLevel << "; " << step.endCAS << "; "
             << step.endTAS << "; " << step.endMach << "; " << step.Point.Altitude << "; "
             << step.Point.Temperature << "; " << step.Point.CAS << "; " << step.Point.TAS << "; "
             << step.Point.Mach << "; " << step.endTemperature << "; " << std::endl;

        // Collect plot data
        steps.push_back(i + 1);
        endFL.push_back(flightLevel);
        endTemperature.push_back(step.endTemperature);
        endMach.push_back(step.endMach);
        endCAS.push_back(step.endCAS);
        endTAS.push_back(step.endTAS);
    }

    plot.close();

    // Plot
    auto fig = matplot::figure(true);
    fig->title("Mission Profile");

    auto ax1 = fig->add_subplot(2, 1, 0);
    auto ax2 = fig->add_subplot(2, 1, 1);

    ax1->grid(true);
    ax2->grid(true);

    ax1->plot(steps, endFL, "-b", steps, endTemperature, "--k");
    ax1->ylabel("Altitude [FL]");
    ax1->y_axis().tick_label_format("%4.0f");
    ax1->ylim({0, 500});
    ax1->y2_axis().visible(true);
    ax1->y2_axis().label("Temperature [K]");
    ax1->y2_axis().limits({200, 300});
    ax1->y2_axis().tick_label_format("%4.0f");
    ax1->xticks(steps);
    matplot::legend(ax1, {"Flight level", "Temperature"});
    ax1->legend()->location(matplot::legend::general_alignment::topright);

    ax2->hold(true);
    ax2->plot(steps, endMach, "-b");
    ax2->plot(steps, endCAS, "-k")->use_y2(true);
    ax2->plot(steps, endTAS, "--k")->use_y2(true);
    ax2->xlabel("Mission Step");
    ax2->ylabel("Mach Number [-]");
    ax2->y_axis().tick_label_format("%1.1f");
    ax2->ylim({0, 1});
    ax2->y2_axis().label("Flight Speed [m/s]");
    ax2->y2_axis().limits({0, 500});
    ax2->y2_axis().tick_label_format("%3.0f");
    ax2->xticks(steps);
    matplot::legend(ax2, {"Mach number", "CAS", "TAS"});
    ax2->legend()->location(matplot::legend::general_alignment::topright);

    // Save plot
    std::string plotFilePath = rtIO_->getPlotDir() + "/" + data_->configuration.theProgramName + "_mission_steps_plot.svg";
    // Replace backslashes with forward slashes
    std::replace(plotFilePath.begin(), plotFilePath.end(), '\\', '/');
    matplot::save(plotFilePath);
}

void virtual_systems_design::accumulated_power_plot() {
    std::ofstream plot;
    std::string plotDataFile(rtIO_->getCsvFilesDir() + "/" + data_->configuration.theProgramName + "_mission_power_" + "virtual_accumulated_power" + "_plot.csv");
    plot.open(plotDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << plotDataFile + " could not be opened!" << std::endl;;
        exit(1);
    } else {
        plot << "# " + plotDataFile << std::endl;
        /* Adjust data structure: **/
        plot << "# (1) MissionsStep; (2) BleedAir [kg/s]; (3) Electric [W]; (4) Hydraulic [W] " << std::endl;
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            plot << i+1 << "; ";
            plot << data_->data.totalConsumption.Bleed.missionPower.at(i).baseLoad << ";";
            plot << data_->data.totalConsumption.Electric.at(0).missionPower.at(i).baseLoad << ";";
            plot << data_->data.totalConsumption.Hydraulic.at(0).missionPower.at(i).baseLoad << ";";
            plot << std::endl;
        }
        plot.close();
    }

    // Prepare data for plotting
    std::vector<double> missionSteps, bleedAirLoad, electricLoad, hydraulicLoad;
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        missionSteps.push_back(i + 1);
        bleedAirLoad.push_back(data_->data.totalConsumption.Bleed.missionPower.at(i).baseLoad);
        electricLoad.push_back(data_->data.totalConsumption.Electric.at(0).missionPower.at(i).baseLoad / 1000.);
        hydraulicLoad.push_back(data_->data.totalConsumption.Hydraulic.at(0).missionPower.at(i).baseLoad / 1000.);
    }

    // Plot with matplotplusplus
    auto fig = matplot::figure(true);//generate figure in quiet mode
    fig->title("Accumulated Power Requirements of the Systems");
    auto ax = fig->current_axes();
    //add grid
    ax->grid(true);
    //hold plots
    ax->hold(true);

    //Labels
    ax->ylabel("Power, kW");
    ax->xticks(missionSteps);
    ax->xlabel("Mission Step");
    ax->y2_axis().label("Mass Flow, kg/s");
    ax->ytickformat("%.2f");
    ax->y2tickformat("%.2f");

    //Electric
    ax->plot(missionSteps, electricLoad, "-b");

    //Hydraulic
    ax->plot(missionSteps, hydraulicLoad, "-k");

    //Bleed Air
    ax->plot(missionSteps, bleedAirLoad, "--r")->use_y2(true);

    //y-axis scaling
    double BA_max = get_y_axis_scaling_BA(&data_->data.totalConsumption.Bleed);
    ax->y2_axis().limits({0, 1.1 * BA_max}); //extra 10% room for legend
    double EHC_max = get_y_axis_scaling_EHC(&data_->data.totalConsumption.Hydraulic.at(0), &data_->data.totalConsumption.Electric.at(0));
    ax->ylim({0, EHC_max});

    //legend
    matplot::legend(ax, {"Electric Power", "Hydraulic Power", "Bleed Air", "", "", ""});
    ax->legend()->location(matplot::legend::general_alignment::topleft);
    ax->legend()->num_rows(1);
    ax->legend()->font_size(7);
    ax->legend()->box(false);

    // Save the plot
    std::string plotFilePath = rtIO_->getPlotDir() + "/" + data_->configuration.theProgramName + "_virtual_accumulated_power_plot.svg";
    // Replace backslashes with forward slashes
    std::replace(plotFilePath.begin(), plotFilePath.end(), '\\', '/');
    matplot::save(plotFilePath);
}

double virtual_systems_design::get_y_axis_scaling_BA(const powerProfile *bleedAir) {
    double maxBA_tmp = 0.0;
    double BA_max_intern = 0.0;
    int maxSteps = static_cast<int>(data_->mission.Steps.size()) - 1;
    for (int j = 0; j <= maxSteps; j++) {
        double BA_tmp = fabs(bleedAir->missionPower.at(j).baseLoad);
        if (BA_tmp > maxBA_tmp) {
            maxBA_tmp = BA_tmp;
        }
        //check peaks
        for (size_t k = 0; k < bleedAir->missionPower.at(j).peakLoad.size(); k++) {
            double BA_peak_tmp = fabs(bleedAir->missionPower.at(j).peakLoad.at(k));
            if (BA_peak_tmp > maxBA_tmp) {
                maxBA_tmp = BA_peak_tmp;
            }
        }
    }
    if (maxBA_tmp < 5) {
        BA_max_intern = static_cast<int>(std::ceil(maxBA_tmp));
    } else {
        BA_max_intern = static_cast<int>(std::ceil(maxBA_tmp / 5.0) * 5);

    }
    return BA_max_intern;
}

double virtual_systems_design::get_y_axis_scaling_EHC(const powerProfile *hydraulicSum, const powerProfile *electricSum) {
    double maxE_tmp = 0.0;
    double maxH_tmp = 0.0;
    double maxEHC_tmp = 0.0;
    double EHC_max_intern = 0.0;

    int maxSteps = data_->mission.Steps.size() - 1;
    for (int j = 0; j <= maxSteps; j++) {
        double E_tmp = fabs(electricSum->missionPower.at(j).baseLoad) / 1000;
        double H_tmp = fabs(hydraulicSum->missionPower.at(j).baseLoad) / 1000;
        if (E_tmp > maxE_tmp) {
            maxE_tmp = E_tmp;
        }
        if (H_tmp > maxH_tmp) {
            maxH_tmp = H_tmp;
        }
        //check peaks
        for (size_t k = 0; k < electricSum->missionPower.at(j).peakLoad.size(); k++) {
            double E_peak_tmp = fabs(electricSum->missionPower.at(j).peakLoad.at(k)) / 1000;
            if (E_peak_tmp > maxE_tmp) {
                maxE_tmp = E_peak_tmp;
            }
        }
        for (size_t k = 0; k < hydraulicSum->missionPower.at(j).peakLoad.size(); k++) {
            double H_peak_tmp = fabs(hydraulicSum->missionPower.at(j).peakLoad.at(k)) / 1000;
            if (H_peak_tmp > maxH_tmp) {
                maxH_tmp = H_peak_tmp;
            }
        }
    }
    maxEHC_tmp = std::max(maxE_tmp, maxH_tmp);
    if (maxEHC_tmp < 5) {
        EHC_max_intern = static_cast<int>(std::ceil(maxEHC_tmp));
    } else if (maxEHC_tmp < 50) {
        EHC_max_intern = static_cast<int>(std::ceil(maxEHC_tmp / 15.0) * 15);
    } else {
        EHC_max_intern = static_cast<int>(std::ceil(maxEHC_tmp / 150.0) * 150);
    }
    return EHC_max_intern;
}
