/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "virtualSystemsDesign.h"

void virtual_systems_design::initialize() {
    /* Get config file */
    const node& config = rtIO_->moduleConfig;
    /* Initializing IOData */
    data_ = std::make_shared<systemsIOData>(*rtIO_, config, rtIO_->getMissionDataDir());
    /* Initialize virtual systems */
    myRuntimeInfo->out << "Initialize systems... " << std::endl;
    initialize_virtual_systems();
    /* Read wing, fuselage, and accommodation (required for operator items) */
    data_->readFuselages(rtIO_->acxml, rtIO_->getGeometryDir());
    data_->readWing(rtIO_->acxml, rtIO_->getAirfoilDataDir());
    data_->readAccomodation(rtIO_->acxml);
}

void virtual_systems_design::run() {
    /* Calculate operator items */
    myRuntimeInfo->out << "Get operator items masses... " << std::endl;
    myWeightsPt = new weightsAndCGs(data_);
    myWeightsPt->getOperatorItems();
    /* Summarize virtual systems mass and power requirement */
    myRuntimeInfo->out << "Sum system masses... " << std::endl;
    sum_systems();
}

void virtual_systems_design::update() {
    /* update acxml (system mass) */
    if (data_->configuration.designCase == true) {
        myRuntimeInfo->out << "Update systems mass in acxml..." << std::endl;
        data_->updateMassProperties();
    }
    /* update mission xml */
    myRuntimeInfo->out << "Update mission xml... " << std::endl;
    data_->update_mission_xml_virtual(rtIO_->getMissionDataDir());
}

void virtual_systems_design::report() {
    /* generate reports */
    if (rtIO_->plotOn) {
        myRuntimeInfo->out << "Generate plots ..." << std::endl;
        generate_plot_data();
    }
    myRuntimeInfo->out << "Write reports ..." << std::endl;
    set_html_body();
    report_.generateReports(data_->configuration.theProgramName);
}

void virtual_systems_design::save() {
    //not used on this level
}

/* METHODS */
void virtual_systems_design::initialize_virtual_systems() {
    for (int i = 0; i < data_->configuration.numberOfVirtualSystems; i++) {
        std::stringstream nrATA;
        data_->Systems.virtualSystem.push_back(new aircraftSystem(data_->configuration.virtualSystem.at(i).systemName,
                                                                    data_->configuration.virtualSystem.at(i).systemOperating, data_, *rtIO_));
        nrATA << "ATA-virtual" << i + 1;
        data_->Systems.virtualSystem.back()->ATA = nrATA.str();
        data_->Systems.virtualSystem.back()->Mass = data_->configuration.virtualSystem.at(i).Mass;
        data_->Systems.virtualSystem.back()->refPoint = data_->configuration.virtualSystem.at(i).refPoint;
        /*Write power requirements to power profiles*/
        for (size_t j = 0; j < data_->mission.Steps.size(); j++) {
            /*Differentiate between mission segments*/
            if (data_->mission.Steps.at(j).missionPhase == "Departure") {
                //departure
                data_->Systems.virtualSystem.back()->Electric.at(0).missionPower.at(j).baseLoad = data_->configuration.virtualSystem.at(i).powerElectric[0];
                data_->Systems.virtualSystem.back()->Hydraulic.at(0).missionPower.at(j).baseLoad = data_->configuration.virtualSystem.at(i).powerHydraulic[0];
                data_->Systems.virtualSystem.back()->bleedAir.missionPower.at(j).baseLoad = data_->configuration.virtualSystem.at(i).powerBleed[0];
            } else if (data_->mission.Steps.at(j).missionPhase == "Cruise") {
                //cruise
                data_->Systems.virtualSystem.back()->Electric.at(0).missionPower.at(j).baseLoad = data_->configuration.virtualSystem.at(i).powerElectric[1];
                data_->Systems.virtualSystem.back()->Hydraulic.at(0).missionPower.at(j).baseLoad = data_->configuration.virtualSystem.at(i).powerHydraulic[1];
                data_->Systems.virtualSystem.back()->bleedAir.missionPower.at(j).baseLoad = data_->configuration.virtualSystem.at(i).powerBleed[1];
            } else {
                //approach
                data_->Systems.virtualSystem.back()->Electric.at(0).missionPower.at(j).baseLoad = data_->configuration.virtualSystem.at(i).powerElectric[2];
                data_->Systems.virtualSystem.back()->Hydraulic.at(0).missionPower.at(j).baseLoad = data_->configuration.virtualSystem.at(i).powerHydraulic[2];
                data_->Systems.virtualSystem.back()->bleedAir.missionPower.at(j).baseLoad = data_->configuration.virtualSystem.at(i).powerBleed[2];
            }
        }
    }
}

void virtual_systems_design::sum_systems() {
    /* Sum up mass and CoG */
    for (size_t i = 0; i < data_->Systems.virtualSystem.size(); i++) {
        /*Total Mass*/
        data_->data.Mass.totalSystemMass += data_->Systems.virtualSystem.at(i)->Mass;
        /*Total CoG*/
        for (size_t i = 0; i < data_->Systems.virtualSystem.size(); i++) {
            for (size_t j = 0; j < 2; j++) {
                data_->data.CofG.total[j] += data_->Systems.virtualSystem.at(i)->refPoint[j] * data_->Systems.virtualSystem.at(i)->Mass;
            }
        }
    }
    // add operator items
    data_->data.Mass.totalSystemMass += (data_->data.Mass.waterAndChemicalsMass + data_->data.Mass.residualFuelMass);
    for (int i = 0; i < 2; i++) {
        data_->data.CofG.total[i] += (data_->data.CofG.waterAndChemicals[i] * data_->data.Mass.waterAndChemicalsMass);
        data_->data.CofG.total[i] += (data_->data.CofG.residualFuel[i] * data_->data.Mass.residualFuelMass);
    }
    // devide CoG by total mass
    for (int i = 0; i < 2; i++) {
        data_->data.CofG.total[i] /= data_->data.Mass.totalSystemMass;
    }
    /*Sum up required power*/
    data_->data.totalConsumption.Hydraulic.resize(1);
    data_->data.totalConsumption.Electric.resize(1);
    initialize_total_consumption(data_->data.totalConsumption.Hydraulic.at(0));
    initialize_total_consumption(data_->data.totalConsumption.Electric.at(0));
    initialize_total_consumption(data_->data.totalConsumption.Bleed);
    //loop over all mission steps
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        //loop over all virtual systems
        for (size_t j = 0; j < data_->Systems.virtualSystem.size(); j++) {
            //add the base loads at this mission step
            data_->data.totalConsumption.Electric.at(0).missionPower.at(i).baseLoad += data_->Systems.virtualSystem.at(j)->Electric.at(0).missionPower.at(i).baseLoad;
            data_->data.totalConsumption.Hydraulic.at(0).missionPower.at(i).baseLoad += data_->Systems.virtualSystem.at(j)->Hydraulic.at(0).missionPower.at(i).baseLoad;
            data_->data.totalConsumption.Bleed.missionPower.at(i).baseLoad += data_->Systems.virtualSystem.at(j)->bleedAir.missionPower.at(i).baseLoad;
        }
    }
}

void virtual_systems_design::initialize_total_consumption(auto& system) {
    system.missionPower.clear();
    system.designPower = 0.0;
    system.maxMissionPower = 0.0;
    system.missionPower.resize(data_->mission.Steps.size(), missionLoads());
}
