/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2025 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef SYSTEMS_DESIGN_SRC_VIRTUAL_VIRTUALSYSTEMSDESIGN_H_
#define SYSTEMS_DESIGN_SRC_VIRTUAL_VIRTUALSYSTEMSDESIGN_H_

#include <memory>
#include <moduleBasics/strategySelector.h>
#include <moduleBasics/report.h>
#include <moduleBasics/plot.h>
#include "../common/systemsIOData.h"
#include "../common/weightsAndCGs.h"

class virtual_systems_design : public Strategy {
 public:
    explicit virtual_systems_design(std::shared_ptr<RuntimeIO> rtIO): rtIO_(rtIO), report_(rtIO) {}
    ~virtual_systems_design() = default;
    /* Inherited functions from class strategy*/
    /** \brief Initialize virtual system from config. */
    void initialize();
    /** \brief Run */
    void run();
    /** \brief update acxml and mission xml */
    void update();
    /** \brief create reports */
    void report();
    /** \brief save */
    void save();

    /** \brief */
    void initialize_virtual_systems();
    /** \brief */
    void sum_systems();

    /** \brief initializes power profiles according to mission steps */
    void initialize_total_consumption(auto& system);

    /* Reporting elements */
    /**
    * \brief Defines the content of the HTML Report body. The content depends on wether the sizing or study mode
    * is used. Since that masses or only calculated in the sizing mode, they are also only written in this mode.
    * \return void
    */
    void set_html_body();

    /* Plotting elements */
    /** \brief Calls all plot generation functions */
    void generate_plot_data();

    /** \brief plots the mission data (altitude, temperature, mach, CAS, TAS) over the mission
     * steps. This data is read from the mission file and not altered in systemsDesign. (Plot can
     * serve as a reference when reading mission plots over mission steps of the systems)*/
    void generate_mission_plot();

    /** \brief Generates plot data and plot for total required mission power */
    void accumulated_power_plot();

    /** \brief Calculates y-axis scaling for electric / hydraulic power based on the maximum value and including room for a legend
     * \param hydraulicSum pointer to the hydraulic power profile that will be plotted
     * \param electricSum pointer to the electric power profile that will be plotted
     * \return double: maximum value for power y-axis
     */
    double get_y_axis_scaling_EHC(const powerProfile *hydraulicSum, const powerProfile *electricSum);

    /** \brief Calculates y-axis scaling for bleed air mass flow based on the maximum value and including room for a legen
     * \note Depending on values the extra room for the legend is sometimes smaller than that for electric/hydraulic power. It can be useful to add
     * an extra 10-20% of room by increasing the upper y-axis limit by this factor.
     * \param bleedAir pointer to the bleed air power profile
     * \return double: maximum value for mass flow y-axis
     */
    double get_y_axis_scaling_BA(const powerProfile *bleedAir);

 private:
    std::shared_ptr<RuntimeIO> rtIO_; /*<-- RuntimeIO object*/
    std::shared_ptr<systemsIOData> data_; /*<-- IOData object*/
    Report report_; /*Report object*/
    weightsAndCGs *myWeightsPt; /*Pointer to weightsAndCGs object*/
};

#endif //SYSTEMS_DESIGN_SRC_VIRTUAL_VIRTUALSYSTEMSDESIGN_H_
