# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard libraries.
import numpy as np
from matplotlib import pyplot as plt

# Import own libraries.
import pyenergycarriers
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculateparameterendcap import (
    calculate_parameter_end_cap)
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculatethicknesswall import (
    calculate_thickness_wall)
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculateparametertruncatedcone import (
    calculate_parameter_truncated_cone)
from src.tube_and_wing.empirical.tank_design_tu_berlin.general.calculatecenterofgravity import (
    calculate_center_of_gravity)
from src.tube_and_wing.empirical.tank_design_tu_berlin.liquid_hydrogen.calculatezposition import (
    calculate_z_position)


def calculate_conical_tank(tank_entity, dict_ac_data, dict_tank_design, runtime_output):
    """Calculate volume and masses of conical tank.

        [Add some text here.]

    :param str tank_entity: ...
    :param dict dict_ac_data: Dict containing parameter and values from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dict containing results from tank design
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_tank_design: Dict containing results from tank design
    """
    # Extract data from 'dict_ac_data'.
    mass_pumps = dict_ac_data['mass_pumps']
    mass_baffle = dict_ac_data['mass_baffle']
    type_end_cap = dict_ac_data['type_end_cap']
    density_wall = dict_ac_data['density_wall']
    material_wall = dict_ac_data['material_wall']
    internal_pressure = dict_ac_data['internal_pressure']
    density_insulation = dict_ac_data['density_insulation']
    mass_vapor_barrier = dict_ac_data['mass_vapor_barrier']
    factor_volume_allowance = dict_ac_data['factor_volume_allowance']
    tensile_stress_allowable = dict_ac_data['tensile_stress_allowable']
    thickness_wall_calculation_method = dict_ac_data['thickness_wall_calculation_method']

    thickness_wall = 0.003
    thickness_insulation = 0.20
    thickness_wall_and_insulation = thickness_wall + thickness_insulation
    thickness_wall_minimum_aluminum = 0.003  # AD-2000: minimum wall thickness AL-2219 is 3 mmm.

    # Get density of liquid hydrogen.
    for key, value in dict_ac_data['energy_carrier_name'].items():
        if value == 'liquid_hydrogen':
            density_liquid_hydrogen = dict_ac_data['energy_carrier_density'].get(
                key.replace('energy_carrier_name', 'energy_carrier_density'))
            break
    liquid_hydrogen_volumetric_energy_density = (
        pyenergycarriers.EnergyCarrier('liquid_hydrogen', density_liquid_hydrogen).volumetric_energy_density)
    liquid_hydrogen_gravimetric_energy_density = (
        pyenergycarriers.EnergyCarrier('liquid_hydrogen', density_liquid_hydrogen).gravimetric_energy_density)

    """Calculation. """
    max_length = round(dict_tank_design[tank_entity]['geometry']['maximum_length'],2)
    # Note: The calculated diameters are actual sizes. This means that the thicknesses for the tank wall and insulation
    # have already been taken into account following the subsequent assumption.
    # Assumption: Maximum outer tank diameter equals maximum useable fuselage diameter minus twice the
    # thickness of insulation and wall. The calculated maximum outer diameter is assumed to be the actual tank
    # diameter.
    max_diameter_front = (dict_tank_design[tank_entity]['geometry']['maximum_diameter_front']
                          - 2*thickness_wall_and_insulation)
    max_diameter_rear = (dict_tank_design[tank_entity]['geometry']['maximum_diameter_rear']
                         - 2*thickness_wall_and_insulation)
    diameters = np.linspace(max_diameter_front, max_diameter_rear,int(max_length*100))

    if dict_tank_design[tank_entity]['tank_position'] == 'tail_cone':
        # Special case since the geometry of the tail remains the same across the iteration loops.
        # Find position of end caps.
        # The position of the bulkhead is known. This is assumed to be the rearmost position of a tank component,
        # i.e. the tank must end there. Due to this, the rear end cap must start slightly before this position, so
        # that it does not extend beyond the position of the bulkhead. The beginning is determined iteratively.
        # Starting from the position of the bulkhead, the distance (dist_to_rear) is increased until it equals the
        # minimum limit (boundary_rear). The latter depends on the type of end cap. In the case of a hemispherical
        # design, the required distance to the end cap equals half the diameter. For a torispherical end, it must be
        # at least equal to the determined length of the end.

        # Set iteration values.
        idx_rear = 1  # beginning of rear end cap
        idx_front = 0  # beginning of front end cap
        increment = 0.01
        boundary_rear = 100
        boundary_front = 100
        distance_to_rear = 0
        distance_to_front = 0

        # Determine foremost position of front tank end cap.
        while boundary_front > distance_to_front:
            boundary_front = calculate_parameter_end_cap(runtime_output, diameters[idx_front], type_end_cap,
                                                         thickness_wall, thickness_insulation, density_wall,
                                                         density_insulation)[3]
            maximum_tank_diameter_front = diameters[idx_front]
            idx_front += 1
            distance_to_front += increment

        # Determine rearmost position of rear tank end cap.
        while boundary_rear > distance_to_rear:
            boundary_rear = calculate_parameter_end_cap(runtime_output, diameters[-idx_rear], type_end_cap,
                                                        thickness_wall, thickness_insulation, density_wall,
                                                        density_insulation)[3]
            maximum_tank_diameter_rear = diameters[-idx_rear]
            idx_rear += 1
            distance_to_rear += increment

    else:
        runtime_output.print('Normal conical tank.')

    """Calculate wall thickness."""
    # Assumption: Only dependent on diameter (if pressure etc. constant) -> calculation for biggest diameter.
    maximum_diameter = max(maximum_tank_diameter_front, maximum_tank_diameter_rear)
    thickness_wall = calculate_thickness_wall(runtime_output, maximum_diameter, internal_pressure,
                                              tensile_stress_allowable, thickness_wall_calculation_method)
    if (material_wall == 'aluminum') and (thickness_wall < thickness_wall_minimum_aluminum):
        thickness_wall = thickness_wall_minimum_aluminum

    """Calculate volumes, masses, and lengths of tank parts."""
    fuselage_geometry = dict_tank_design['geometry']['fuselage']
    # Calculate maximum length of conical tank part w/o end cap.
    # Note that the lengths of the end caps already take into account the tank wall and the insulation thickness.
    length_conical_tank_excl_end_caps = max_length - distance_to_front - distance_to_rear

    # 1) Front end cap.
    (volume_tank_front_end_cap, mass_wall_front_end_cap, mass_insulation_front_end_cap,
     length_front_end_cap) = calculate_parameter_end_cap(runtime_output, maximum_tank_diameter_front, type_end_cap,
                                                         thickness_wall, thickness_insulation, density_wall,
                                                         density_insulation)

    dict_tank_design[tank_entity]['geometry']['front_end_cap'] = {
        'section_name': 'front_end_cap',
        'shape': 'circular',
        'height': 0.0,
        'width': 0.0,
        'length': length_front_end_cap,
        'global_position': {
            'x': dict_tank_design[tank_entity]['geometry']['idx_front']/100,
            'y': 0.0,
            'z': None
        }
    }

    # 2) Front tank section (begin of conical section).
    (volume_tank_conical_section, mass_wall_conical_section,
     mass_insulation_conical_section) = calculate_parameter_truncated_cone(runtime_output, maximum_tank_diameter_front,
                                                                           maximum_tank_diameter_rear,
                                                                            length_conical_tank_excl_end_caps,
                                                                            thickness_wall, thickness_insulation,
                                                                            density_wall, density_insulation)

    dict_tank_design[tank_entity]['geometry']['front_tank_section'] = {
        'section_name': 'front_tank_section',
        'shape': 'circular',
        'height': maximum_tank_diameter_front,
        'width': maximum_tank_diameter_front,
        'length': 0.0,
        'global_position': {
            'x': (dict_tank_design[tank_entity]['geometry']['front_end_cap']['global_position']['x']
                  + length_front_end_cap),
            'y': 0.0,
            'z': None
        }
    }
    idx_front_tank_section = round(
        dict_tank_design[tank_entity]['geometry']['front_tank_section']['global_position']['x']*100)
    dict_tank_design[tank_entity]['geometry']['front_tank_section']['global_position']['z'] = calculate_z_position(
        fuselage_geometry, idx_front_tank_section)
    # Set z-coordinate of front end cap.
    dict_tank_design[tank_entity]['geometry']['front_end_cap']['global_position']['z'] = (
        dict_tank_design[tank_entity]['geometry']['front_tank_section']['global_position']['z'])

    # 3) Rear tank section (end of conical section).
    dict_tank_design[tank_entity]['geometry']['rear_tank_section'] = {
        'section_name': 'rear_tank_section',
        'shape': 'circular',
        'height': maximum_tank_diameter_rear,
        'width': maximum_tank_diameter_rear,
        'length': 0.0,
        'global_position': {
            'x': (dict_tank_design[tank_entity]['geometry']['front_tank_section']['global_position']['x']
                  + length_conical_tank_excl_end_caps),
            'y': 0.0,
            'z': None
        }
    }

    idx_rear_tank_section = round(
        dict_tank_design[tank_entity]['geometry']['rear_tank_section']['global_position']['x']*100)
    dict_tank_design[tank_entity]['geometry']['rear_tank_section']['global_position']['z'] = calculate_z_position(
        fuselage_geometry, idx_rear_tank_section)

    # 4) Rear end cap.
    (volume_tank_rear_end_cap, mass_wall_rear_end_cap, mass_insulation_rear_end_cap,
     length_rear_end_cap) = calculate_parameter_end_cap(runtime_output, maximum_tank_diameter_rear, type_end_cap,
                                                        thickness_wall, thickness_insulation, density_wall,
                                                        density_insulation)

    dict_tank_design[tank_entity]['geometry']['rear_end_cap'] = {
        'section_name': 'rear_end_cap',
        'shape': 'circular',
        'height': 0.0,
        'width': 0.0,
        'length': -length_rear_end_cap,
        'global_position': {
            'x': (dict_tank_design[tank_entity]['geometry']['rear_tank_section']['global_position']['x']
                  + length_rear_end_cap),
            'y': 0.0,
            'z': dict_tank_design[tank_entity]['geometry']['rear_tank_section']['global_position']['z']
        }
    }


    """Calculate overall values."""
    volume_tank_available = volume_tank_front_end_cap + volume_tank_conical_section + volume_tank_rear_end_cap
    volume_tank_usable = volume_tank_available/(1 + factor_volume_allowance)
    energy_hydrogen_available = volume_tank_usable * liquid_hydrogen_volumetric_energy_density
    mass_hydrogen_available = energy_hydrogen_available/liquid_hydrogen_gravimetric_energy_density

    # Tank masses.
    dict_tank_design[tank_entity]['masses'] = {
        'masses_tank_structure': {
            'mass_wall': (mass_wall_front_end_cap + mass_wall_conical_section + mass_wall_rear_end_cap),
            'mass_pumps': mass_pumps,
            'mass_baffle': mass_baffle
            },
        'masses_tank_insulation': {
            'mass_insulation': (mass_insulation_front_end_cap + mass_insulation_conical_section
                                + mass_insulation_rear_end_cap),
            'mass_vapor_barrier': mass_vapor_barrier
            },
        'mass_tank_overall': None
    }
    mass_tank_structure = sum(dict_tank_design[tank_entity]['masses']['masses_tank_structure'].values())
    mass_tank_insulation = sum(dict_tank_design[tank_entity]['masses']['masses_tank_insulation'].values())
    dict_tank_design[tank_entity]['masses']['mass_tank_overall'] = mass_tank_structure + mass_tank_insulation
    # Tank length.
    # length_tank_overall = length_front_end_cap + length_conical_tank_excl_end_caps + length_rear_end_cap
    dict_tank_design[tank_entity]['geometry']['length_tank_overall'] = float(length_front_end_cap
                                                                             + length_conical_tank_excl_end_caps
                                                                             + length_rear_end_cap)
    # Calculate gravimetric efficiency tank.
    gravimetric_efficiency = mass_hydrogen_available/(mass_hydrogen_available
                                                      + dict_tank_design[tank_entity]['masses']['mass_tank_overall'])
    dict_tank_design[tank_entity]['gravimetric_efficiency'] = float(gravimetric_efficiency)

    # Inertia (set inertia to zero).
    dict_tank_design[tank_entity]['geometry']['total'] = {}
    dict_tank_design[tank_entity]['geometry']['total']['inertia'] = {}  
    inertia_list = ['j_xx', 'j_yy', 'j_zz', 'j_xy', 'j_xz', 'j_yx', 'j_yz', 'j_zx', 'j_zy']
    for inertia in inertia_list:
        dict_tank_design[tank_entity]['geometry']['total']['inertia'][inertia] = 0.0

    # Calculate center of gravity.
    dict_tank_design[tank_entity]['geometry']['total']['center_of_gravity'] = calculate_center_of_gravity(
        runtime_output, dict_tank_design[tank_entity]['tank_shape'],
         dict_tank_design[tank_entity]['geometry']['maximum_diameter_front'],
         dict_tank_design[tank_entity]['geometry']['maximum_diameter_rear'],
         length_conical_tank_excl_end_caps)

    # Calculate position of tank entity (foremost point of tank).
    dict_tank_design[tank_entity]['geometry']['total']['global_position'] = {
        'x': dict_tank_design[tank_entity]['geometry']['front_end_cap']['global_position']['x'],
        'y': dict_tank_design[tank_entity]['geometry']['front_end_cap']['global_position']['y'],
        'z': dict_tank_design[tank_entity]['geometry']['front_end_cap']['global_position']['z'],
    }

    # Direction.
    dict_tank_design[tank_entity]['geometry']['total']['direction'] = {
        'x': 1,
        'y': 0,
        'z': 0
    }

    # Centroid.
    dict_tank_design[tank_entity]['geometry']['total']['centroid'] = {
        'x': dict_tank_design[tank_entity]['geometry']['total']['center_of_gravity']['x'],
        'y': dict_tank_design[tank_entity]['geometry']['total']['center_of_gravity']['y'],
        'z': dict_tank_design[tank_entity]['geometry']['total']['center_of_gravity']['z']
    }

    # Prepare further data output.
    dict_tank_design[tank_entity]['energy_available'] = float(energy_hydrogen_available)
    dict_tank_design[tank_entity]['volume_available'] = float(volume_tank_usable)
    dict_tank_design[tank_entity]['energy_required_for_mission'] = True
    dict_tank_design[tank_entity]['volume_required_for_mission'] = True

    # Debug print.
    runtime_output.debug('Debug: The "calculate_conical_tank" function was successfully executed.')

    return dict_tank_design
