# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2025 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.


# Import standard modules.
import sys
import logging
from sys import argv
import traceback
from termcolor import colored

# Import own modules.
from runmodule import run_module
from src.datapreprocessing import data_preprocessing
from src.datapostprocessing import data_postprocessing


def main():
    """Execute the main program for weight and balance analysis.

    This function serves as the main entry point for performing the weight and balance analysis.
    It goes through the following key steps:
        (1) Preprocessing - Acquire necessary data and paths: Call the 'data_preprocessing' function from
        'datapreprocessing.py' to set up data and routing information.
        (2) Run (main processing) - Execute code depending on method layers: Execute the 'run_module' function from the
        'methodexecutionpackage' library. The 'run_module' function is responsible for the programs primary logic.
        (3) Postprocessing - Write data to the aircraft exchange file and generate plots and reports: Call the
        'data_postprocessing' function from 'datapostprocessing.py' to handle postprocessing tasks. This step receives
        data from both the preprocessing and the main processing step.

    Note: The 'routing_dict' dictionary is used to manage the routing and execution of different program components.

    :raises AttributeError: Raised to handle attribute exceptions
    :raises FileNotFoundError: Raised to handle file not found exceptions
    :raises IndexError: Raised to handle index exceptions
    :raises KeyError: Raised to handle key exceptions
    :raises ModuleNotFoundError: Raised to handle module not found exceptions
    :raises OSError: Raised to handle operating system exceptions
    :raises TypeError: Raised to handle type exceptions
    :raises ValueError: Raised to handle value exceptions
    :raises ZeroDivisionError: Raised to handle zero division exceptions
    :raises Exception: Raised to handle other exceptions
    :return: None
    """
    module_name = "Weight and balance analysis"
    module_config_file = "weight_and_balance_analysis_conf.xml"

    # Initialize exception string and runtime output logger.
    runtime_output = logging.getLogger('module_logger')
    tool_name = "Weight and Balance"
    tool_version = '0.5.0'

    try:
        """Preprocessing: Acquire necessary data and paths."""
        # Run 'data_preprocessing' function from 'datapreprocessing.py'.
        paths_and_names, routing_dict, runtime_output = data_preprocessing(
            module_config_file, tool_version, argv)
        runtime_output.print(f'{module_name} started...')

        """Run: Execute code depending on method layers."""
        # Execute 'run_module' function from 'methodexecutionpackage' library. This function is responsible for the
        # main logic of the program.
        run_output_dict = run_module(
            paths_and_names, routing_dict, runtime_output)

        """Postprocessing: Write data to aircraft exchange file and generate plots and reports."""
        # Run 'data_postprocessing' function from 'datapostprocessing.py' to handle postprocessing tasks. Receives data
        # from preprocessing and main processing step.
        data_postprocessing(paths_and_names, routing_dict,
                            run_output_dict, runtime_output)
        runtime_output.print(f'{module_name} finished.')

    except Exception as e:  # pylint: disable=broad-exception-caught
        # Handle other exceptions.

        # Check if exception raised.
        runtime_output.critical(exception_string_msg(e, tool_name))
        sys.exit(1)


def exception_string_msg(error, tool_name: str) -> str:
    error_type = str(type(error).__name__)
    error_trace = traceback.extract_tb(error.__traceback__)
    error_file, error_line, error_func, error_code = error_trace[-1]
    error_file = error_file.split('/')[-1]

    exception_string = f"{error_type}      : \n"
    exception_string += f"File             : {error_file} \n"
    exception_string += f"Function / Method: {error_func} \n"
    exception_string += f"Line             : {error_line} \n"
    exception_string += f"Code             : {error_code} \n"
    exception_string += f"Error message    : {str(error)} \n"

    return colored(exception_string + "\n" + "                                     "
                                    + f"Main execution of {tool_name} module failed! \n"
                                    + "                                     " + "Program aborted.", 'red')


if __name__ == '__main__':
    main()
